#include <dbsim.h>
#include <node.h>
#include <op.h>
#include <stdio.h>
#include <cost.h>

#ifndef BUFMGR_H
#include <bufmgr.h>
#endif

#include "bfbucket.h"
#include "freeframe.h"

// Constructor.
BufMgr::BufMgr(int numBuffers = DEFAULT_BUFSIZE, Node *owningNode)
{
  int i;


  myNode = owningNode;

  FreeListRecord *tmpFreeListRec;
 
  if (myNode == 0) {
    printf("BufMgr: constructor fed bad myNode.\n");
  }

  numFree = numBuf = numBuffers;

//  numIoComplete = 0;

  // initially no loved or hated pages.
  HatedPages = HatedPagesTail = NIL;
  LovedPages = LovedPagesTail = NIL;

  ClearStat();
  bufPool = new Page[numBuf];
  if ( ! bufPool ) {
    printf("BufMgr() node %d: Couldn't create buffer pool.\n", myNode->id);
  }

  bufTable = new BufDesc[numBuf];
  if (!bufTable ) {
    printf("BufMgr() node %d: Couldn't create buffer table.\n", myNode->id);
  }


  // initially all pages are invalid.

  for (i = 0; i < numBuf; i++) {
    bufTable[i].valid = FALSE;
    tmpFreeListRec = new FreeListRecord;
    tmpFreeListRec->bufIndex = i;
    InsertInvalidPage(tmpFreeListRec);
  }





  hashSize = (numBuf) + 1;
  hashTable = (BFBucket**) new char[sizeof(BFBucket*) * hashSize];
  if ( ! hashTable )  {
    printf("BufMgr() node %d: Couldn't create hashTable.\n", myNode->id);
  }

  for (i = 0; i < hashSize; i++) {
    hashTable[i] = NIL;
//    hashTable[i] = new BFBucket;
//    numBFBuckets++;
  }
  
}

BufMgr::~BufMgr()
{
  
  delete (char*) hashTable;
  delete [numBuf] bufTable;
  delete bufPool;

}


// insert f at tail of list of invalid pages.
// The list is maintained as a doubly-linked
// circular list.
void BufMgr::InsertInvalidPage(FreeListRecord *f) {

  if (bufTable[f->bufIndex].valid == TRUE) {
    printf("InsertInvalidPage:: asked to insert valid page!\n");
  } else if (bufTable[f->bufIndex].reserved == TRUE) {
    printf("InsertInvalidPage:: asked to insert reserved page!\n");
  }
  if (InvalidPages == NIL) {
    InvalidPages = f;
    InvalidPagesTail = f;
    f->next = f->prev = f;
  } else {
    InvalidPages->prev = f;      // circular link from head
    f->prev = InvalidPages;      // circular link to the head
    InvalidPagesTail->next = f;
    f->next = InvalidPagesTail->next;
    InvalidPagesTail = f;
  }
}

// delete f from circular doubly-linked list of invalid pages.
void BufMgr::DeleteInvalidPage(FreeListRecord *f) {

  if (InvalidPages == NIL) {
    printf("BufMgr::DeleteInvalidPage(): deleted from empty list!\n");
  } else if (bufTable[f->bufIndex].valid != FALSE) {
    printf("BufMgr::DeleteInvalidPage(): page was not invalid!\n");
  } else if (InvalidPages == InvalidPagesTail) {
    // only one element
    InvalidPages = InvalidPagesTail = NIL;
  } else {
    f->prev->next = f->next;
    f->next->prev = f->prev;
    if (f == InvalidPagesTail) {
      InvalidPagesTail = InvalidPagesTail->prev;
    } else if (f == InvalidPages) {
      InvalidPages = InvalidPages->next;
    }
  }
}

// return a pointer to an invalid frame.  It should never
// be intransit or fixed.
FreeListRecord *BufMgr::FindInvalidFrame() {

  register FreeListRecord *f;

  static activeCnt = 0;


  activeCnt++;
  if (activeCnt > 1) {
    printf("Gag! %d active FindInvalidFrame() instances!\n", activeCnt);
  }
  f = InvalidPages;
  if (f != NIL) {
    // delete f from the list.
    DeleteInvalidPage(f);
    if (bufTable[f->bufIndex].valid != FALSE) {
      printf("FindInvalidFrame:Node%d: found valid page %d in invalid list!\n",
             myNode->id, f->bufIndex);
      return NIL;
    }
    activeCnt--;
    return f;
  } else {
    activeCnt--;
    return NIL;
  }
}

// insert f at tail of list of hated pages.
// The list is maintained as a doubly-linked
// circular list.
void BufMgr::InsertHatedPage(FreeListRecord *f) {

  if (bufTable[f->bufIndex].hated != HATE) {
    printf("Hey, trying to insert non-hated record in hated list!\n");
    fflush(stdout);
    return;
  }
  bufTable[f->bufIndex].flRec = f;

  if (HatedPages == NIL) {
    HatedPages = f;
    HatedPagesTail = f;
    f->next = f->prev = f;
  } else {
    HatedPages->prev = f;      // circular link from head
    f->next = HatedPages;      // circular link to the head
    HatedPagesTail->next = f;
    f->prev = HatedPagesTail;
    HatedPagesTail = f;
  }
}

// delete f from circular doubly-linked list of Hated pages.
void BufMgr::DeleteHatedPage(FreeListRecord *f) {

  if (f == NIL) {
    printf("DeleteHatedPage: asked to delete NIL record.\n");
    return;
  } else if (bufTable[f->bufIndex].hated != HATE) {
    printf("DeleteHatedPage: asked to delete loved page %d.\n",
	   f->bufIndex);
    return;
  }

  if (HatedPages == NIL) {
    printf("BufMgr::DeleteHatedPage(): deleted from empty list!\n");
    printf("Page to delete was index %d.\n", f->bufIndex);
  } else if (HatedPages == HatedPagesTail) {
    // only one element
    HatedPages = HatedPagesTail = NIL;
  } else {
    f->prev->next = f->next;
    f->next->prev = f->prev;
    if (f == HatedPagesTail) {
      HatedPagesTail = HatedPagesTail->prev;
    } else if (f == HatedPages) {
      HatedPages = HatedPages->next;
    }
  }
}

// return a pointer to an Hated frame that is not
// be intransit or fixed; NIL if no such frame
// exists.
FreeListRecord *BufMgr::FindHatedFrame() {

  register FreeListRecord *f;
  register	int bufIdx;
  BFBucket *htEntry;

   if (HatedPages == NIL) {
     return NIL;
   }

  // do first element outside of while loop, rest within.
  f = HatedPages;
  bufIdx = f->bufIndex;
  htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
  if ((htEntry->inTransit == FALSE) &&
      (bufTable[bufIdx].fixCnt == 0)) {
    if (bufTable[bufIdx].hated != HATE) {
      printf("FindHatedFrame: found a non-hated page in hated list!\n");
      printf("(frame had index %d, valid = %d, hated = %d.\n",
	     f->bufIndex, 
	     bufTable[bufIdx].valid, 
	     bufTable[bufIdx].hated);
           return NIL;
    }
    DeleteHatedPage(f);
    return (f);
  }
  f = f->next;
  while (f != HatedPages) {
    bufIdx = f->bufIndex;
    htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
    if ((htEntry->inTransit == FALSE) &&
      (bufTable[bufIdx].fixCnt == 0)) {
       // got one!
	 if (bufTable[bufIdx].hated != HATE) {
	   printf("FindHatedFrame: found a non-hated page in hated list!\n");
	   printf("(frame had index %d, valid = %d, hated = %d.\n",
		  bufIdx, bufTable[bufIdx].valid, bufTable[bufIdx].hated);
	   return NIL;
	 }
       DeleteHatedPage(f);
       return f;
    }
    f = f->next;
  } 
  return NIL;
}

// return TRUE if all frames on hated list are hated;
// FALSE otherwise.
int BufMgr::CheckHatedFrameList() {

  register FreeListRecord *f;
  register int   bufIdx;

  BFBucket *htEntry;
  int sanityCount = 0;

   if (HatedPages == NIL) {
     return TRUE;
   }

  // do first element outside of while loop, rest within.
  f = HatedPages;
  bufIdx = f->bufIndex;
  htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
    if (bufTable[bufIdx].hated != HATE) {
    return (FALSE);
  }
  f = f->next;
  bufIdx = f->bufIndex;

  while ((f != HatedPages) && (sanityCount < numBuf)) {
    htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
    if (bufTable[bufIdx].hated != HATE) {
      return FALSE;
    }
    f = f->next;
    bufIdx = f->bufIndex;
    sanityCount++;
  } 
  return TRUE;
}

// insert f at tail of list of Loved pages.
// The list is maintained as a doubly-linked
// circular list.
void BufMgr::InsertLovedPage(FreeListRecord *f) {

  if (bufTable[f->bufIndex].hated != LOVE) {
    printf("Hey, tried to insert an unloved page in loved list!\n");
    return;
  }

  bufTable[f->bufIndex].flRec = f;
  if (LovedPages == NIL) {
    LovedPages = f;
    LovedPagesTail = f;
    f->next = f->prev = f;
  } else {
    LovedPages->prev = f;      // circular link from head
    f->next = LovedPages;      // circular link to the head
    LovedPagesTail->next = f;
    f->prev = LovedPagesTail;
    LovedPagesTail = f;
  }
}

// delete f from circular doubly-linked list of Loved pages.
void BufMgr::DeleteLovedPage(FreeListRecord *f) {

  if (LovedPages == NIL) {
    printf("BufMgr::DeleteLovedPage(): deleted from empty list!\n");
  } else if (LovedPages == LovedPagesTail) {
    // only one element
    LovedPages = LovedPagesTail = NIL;
  } else {
    f->prev->next = f->next;
    f->next->prev = f->prev;
    if (f == LovedPagesTail) {
      LovedPagesTail = LovedPagesTail->prev;
    } else if (f == LovedPages) {
      LovedPages = LovedPages->next;
    }
  }
}

// return a pointer to an Loved frame that is not
// be intransit or fixed; NIL if no such frame
// exists.
FreeListRecord *BufMgr::FindLovedFrame() {

  register FreeListRecord *f;
  register int   bufIdx;
  BFBucket *htEntry;

   if (LovedPages == NIL) {
     return NIL;
   }

  // do first element outside of while loop, rest within.
  // this is ugly, but is a result of using a circular
  // list.
  f = LovedPages;
  bufIdx = f->bufIndex;

  htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
  if ((htEntry->inTransit == FALSE) &&
      (bufTable[bufIdx].fixCnt == 0)) {
    if (bufTable[bufIdx].hated != LOVE) {
      printf("FindLovedFrame: found a non-loved page in loved list!\n");
      return NIL;
    }
    DeleteLovedPage(f);
    return (f);
  }
  f = f->next;
  while (f != LovedPages) 
  {
    bufIdx = f->bufIndex;
    htEntry = GetHashTableEntry(bufPool[bufIdx].pid);
    if ((htEntry->inTransit == FALSE) &&
      (bufTable[bufIdx].fixCnt == 0)) 
    {
       // got one!
        if (bufTable[bufIdx].hated != LOVE) 
	{
      	   printf("FindLovedFrame: found a non-loved page in loved list!\n");
           return NIL;
        }
        DeleteLovedPage(f);
        return f;
    }
    f = f->next;
  } 
  return NIL;
}

// same as lookup except return pointer to
// hash table entry and return even if page pid is in transit.
BFBucket *BufMgr::GetHashTableEntry(PID pid) {

  register BFBucket* p;

  int waited = FALSE;  // for debugging

  p = hashTable[Hash(pid)];
  while (p != NIL) {
    if (p->pid.diskId < 0) {
      printf("GetHashTableEntry: got a negative disk %d in pid.\n",
	     p->pid.diskId);
      fflush(stdout);
      p = NIL;
      p->pid.diskId = 10;
      return NIL;
    } else if (p->pid.pageId < 0) {
      printf("GetHashTableEntry: got a negative page id %d in pid.\n");
      fflush(stdout);
      return NIL;
    }
    if ((p->pid.diskId != pid.diskId) ||
	(p->pid.pageId != pid.pageId))  {
      // not in this slot, skip it 
      p = p->next;
    } else {
      return p;
    }
  }

  // there is no mapping for the page 
  printf("Node %d:BufMgr::GetHashTableEntry couldn't find pid %d.%d.\n", 
	 myNode->id, pid.diskId, pid.pageId);
  return NIL;
}

// Write frame "index" to disk.  Sets "inTransit" to
// TRUE before IO, FALSE afterwards.  Also, wakes up
// any waiters on ioComplete when the io is finished.
void BufMgr::WriteFrame(int index) {

  register BFBucket *htEntry;  // the entry corresponding to the frame
  
  // get corresponding hash table entry.
  htEntry = GetHashTableEntry(bufPool[index].pid);

  // mark frame in-transit.
  htEntry->inTransit = TRUE;

  // do the write
  myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
  // Where to account for this cpu and disk usage in stats?  Nowhere for now!
  myNode->disks[bufPool[index].pid.diskId]->use(bufPool[index].pid.pageId,
					        DISK_WRITE, NO_PREFETCH);

  // signal that I/O is complete
  htEntry->inTransit = FALSE;

  // wake up any waiters.
  if (htEntry->ioComplete != NIL) {
     // wake up waiters
     htEntry->ioComplete->set(); 
  }
  noIOWrites++;
}




// given (diskId, pageNumber) pair, find index of
// page in bufPool and descriptor in bufTable (the two
// indices are the same).
//
// 
//
// This is complicated by the following situations:
// 1.  The page could be mapped into the buffer pool but
//    still "in transit" from disk to the buffer pool.
// 2.  The page could be in the buffer pool but in the process
//     of being written out to disk to "flush" the page.  
// It is possible that a
// process could wait() for the page to be read in, but have the
// page be deleted again before the waiting process had a chance
// to start running again after the wait.
//
// To resolve these points there is a while loop within LookUp() that
// continually checks the page until it is certain that either the
// page is resident or that it is not.
int BufMgr::LookUp(PID pid)
{

  register BFBucket* p;

  p = hashTable[Hash(pid)];
  while (p != NIL) {
    if ((p->pid.diskId != pid.diskId) ||
	(p->pid.pageId != pid.pageId))  {
      // not in this slot, skip it  
      p = p->next;
    } else {
      // there is a mapping for the page   
      if ( !(p->inTransit) ) {
	// no IO in progress, mapping is valid 
	return p->bufIndex;
      } else {
	// page is mapped but undergoing I/O, wait for it   
	if (p->ioComplete == NIL) {
            // allocate an event to wait on
            p->ioComplete = new global_event("ioComplete");
//	    numIoComplete++;
        }

	p->ioComplete->wait();
	// here I/O is complete, wake up and try loop again   
	// return to beginning of bucket 
	p = hashTable[Hash(pid)];
      }
    }
  }
  // there is no mapping for the page  
  return eNOTFOUND;
}


// reserve a mapping for an incoming page
void BufMgr::HashTableReserve(PID pid) {
  register BFBucket* p;

  p = hashTable[Hash(pid)];

  for ( ; p; p = p->next) {
    // Check if hashtable already contains what is to be inserted 
    if ((p->pid.diskId == pid.diskId) &&
	(p->pid.pageId == pid.pageId)) {
      printf("Node %d::HashTableReserve: tried to insert repeat pid %d.%d.\n",
	     myNode->id, pid.diskId, pid.pageId);
    } 
  }

  // fallen through to end of list, insert new entry at head of list 
  p = new BFBucket;

  // give it invalid index, since real index is unknown
  p->bufIndex = ILLEGAL_BUFINDEX;  

  p->pid = pid;

  // mark undergoing I/O
  p->inTransit = TRUE;

  // an event will be created by waiter if needed
  p->ioComplete = NIL;

  p->next = hashTable[Hash(pid)];
  hashTable[Hash(pid)] = p;
}


// install mapping from pid to buffer pool index.
// this routine should only be called on a page known
// not to be mapped or intransit.
void BufMgr::InsertInHashTable(int index)
{
  char* routine = "BufMgr::InsertInHashTable()";
  register BFBucket* 	p;
  PID pid;

  pid = bufPool[index].pid;
  p = hashTable[Hash(pid)];

  for ( ; p; p = p->next) {
    // Check if hashtable already contains what is to be inserted 
    if (p->bufIndex == index) {
      printf("Node %d::InsertInHashTable: pid %d.%d already allocated %d.\n",
	     myNode->id, pid.diskId, pid.pageId, index);
    } 
  }

  // fallen through to end of list, insert new entry at head of list 
  p = new BFBucket;
//  numBFBuckets++;

  p->bufIndex = index;
  p->pid = pid;
  p->ioComplete = NIL;
  p->inTransit = FALSE;
  p->next = hashTable[Hash(pid)];
  hashTable[Hash(pid)] = p;
}

// Invalidate mapping from pid to buffer pool index.
void BufMgr::RemoveFromHashTable(int index)
{
  register BFBucket*	p;
  register BFBucket**	pp;

  Page& page = bufPool[index];

  if (bufTable[index].fixCnt > 0) {
    printf("Node %d: BufMgr::RemoveFromHashTable removed fixed page %d.%d.\n",
	   myNode->id, page.pid.diskId, page.pid.pageId);
  }

  pp = &hashTable[Hash(page.pid)];
  p = *pp;

  for ( ; p; pp = &(p->next), p = p->next)  {
    if (index == p->bufIndex)  {
      if (p->inTransit) {
	printf("Node %d:BufMgr::RemoveFromHashTable: removed in-transit page!\n",
	       myNode->id);
      } else {
	*pp = p->next;
	if (p->ioComplete != NIL) {
	  delete p->ioComplete;
//	  numIoComplete--;
	  p->ioComplete = NIL;
	}
	delete p;
//	numBFBuckets--;
	return;
      }
    } 
  }
  printf("Node %d:BufMgr::RemoveFromHashTable: couldn't find index %d in frame %d.\n",
	 myNode->id,
	 index, 
	 Hash(page.pid));
}



// find a buffer for a new page to be read into
//
// Currently priorities are as follows: returns
//    1. An invalid page if one exists.
//    2. An unfixed and hated page that isn't intransit if one exists 
//       and no invalid page exists.  Priority within hated pages
//       is LRU.
//    3. An unfixed "loved" page that isn't intransit if one exists.
//       priority within "loved" pages is LRU.
int BufMgr::NewBuffer()
{
//  BFBucket *htEntry;
  register FreeListRecord *victim = NIL;
  register int i;
  int foundHatedPage = FALSE;
  int index;


  victim = FindInvalidFrame();
  if (victim != NIL) {
    // found an invalid page, return it!
    index = victim->bufIndex;
    bufTable[index].valid = TRUE;
    delete victim;
    return index;
  } 

  // at this point know there are no invalid frames available.

  victim = FindHatedFrame();
  if (victim != NIL) {
    // found a hated page, return it!
    index = victim->bufIndex;
    i = index;
    delete victim;
//    numFreeListRecs--;
    foundHatedPage = TRUE;
  } 

  // at this point we know there are no hated or invalid frames available.

  if (!foundHatedPage) {
    victim = FindLovedFrame();
    if (victim == NIL) {
      // bad news, no more available frames!
      printf("BufMgr node %d::NewBuffer(): no free buffers!\n", myNode->id);
      return eNOFREEBUFFER;
    } else {
      index = victim->bufIndex;
      i = index;
      delete victim;
    }
  }


  // here it was valid but not fixed, ref'ed,  or in transit.
  // (may or may not be hated)

  if (bufTable[i].dirty)  {
    // write it to disk before reusing frame
    WriteFrame(i);
  } 

  // invalidate mapping for the frame.
  RemoveFromHashTable(i);

  return i;
}




// install new page (diskId, pageId) in buffer pool of node,
// return index, also fix the page.  Error if page is already
// in the buffer pool.  Marks the page as "hated" if "hate"
// is TRUE; hate defaults to false.
int BufMgr::AllocPage(int diskId, int pageId, File *file, int hate)
{
  register int i;
  PID pid;
  FreeListRecord *f;

  pid.diskId = diskId;
  pid.pageId = pageId;

  // see if page already has been mapped.
  i = LookUp(pid);


  // Assert that the page does not exist in buffer

  if (i >= 0) {
    printf("BufMgr node %d::AllocPage: Tried to allocate existing page.\n", 
	   myNode->id);
    printf("page was disk = %d, pageId = %d\n", 
	   bufPool[i].pid.diskId, bufPool[i].pid.pageId);
    return(NOTOK);
  }


   // Get a new frame
  if ((i = NewBuffer()) < 0) return (NOTOK);
  else {
    bufTable[i].Clear();
    bufTable[i].fixCnt = 1;
    bufPool[i].pid = pid;
    bufPool[i].file = file;
    bufPool[i].op = (Op *) ILLEGAL_OP;
    bufTable[i].hated = hate;
    numFree--;
    InsertInHashTable(i);

    f = new FreeListRecord;
//    numFreeListRecs++;
    f->bufIndex = i;
    bufTable[i].flRec = f;

    if (hate == HATE) {
      // get a record and link this page into hated pages.
      InsertHatedPage(f);
    } else {
      // get a record and link this page into loved pages.
      InsertLovedPage(f);
    }
    return (OK);
  }
}

// Fake AllocPage --- does absolutely nothing.
// Included here for possible future use and
// to make all stores similar (AllocPage, WritePage).
// Used to model store writing out a bucket
// of a join using a pre-reserved join buffer frame.
int BufMgr::BucketAllocPage(int diskId, int pageId, File *file, int hate) 
{
  // do a bunch of nothing to keep CC from complaining about
  // parameters not being used.
  diskId = diskId;
  pageId = pageId;
  file = file;
  hate = hate;
  return (OK);
}

// Fake read --- just does a disk read, doesn't affect the
// buffer pool at all.  Used to model select reading bucket
// of a join into pre-reserved join buffer frame.
int BufMgr::BucketReadPage(int diskId, int pageId, int prefetch, Op *op)
{
    // read the page 
    myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->cnt_CPUInstr += Inst_StartIO;
    myNode->disks[diskId]->use(pageId, DISK_READ, prefetch, op);
    noIOReads++;

    return (OK);
}



// returns index of page with PID pid.  Causes I/O if
// page is not already in buffer pool; also fixes page.
// notes file if given.  Marks page as hated if "hate"
// is set to TRUE; hate defaults to "FALSE."
int BufMgr::ReadPage(int diskId, int pageId, File *file, int hate,
		     int prefetch, Op *op)
{
  register int i;
  BFBucket *htEntry;
  PID pid;
  FreeListRecord *f;

  pid.diskId = diskId;
  pid.pageId = pageId;
  bufferRefs++;
  if (op != NULL) op->cnt_bufRef++;  // counts refs in op, if present


  if ((i = LookUp(pid)) >= 0)  {

    // at this point we know page is mapped and resident. 
    if (bufTable[i].fixCnt++ == 0)  numFree--;

    bufferHits++;
    if (op != NULL) op->cnt_bufHit++; // counts hits in op, if present

    // if previous requestor hated the page, and
    // this requestor loves the page, change the page
    // to "love". 

    if ((hate == LOVE) && (bufTable[i].hated == HATE)) {
       f = bufTable[i].flRec;
       DeleteHatedPage(f);
       bufTable[i].hated = LOVE;
       InsertLovedPage(f);
    }
  } else {

    // must read the page in.  First reserve mapping to block
    // concurrent readers/writers
    HashTableReserve(pid);

    // now get a buffer
    if ((i = NewBuffer()) < 0) {  
      printf("BufMgr node %d::ReadPage: couldn't find free buffer.\n", 
	     myNode->id);
      return (NOTOK);
      // KLUDGE!!! should really "panic" here.
    }

    numFree--;

    // prepare frame for new page. 
    // set id's in buffer pool page  
    bufPool[i].pid = pid;
    bufPool[i].op = (Op *) ILLEGAL_OP;   // avoid coincidental release
    bufPool[i].file = file;

    // clear descriptor 
    bufTable[i].Clear();

    // fix the page 
    bufTable[i].fixCnt = 1;

    // install mapping from pid to index i. 
    htEntry = GetHashTableEntry(pid);
    if (htEntry == NIL) {
      printf("BufMgr node %d::ReadPage: couldn't find entry for pid %d.%d.\n",
              myNode->id, pid.diskId, pid.pageId);
    } 
    htEntry->bufIndex = i;
    htEntry->inTransit = TRUE;

    // mark love/hate status
    bufTable[i].hated = hate;

    f = new FreeListRecord;
    f->bufIndex = i;
    bufTable[i].flRec = f;
    if (hate == HATE) {
      // get a record and link this page into hated pages.
      InsertHatedPage(f);
    } else {
      // get a record and link this page into loved pages.
      InsertLovedPage(f);
    }

    // read the page 
    myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->cnt_CPUInstr += Inst_StartIO;
    myNode->disks[pid.diskId]->use(pid.pageId, DISK_READ, prefetch, op);
    noIOReads++;

    // wake up anyone waiting on this page (in lookup.) 
    htEntry->inTransit = FALSE;

    if (htEntry->ioComplete != NIL) {
       htEntry->ioComplete->set();
    }
  }
  return (OK);
}

// makes sure page is in buffer pool and fixed, marks  page
// as dirty, increments buffer refs and hits.
int  BufMgr::WritePage(int diskId, int pageId, Op *op)
{
  register int i;
  register PID  pid;

  pid.diskId = diskId;
  pid.pageId = pageId;

  if ((i = LookUp(pid)) < 0)  {
      printf("BufMgr node%d::WritePage:Non-resident page (disk=%d,pageId=%d)\n",
	         myNode->id, pid.diskId, pid.pageId);
      return(NOTOK);
  } else if (bufTable[i].fixCnt == 0)  {
    printf("BufMgr node %d::WritePage: Tried to write unfixed page.\n", 
	   myNode->id);  
    printf("page was disk = %d, pageId = %d\n", bufPool[i].pid.diskId, 
	   bufPool[i].pid.pageId);
    return(NOTOK);
  } else {
    bufTable[i].dirty = TRUE;
    if (op != NULL) op->cnt_bufDirty++;  // counts dirties in op, if present 
  }
  return (OK);
}


// Fake write --- just does a disk write, doesn't affect the
// buffer pool at all.  Used to model store writing out bucket
// of a join from a pre-reserved join buffer frame.
int BufMgr::BucketWritePage(int diskId, int pageId, Op *op) 
{
    // write the page 
    myNode->cpu.use(((float) Inst_StartIO) / (Mips * 1000)); //milliseconds
    op->cnt_CPUInstr += Inst_StartIO;
    myNode->disks[diskId]->use(pageId, DISK_WRITE, NO_PREFETCH);
    noIOWrites++;
    if (op != NULL) op->cnt_bufDirty++;  // counts dirties in op, if present 
    return (OK);

}

// delete page lid from the buffer pool.
// assumes page was not fixed or dirty.
int BufMgr::DiscardPage(int bufIndex)
{
  register FreeListRecord *f;

  if ((bufIndex < 0) || (bufIndex >= numBuf)) {
      printf("BufMgr node %d::DiscardPage: out of range buffer number.\n", 
	myNode->id);
      return(NOTOK);
  }

  if (bufTable[bufIndex].fixCnt != 0) {
      printf("BufMgr node %d::DiscardPage: page was fixed!.\n", myNode->id);
      return(NOTOK);
  } else if (bufTable[bufIndex].dirty) {
      printf("BufMgr node %d::DiscardPage: page was dirty!.\n", myNode->id);
      return(NOTOK);
  } else if (!bufTable[bufIndex].valid) {
      printf("BufMgr node %d::DiscardPage: page was invalid!.\n", myNode->id);
      return(NOTOK);
  }
  numFree++;

  f = bufTable[bufIndex].flRec;
  if (bufTable[bufIndex].hated == HATE) {
    DeleteHatedPage(f);
  } else {
    DeleteLovedPage(f);
  }

  bufTable[bufIndex].valid = FALSE;
  // link page into
  // list of invalid records.
  f->bufIndex = bufIndex;
  InsertInvalidPage(f);

  RemoveFromHashTable(bufIndex);
  return (OK);
}


//Reset the buffer pool to initial condition
//  i.e. put everything in invalid pages list
int BufMgr::ResetBuffer()
{
  int e;
  register FreeListRecord *f;

  // put HatedPages in invalid page list
  while ((f = HatedPages) != NULL)
    if ((e = DiscardPage(f->bufIndex)) != OK) return(NOTOK);

  // put LovedPages in invalid page list
  while ((f = LovedPages) != NULL)
    if ((e = DiscardPage(f->bufIndex)) != OK) return(NOTOK);

  return(OK);
}


// Decrement fix count for page in slot "bufIndex".
int BufMgr::UnfixPage(int diskId, int pageId)
{

  register int bufIndex;
  register FreeListRecord *f;
  PID pid;

  pid.diskId = diskId;
  pid.pageId = pageId;
  bufIndex = LookUp(pid);

  if (bufTable[bufIndex].valid == FALSE) {
    printf("UnfixPage: Asked to unfix an invalid page!\n");
    return (NOTOK);
  } else if ((bufIndex < 0) || (bufIndex >= numBuf)) {
      printf("BufMgr node %d::UnfixPage: couldn't find page with index %d.\n",
	   myNode->id, bufIndex);
      printf("Was looking for PID (disk = %d, pageId = %d)\n",
	   pid.diskId, pid.pageId);
      fflush(stdout);
      return (NOTOK);
  } else if (bufTable[bufIndex].fixCnt == 0) {
      printf("BufMgr node %d::UnfixPage: page %d was not fixed!\n", 
	     myNode->id,
	     bufIndex);
      printf("\t pid was %d.%d\n", 
	     bufPool[bufIndex].pid.diskId,
	     bufPool[bufIndex].pid.pageId);
      fflush(stdout);
      return (NOTOK);
  } else if (--bufTable[bufIndex].fixCnt == 0) {
      numFree++;
     // bump it up in LRU chain (either hated or loved.) 

      f = bufTable[bufIndex].flRec;

      if (f->bufIndex != bufIndex) {
	printf("Unfix page: free list record and Lookup() mismatch on idx!\n");
      } else if (bufTable[bufIndex].hated) {
	DeleteHatedPage(f);
	InsertHatedPage(f);
      } else {
	DeleteLovedPage(f);
	InsertLovedPage(f);
      }
      return (OK);
  } else return (OK);
}

// Release all pages reserved by op.
int BufMgr::ReleaseBufferPages(Op *op)
{
  // do sequential scan of buffer pool.  If this
  // routine becomes a bottleneck, should be
  // replaced by routine that uses auxiliary data
  // structure mapping op's to pages.

  register int i;
  register FreeListRecord *f;

//  int relCnt = 0;

//  printf("Node %d: Op %d releasing pages.\n",
//          myNode->id, (int) op);

  for (i = 0; i < numBuf; i++) {
    if ((bufPool[i].op == op) && (bufTable[i].reserved == TRUE)) {  // got one!
  //     relCnt++;
       bufTable[i].reserved = FALSE;
       bufTable[i].valid = FALSE;  // good candidate to replace
       bufPool[i].op = (Op *) ILLEGAL_OP;

       // get a free list record for new page and link it into
       // list of invalid record.
       f = new FreeListRecord;
       f->bufIndex = i;
       InsertInvalidPage(f);
       numFree++;
    }
  }
//  printf("Node %d: Op %d released %d pages.\n",
//          myNode->id, (int) op, relCnt);
  return (OK);
}


// Release numReleasePages of the pages reserved by op.
int BufMgr::ReleaseReservedPages(Op *op, int numReleasePages)
{
  // do sequential scan of buffer pool.  If this
  // routine becomes a bottleneck, should be
  // replaced by routine that uses auxiliary data
  // structure mapping op's to pages.

  register int i;
  register FreeListRecord *f;
  int relCnt = 0;

//  printf("Node %d: Op %d releasing pages.\n",
//          myNode->id, (int) op);

  for (i = 0; (i < numBuf) && (relCnt < numReleasePages); i++) {
    if ((bufPool[i].op == op) && (bufTable[i].reserved == TRUE)) {  // got one!
       relCnt++;
       bufTable[i].reserved = FALSE;
       bufTable[i].valid = FALSE;  // good candidate to replace
       bufPool[i].op = (Op *) ILLEGAL_OP;

       // get a free list record for new page and link it into
       // list of invalid record.
       f = new FreeListRecord;
       f->bufIndex = i;
       InsertInvalidPage(f);
       numFree++;
    }
  }
//  printf("Node %d: Op %d released %d pages.\n",
//          myNode->id, (int) op, relCnt);

  // if less reserved pages found flag an error
  if (relCnt != numReleasePages)
  {
      printf("node %d::ReleaseReservedPages: wanted %d found %d",
             myNode->id, numReleasePages, relCnt);
      return(NOTOK);
  }
  return (OK);
}


// Reserve numPages for op.  Returns number
// of pages actually reserved.  Actual reservation is
// made by "fix"ing the pages in the name of the op.
// File is set to ILLEGAL_FILE to avoid flushing, 
// diskId set to ILLEGAL_FILE to avoid false "hits."
// Returns number of pages actually reserved.
//
// Note: if the option-taking is working properly,
// it should never fail to reserve the requested
// number of pages.
int BufMgr::ReserveBufferPages(int numPages, Op *op)
{
  register int i;
  register int curPage;
  int actualNumberReserved;

//  printf("Node %d: Op %d reserving %d pages.\n",
//	 myNode->id, (int) op, numPages);


  if (numPages > numFree) {
    actualNumberReserved = numFree;
    numFree = 0;
    printf("BARF! Carey said this would never happen.\n");
  } else {
    actualNumberReserved = numPages;
    numFree -= numPages;
  }

  // now actually reserve the pages.
  for (i = 0; i < actualNumberReserved; i++) {
    curPage = NewBuffer();
    if (curPage < 0) {
      // if we get here, we have failed to grab enough pages.
      printf("Barf, Naughton said this wouldn't happen.\n");
      printf("(Couldn't get promised number of pages.)\n");
      ReleaseBufferPages(op);
      return 0;
    } else {
      bufTable[curPage].reserved = TRUE;
      bufPool[curPage].op = op;
      bufPool[curPage].file = (File *) ILLEGAL_FILE; // prevent coinc flushings.
      bufPool[curPage].pid.diskId = ILLEGAL_DISK;  // prev othr ops frm finding
    }
  }
  return actualNumberReserved;
}


// Destructively flush a frame.  That is, after
// the flush, the frame is free and invalid.  Does
// a disk write if necessary.
int BufMgr::FlushFrame(int frame) {

  BFBucket *htEntry;
  FreeListRecord *f;

  if(bufTable[frame].fixCnt != 0) {
    printf("node %d::FlushFRAME: tried to flush fixed frame %d!\n", 
	   myNode->id, frame);
    return (NOTOK);
  } else if (!bufTable[frame].valid) {
    printf("node %d::FlushFRAME: tried to flush invalid frame %d!\n", 
	   myNode->id, frame);
    return (NOTOK);
  } else if (!bufTable[frame].dirty) {
    // clean, just delete hash table mapping and set to 
    // invalid. 
    RemoveFromHashTable(frame);


    // delete it from hated or loved list.
    f = bufTable[frame].flRec;

    if (f->bufIndex != frame) {
      printf("FlushFrame(): f->bufIndex = %d, frame = %d.\n",
	     f->bufIndex, frame);
    } else if (bufTable[frame].hated == HATE) {
      DeleteHatedPage(f);
    } else { // must be loved 
      DeleteLovedPage(f);
    }
    // link it into list of invalid records.
    bufTable[frame].valid = FALSE;
    InsertInvalidPage(f);

  } else if (bufTable[frame].dirty) {
    htEntry = GetHashTableEntry(bufPool[frame].pid);
    if (htEntry->inTransit) {
      // page is being written.  Only legal way this
      // could happen is if NewBuffer() is trying to
      // steal the frame.  In this case, don't do anything,
      // since NewBuffer() is doing the flush.
      ;
    } else {
      // frame is dirty but not in transit, write it out. 
      WriteFrame(frame);
      // delete hash table mapping. 
      RemoveFromHashTable(frame);

      f = bufTable[frame].flRec;
      if (f->bufIndex != frame) {
         printf("FlushFrame():2: f->bufIndex = %d, frame = %d.\n",
	     f->bufIndex, frame);
      } else if (bufTable[frame].hated == HATE) {
        DeleteHatedPage(f);
      } else { // must be loved 
        DeleteLovedPage(f);
      }

      // invalidate 
      bufTable[frame].valid = FALSE;

      // link into list of invalid records.
      InsertInvalidPage(f);
    }
  }
  return (OK);
}


// flushes all pages with homes in "file".  ASSUMPTION:
// FlushFile() should only be called when caller is
// certain that there is no other activity on the file,
// since this is a "destructive flush" that results in
// the buffer pool being void of any pages from "file."
int BufMgr::FlushFile(File *file)
{
    register int i;
    int status;

    // sequential scan of buffer pool.  Might need to be replaced
    // if this becomes a bottleneck.
    for (i = 0; i < numBuf; i++)  {
      if ((bufTable[i].valid) && (bufPool[i].file == file)) {
        // got a candidate to flush
	if (bufTable[i].reserved) {
	  printf("Hey, flushFile is flushing a reserved page!\n");
	}
	status = FlushFrame(i);
	if (status != OK) {
	  return (NOTOK);
	}
      }
    }
    return (OK);
}



int BufMgr::FlushDisk(int diskId)
{
    register int i;
    int flushCount = 0;
    int status;
  
    for (i = 0; i < numBuf; i++)  {
      if ((bufTable[i].valid) && (bufPool[i].pid.diskId == diskId)) {
	status = FlushFrame(i);
	if (status != OK) {
	  return (NOTOK);
	}
      }
    }
    return (OK);
}

int BufMgr::FlushPage(int diskId, int pageId)
{
    PID pid;
  
    pid.diskId = diskId; 
    pid.pageId = pageId;

    int bufIndex = LookUp(pid);

    if ((bufIndex < 0) || (bufIndex >= numBuf)) {
        printf("BufMgr node %d::FlushPage: couldn't find page.\n", myNode->id);
	return (NOTOK);
    } else if (!bufTable[bufIndex].valid) {
    	printf("BufMgr node %d::FlushPage: page was not valid!\n", myNode->id);
	return (NOTOK);
    } else return FlushFrame(bufIndex);
}

// Print stats from buffer pool.
void BufMgr::PrintBufferPoolStats() 
{
    printf("  %4d  %5.3f  %8d  %8d  %8d  %8d\n", 
	   myNode->id, 
	   (bufferRefs == 0) ? 0 : (float) bufferHits/(float) bufferRefs, 
	   bufferHits, 
	   bufferRefs, 
	   noIOWrites, 
	   noIOReads);
}

// Returns nonzero if page is dirty.
int BufMgr::IsDirty(int bufIndex)
{
    return bufTable[bufIndex].dirty;
}

void BufMgr::PrintBufferPool() 
{
    for (int i = 0; i < numBuf; i++) 
    {
    	printf("Buffer pool for node %d, page %d:\n", myNode->id, i);
    	printf("\t valid = %d.\n", bufTable[i].valid);
    	printf("\t dirty = %d.\n", bufTable[i].dirty);
    	printf("\t reserved = %d.\n", bufTable[i].reserved);
    	printf("\t fixcnt = %d.\n", bufTable[i].fixCnt);
    	printf("\t hated = %d.\n", bufTable[i].hated);
    	printf("\t PID: %d.%d\n", bufPool[i].pid.diskId,bufPool[i].pid.pageId);
    }
}

void BufMgr::PrintHashTable() {

  BFBucket* p;

  printf("Printing Hash Table for node %d.\n", myNode->id);
  for (int i = 0; i < hashSize; i++) {
    printf("\t ht[ %d ]: ", i);
    p = hashTable[ i ];
    while (p != 0) {
      if (p->bufIndex >= 0) {
	printf("(index %d, pid %d.%d, intransit %d, address %x) ", 
	       p->bufIndex, 
	       p->pid.diskId,
	       p->pid.pageId,
	       p->inTransit,
	       (int) p);
      }
      p = p->next;
    }
    printf("\n");
  }
}

void BufMgr::PrintHatedList() {

  FreeListRecord *f;
  int count = 0;  // for sanity bound!!

  f = HatedPages; 

  printf("Node %d: Printing Hated Page List.\n", myNode->id);

  if (f == NIL) {
    printf("\tempty list.\n");
  } else {
    printf("\t index = %d.\n", f->bufIndex);
    printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
    printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
    printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
    printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
    printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
    printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
    printf("\n");
    f = f->next;
    while ((f != HatedPagesTail) && (count < numBuf)) {
      printf("\t index = %d.\n", f->bufIndex);
      printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
      printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
      printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
      printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
      printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
      printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
      printf("\n");
      f = f->next;
      count++;
    }
  }
}

void BufMgr::PrintLovedList() {

  FreeListRecord *f;
  int count = 0;  // for sanity bound

  f = LovedPages; 

  printf("Node %d: Printing Loved Page List.\n", myNode->id);

  if (f == NIL) {
    printf("\tempty list.\n");
  } else {
    printf("\t index = %d.\n", f->bufIndex);
    printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
    printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
    printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
    printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
    printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
    printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
    printf("\n");
    f = f->next;
    while ((f != LovedPagesTail) && (count < numBuf)) {
      printf("\t index = %d.\n", f->bufIndex);
      printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
      printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
      printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
      printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
      printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
      printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
      printf("\n");
      f = f->next;
      count++;
    }
  }
}

void BufMgr::PrintInvalidList() {

  FreeListRecord *f;
  int count = 0;  // for sanity bound

  f = InvalidPages; 

  printf("Node %d: Printing Invalid Page List.\n", myNode->id);

  if (f == NIL) {
    printf("\tempty list.\n");
  } else {
    printf("\t index = %d.\n", f->bufIndex);
    printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
    printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
    printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
    printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
    printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
    printf("\t op = %d.\n", bufPool[f->bufIndex].op);
    printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
    printf("\n");
    f = f->next;
    while ((f != InvalidPagesTail) && (count < numBuf)) {
      printf("\t index = %d.\n", f->bufIndex);
      printf("\t valid = %d.\n", bufTable[f->bufIndex].valid);
      printf("\t dirty = %d.\n", bufTable[f->bufIndex].dirty);
      printf("\t reserved = %d.\n", bufTable[f->bufIndex].reserved);
      printf("\t fixcnt = %d.\n", bufTable[f->bufIndex].fixCnt);
      printf("\t hated = %d.\n", bufTable[f->bufIndex].hated);
      printf("\t op = %d.\n", bufPool[f->bufIndex].op);
      printf("\t PID: %d.%d\n", bufPool[f->bufIndex].pid.diskId,bufPool[f->bufIndex].pid.pageId);
      printf("\n");
      f = f->next;
      count++;
    }
  }
}




void BufMgr::PrintEverything() {
  PrintHatedList();
  PrintLovedList();
  PrintInvalidList();
  PrintBufferPool();
  PrintHashTable();
  fflush(stdout);
}
