#define BUFMGR_H


#define eNOERROR 0
#define eNOTFOUND (-1)
#define eNOFREEBUFFER (-2)
#define ILLEGAL_DISK (-1)
#define ILLEGAL_FILE (-1)
#define ILLEGAL_PAGE (-1)
#define ILLEGAL_OP (-1)
#define ILLEGAL_BUFINDEX (-1)

#define HATE 1
#define LOVE 0

extern class Node;
extern class Op;
extern class File;
extern class FreeListRecord;

class BFBucket;


// a physical page id
typedef struct {
    int pageId;
    int diskId;
} PID;


// a page in the buffer pool.
typedef struct {
  struct PID pid;          // logical id of page
  Op   *op;                // op "owning" the page
  File *file;              // file for the page
} Page;



// descriptor for page in buffer pool
struct  BufDesc {
  FreeListRecord *flRec;        // pointer to record in hated or loved use chain 
  short		fixCnt;		// fix charges
  unsigned	dirty : 1;		// TRUE if dirty
  unsigned	valid : 1;		// TRUE if page is valid
  unsigned      reserved : 1;       // TRUE if page has been reserved
  unsigned      hated : 1;          // TRUE if page is "hated"

  BufDesc() {
    fixCnt = 0;
    dirty = valid = reserved = hated = FALSE;
  }

  void Clear()	{               // initialize a page descriptor for a new user
    fixCnt = 0;
    dirty = FALSE;
    valid = TRUE;
    reserved = FALSE;
  }
};

class Op;

class BufMgr {
  int		numFree;          // # remaining free buffers
  int           hateCount;        // # hated pages in buffer pool
  int		numBuf;           // total # buffers

  FreeListRecord *HatedPages;     // list of currently hated pages.
  FreeListRecord *HatedPagesTail;     // Tail of list of currently hated pages.
  FreeListRecord *InvalidPages;   // list of currently invalid pages.
  FreeListRecord *InvalidPagesTail;   // Tail list of currently invalid pages.
  FreeListRecord *LovedPages;     // list of currently invalid pages.
  FreeListRecord *LovedPagesTail;     // Tail of list of currently invalid pages.

  // various clock hands --- should go away when we use
  // linked lists.
  int         clockHand, hatedClockHand;

  // variables for maintaining statistics
  int 		bufferHits;
  int		bufferRefs;
  int           noIOReads;
  int           noIOWrites;

  Node  *myNode;

  Page*		bufPool;          // the pool itself
  BufDesc*	bufTable;         // descriptors for pool
                                  // (assumption: descriptor
                                  // for bufPool[i] is in bufTable[i].)

  BFBucket** 	hashTable;        // maps LIDS to buffer pool pages
  int		hashSize;
  int	Hash(PID pid) 	
    { return (pid.pageId * (pid.diskId+1)) % hashSize; }

  int	LookUp(PID pid);  // return index of page PID if found


  // insert f in list of invalid pages.
  void BufMgr::InsertInvalidPage(FreeListRecord *f);

  // delete f from list of invalid pages.
  void BufMgr::DeleteInvalidPage(FreeListRecord *f);

  // find an invalid frame in the invalid list.  Return
  // nil if none is found, otherwise pointer to the
  // record for the invalid frame.  As a side effect
  // deletes the frame from the invalid list.
  FreeListRecord *BufMgr::FindInvalidFrame();

  // insert f in list of hated pages.
  void BufMgr::InsertHatedPage(FreeListRecord *f);

  //  delete f from list of hated pages.
  void BufMgr::DeleteHatedPage(FreeListRecord *f);
 
  // return a pointer to an Hated frame that is not
  // be intransit or fixed; NIL if no such frame
  // exists.
  FreeListRecord *BufMgr::FindHatedFrame();

  // return TRUE if all frames on hated list are hated;
  // FALSE otherwise.
  int BufMgr::CheckHatedFrameList();


  // insert f in list of Loved pages.
  void BufMgr::InsertLovedPage(FreeListRecord *f);

  //  delete f from list of hated pages.
  void BufMgr::DeleteLovedPage(FreeListRecord *f);
 
  // return a pointer to an Hated frame that is not
  // be intransit or fixed; NIL if no such frame
  // exists.
  FreeListRecord *BufMgr::FindLovedFrame();



  // Write frame "index" to disk.  Sets "inTransit" to
  // TRUE before IO, FALSE afterwards.  Also, wakes up
  // any waiters on ioComplete when the io is finished.
  void BufMgr::WriteFrame(int index);

  // Destructively flush a frame.  That is, after
  // the flush, the frame is free and invalid.  Does
  // a disk write if necessary.
  int BufMgr::FlushFrame(int frame);

  // initialize descriptor for page in bufPool[i]. 
  void	InsertInHashTable(int index); 
                                      

  // delete descriptor for page in bufPool[i].
  void	RemoveFromHashTable(int index); 
                                        

  // reserve mapping for pid.  This has nothing to do with
  // reserving buffer pool frames!!!
  void  HashTableReserve(PID pid);  
    
  // return pointer to node in HT
  BFBucket* GetHashTableEntry(PID pid);  


  void	ClearStat()	   // initialize stats.
           { bufferHits = bufferRefs = noIOReads = noIOWrites = 0; }

  // return index of a free buffer
    int		NewBuffer(); 


 public:
  BufMgr(int numBuffers, Node *owningNode = 0);
  ~BufMgr();

  // install page lid in buffer pool of node,
  // return error.  Record file if given; this
  // allows the pages of a file to be flushed as
  // a group.  Marks the page as hated if "hate" is
  // TRUE; hate defaults to FALSE.
  int 	AllocPage(int diskId, int pageId, File *file = 0, int hate = FALSE);

  // returns index of page (diskId, pageId) .  (causes I/O if
  // page is not already in buffer pool; also fixes page.)
  // file arg allows pages of a file to be flushed as a group.  
  // Marks page as hated if "hate" is TRUE.
  // op should be specified if buffer and disk stats are to be recorded.
  int   ReadPage(int diskId, int pageId, File *file, int hate, int prefetch,
		 Op* op = NULL);

  // Fake read --- just does a disk read, doesn't affect the
  // buffer pool at all.  Used to model select reading bucket
  // of a join into pre-reserved join buffer frame.
  // op should be specified if buffer and disk stats are to be recorded.
  int BufMgr::BucketReadPage(int diskId, int pageId, int prefetch,
  			     Op* op = NULL);


  // makes sure page is in buffer pool and fixed, marks  page
  // as dirty, increments buffer refs and hits.  File is
  // not needed here, since it is set by preceding AllocPage();
  // return error.  
  // op should be specified if buffer and disk stats are to be recorded.
  int   WritePage(int diskId, int pageId, Op* op = NULL);

  // Fake AllocPage --- does absolutely nothing.
  // Used to model store writing out a bucket
  // of a join using a pre-reserved join buffer frame.
  int BufMgr::BucketAllocPage(int diskId, 
			      int pageId, 
			      File *file = 0, 
			      int hate = FALSE);

  // Fake write --- just does a disk write, doesn't affect the
  // buffer pool at all.  Used to model store writing out bucket
  // of a join from a pre-reserved join buffer frame.
  // op should be specified if buffer and disk stats are to be recorded.
  int BufMgr::BucketWritePage(int diskId, int pageId, Op* op = NULL);

  // delete page lid from the buffer pool.
  // assumes page was not fixed or dity.
  int 	DiscardPage(int bufIndex);

  // reset the buffer pool to the initial condition
  int   ResetBuffer();

  // Decrement fix count for page (diskId, pageId).  
  // returns error.
  int  UnfixPage(int diskId, int pageId);

  // Returns nonzero if page is dirty.
  int	IsDirty(int bufIndex);

  // Try to reserve numPages for op.  Returns number
  // of pages actually reserved.  Actual reservation is
  // made by "fix"ing the pages in the name of the op.  
  int  ReserveBufferPages(int numPages, Op *op);

  // Release all pages reserved by op.
  int  ReleaseBufferPages(Op *op);

  // Release numReleasePages reserved by op.
  int  ReleaseReservedPages(Op *op, int numReleasePages);

  // Flush all pages for a given file to disk.  This
  // is a destructive flush --- after flushing the
  // page frames are available for another page.
  int  FlushFile(File *file);

  // flush all dirty pages from disk diskId
  int	FlushDisk(int diskId);

  // flush page (diskId, pageId);
  int BufMgr::FlushPage(int diskId, int pageId);

  // Print stats from buffer pool.
  void BufMgr::PrintBufferPoolStats(); 

  // The following Print routines are useful in debugging.
  // Warning: in large configurations, each routine will
  // print a lot of information...
  void BufMgr::PrintBufferPool();

  void BufMgr::PrintHashTable();

  void BufMgr::PrintHatedList();

  void BufMgr::PrintLovedList();

  void BufMgr::PrintInvalidList();

  void BufMgr::PrintEverything();
};

