#include <dbsim.h>
#include <object.h>
#include <node.h>
#include <language.h>
#include <db.h>
#include <strings.h>
#include <math.h>

//---------------------------------------------------------------------------
// class implementation for Files
//---------------------------------------------------------------------------

File::File(int fileId, int nodeCount, Node** nodes, int diskCount, int* disks,
           int tupCount, int tupSize)
{
    // variables used to compute sizes and ids of file fragments, etc.
    int tupsPerFragment, extraTups, fragTups, fragId, n, d;
    Disk *curDisk;

    // record the file id for display/debugging purposes
    fileNo = fileId;

    // record the set of storage sites for the file
    numNodes = nodeCount;
    for (n = 0; n < numNodes; n++) {
	fragNode[n] = nodes[n];
    }

    // record the particular disks where the file is to reside
    numDisks = diskCount;
    for (d = 0; d < numDisks; d++) {
	fragDisk[d] = disks[d];
    }

    // record the file-specific database characteristics
    tupleSize = tupSize;
    numTuples = tupCount;
    tuplesPerPage = PageSize / tupleSize;

    // allocate a set of file fragments for representing the file, with
    // one fragment residing on each storage disk of each file storage node,
    // using a sparse data structure that will mesh well with the related
    // data structures that David is using for identifing active sites
    numFragments = numNodes * numDisks;
    tupsPerFragment = numTuples / numFragments;
    extraTups = numTuples % numFragments;
    // begin by clearing the fragment pointer array
    for (n = 0; n < MAX_NODES; n++) {
        for (d = 0; d < MAX_DISKSPERNODE; d++) {
	    fragment[n][d] = NULL;
	}
    }
    // now initialize the actual fragment pointers
    fragId = 0;
    for (n = 0; n < numNodes; n++) {
        for (d = 0; d < numDisks; d++) {
	    curDisk = storageNode(n)->disks[storageDisk(d)];
	    fragTups = tupsPerFragment;		// all sites get this many
	    if (extraTups-- > 0) fragTups++;	// some sites get one extra
	    fragment[storageNode(n)->id][storageDisk(d)] =
	        new FileFrag(fragId, curDisk, tuplesPerPage, fragTups);
	    fragId++;
        }
    }
}


File::~File()
{
    if (numFragments > 0) {
        for (int n = 0; n < numNodes; n++) {
            for (int d = 0; d < numDisks; d++) {
	        delete fragment[storageNode(n)->id][storageDisk(d)];
            }
        }
    }
}


void File::deleteFrag(Node* node, int disk)
{
//  printf("File::deleteFrag(node %d, disk %d)", node->id, disk);
    delete fragment[node->id][disk];
    numFragments--;
    if (numFragments == 0) {
	delete this;
//      printf(" -- deleted file itself");
    }
//  printf(" -- %d fragments left\n", numFragments);
};


void File::print()
{
    int n, d;

    printf("File summary for file %d...\n", fileNo);
    printf("\tnumNodes = %d, numDisks = %d\n", numNodes, numDisks);
    printf("\tnodes = { ");
    for (n = 0; n < numNodes; n++) {
	printf("%d ", storageNode(n)->id);
    }
    printf("}\n");
    printf("\tdisks = { ");
    for (d = 0; d < numDisks; d++) {
	printf("%d ", storageDisk(d));
    }
    printf("}\n");
    printf("\ttuplesPerPage = %d, numTuples = %d, tupleSize = %d\n",
            tuplesPerPage, numTuples, tupleSize);
    printf("\tFile contains %d fragments...\n", numFragments);
    for (n = 0; n < numNodes; n++) {
        for (d = 0; d < numDisks; d++) {
  	    printf("\t  Displaying file fragment[%d, %d]\n",
  	            storageNode(n)->id, storageDisk(d));
  	    fragment[storageNode(n)->id][storageDisk(d)]->print();
        }
    }
}

