%{
#define isparser "yup"
#include <stdio.h>
#include <math.h>
#include "string.h"
#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "query.h"
#include "language.h"
#include "disk.h"
#include "db.h"
#include "cost.h"

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       parser.y (yyparse, yyerror, yywhere, yywrap)
//  
//    Function:   A lex/yacc generated parser which translates hardware and 
//                software configuration, database definition, and workload
//                definition statements into "template objects" (see 
//                language.h).  These template objects are then used to cReate 
//                specific instances of simulator objects:  nodes, disks, 
//                terminals, and queries.
//
//    Change Activity:
//
//     09/91 - Original version.  KPB
//  11/01/91 - Added index support.  KPB
//
///////////////////////////////////////////////////////////////////////////////

extern SymTable   RealSyms;      // real num symbol table
extern SymTable   FileSyms;      // file symbol table
extern SymTable   PlanSyms;      // plan symbol table
extern SymTable   JobClassSyms;  // job class symbol table
extern Node       *Nodes[MAX_NODES];
extern Index      *Indices[MAX_FILES];
extern File       *Relations[MAX_FILES];
extern JobClass   *JobClasses[MAX_CLASSES];

class IntegerSet;

ProbDist   *currProbDist; // working ptr to a user specified prob distr array
RandomReal *currRealNum;  // working ptr to a random real number 
RandomFile *currFile;     // working ptr to a random file 
RandomPlan *currPlan;     // working ptr to a random plan 
JobClass   *currJClass;   // working ptr to a job class 
IntegerSet *currIntSet;   // working ptr to a set of integers
MemReq     *currMemReq;   // working ptr to a memory requirement
int        i;
bool       stmtErr = FALSE; // some syntax/semantic error on current stmt

static char errmsg[80];
extern char yytext[];     // lex managed, holds current input file line
extern int yyleng;
extern int yylineno;
FILE*      yyerfp = stdout;
extern char *SourceFileName;

void yyerror(char* s);
void       yywhere();
int        yylook();
int        yyback(int*, int);
int        yywrap();
void       CreateFile(char* sym, RandomFile* f, int fileID);
void       ResetFile();
void       CreateJobClass(char* sym, JobClass* jc);
void       CreatePlan(char* sym, RandomPlan* p, bool stats);

struct fileinfo {         // place to accumulate file attributes during parse
  fileType type;          // index_file or data_file
  int      fileNo;
  int      numNodes;
  Node**   storageNode;   // list of specific nodes to decluster over
  int      numDisks;
  int*     storageDisk;   // list of disks within each node to place fragments
  int      tupleSize;
  int      numTuples;
  int      keySize;
  int      indexType;
  RandomFile* indexedFile; 
} currFileInfo = {data_file, -1, -1, NULL, -1, NULL, -1, -1, -1, -1, NULL};

typedef enum intset_type {random_intset, constant_intset};

class IntegerSet {          // random or fixed set of integers
  public:			    
    intset_type setType;    // random or constant			      
    int      max;           // max cardinality and element value allowed
    int      card;          // current cardinality (number of elements)
    int*     set;           // actual set (array) of integers
    RandomSet* randSet;     // points to RandomSet if specified that way

    IntegerSet(int mx) {

      // Prepare for default of constant_intset

      setType = constant_intset;
      max = mx;  
      card = 0;  
      set = new int[max];
      randSet = NULL;
    };

    ~IntegerSet() {delete set;};

    int add(int e) {
      if ((int) e < max && card < max) {
	set[card++] = (int) e;
	return(0);          // indicate successful
      }
      else {
	sprintf(errmsg, "max cardinality is %d and max element value is %d for this set", max, max-1);
	yyerror(errmsg);
	return(-1);         // indicate some problem
      }
    };
};

%}

/* Union of all the (compatible) types that can live on the semantic stack */

%union {               
        int intval;
	float floatval;
	bool boolval;
	char *string;
	joinTypes jtype;
	scanTypes stype;
        ProbDist* probdist;
        RandomReal* randreal;
        RandomFile* randfile;
	RandomPlan* randplan;
        JobClass* jclass;
	IntegerSet* iset;
	MemReq* mreq;
        }

/* Mappings of object types to non-terminals which detect those object types */

%type<probdist> ProbDist
%type<randreal> RealNum
%type<randfile> File
%type<randplan> Plan PlanList Query QueryOp
%type<jclass> JobClass
%type<jtype> JoinType
%type<stype> ScanType
%type<iset> Sites
%type<mreq> MemReq
%type<intval> MemUnit
%type<boolval> PlanAttrList

/* Tokens recognized by the lexical analyzer (tokenizer.l) */

%token <string> FILENAME
%token <string> NAME
%token <floatval> NUMBER
%token SEMI COMMA EQUALS SLASH DOTDOT SPLAT
%token LPAREN RPAREN LBRACK RBRACK LBRACE RBRACE LESSTHAN GREATERTHAN
%token REAL PLAN FILEKW INDEXKW JOBCLASS
%token PROJ SORT AGGR XPROD UPDATE WHERE INDEXSCAN FILESCAN STOREOP
%token HASHJOIN INDEXJOIN NLJOIN MERGEJOIN UNIQUE SORTUNIQUE
%token UNION INTERSECTION DIFFERENCE TERMSEND
%token EXPNTL NORMAL UNIFORM RANDOM
%token ID TUPLESIZE NUMTUPLES DISKS NODES KEYSIZE TYPE CLUSTEREDKW 
%token UNCLUSTEREDKW THINKSECS POPULATION PLANS CPUS CPUMIPS KBYTES PAGESIZE
%token PAGESPERCYL SEEKFACTOR MINLATENCY ROTATION SETTLE XFERRATE PRIMARY
%token CACHEPAGES BPOOL MIN MAX BYTES KBYTES MBYTES MAXMPL STATS CACHECONTEXTS
%token STARTIO_INST

/* Goal symbol of the language */

%start SimDef

/* Productions and their associated semantics (actions) within the braces */

%%
 SimDef:         Stmt
             |   SimDef Stmt { yyerrok; stmtErr = FALSE; }
             ;

 Stmt:           RealNumStmt
             |   FileStmt
             |   JobClassStmt
             |   PlanStmt
             |   NodeStmt
             |   DiskStmt

             |   error Semi { yyerror("unknown statement type"); }
             ;

 ProbDist:       LBRACK { currProbDist = new ProbDist(rand_dist); } 
                 ProbList RBRACK {
                   if (!currProbDist->spread()) // if problem spreading prob
		     delete currProbDist;    // dispose of defective list
		   else 
		     $$ = currProbDist;      // push the list on semantic stack
		 }

             |   LESSTHAN { currProbDist = new ProbDist(determ_dist); } 
                 ProbList GREATERTHAN {
	           $$ = currProbDist;        // push the list on semantic stack
		 }
             ;

 ProbList:       ProbListEl
             |   ProbList COMMA ProbListEl
             ;

 ProbListEl:     NAME SLASH NUMBER 
                   { currProbDist->add($1, $3); }

             |   NUMBER SLASH NUMBER 
                   { currProbDist->add($1, $3); }

             |   NAME SPLAT NUMBER 
                   { currProbDist->add($1, (int) $3); }

             |   NUMBER SPLAT NUMBER 
                   { currProbDist->add($1, (int) $3); }

             |   NAME 
                   { currProbDist->add($1); }

             |   NUMBER
                   { currProbDist->add($1); }
             ;

 RealNumStmt:    REAL RealNumList Semi
             ;

 RealNumList:    NAME EQUALS RealNum {
                   if (!stmtErr)
		     if (!RealSyms.add($1, $3))
		       delete $3;             // some error adding symbol
		   stmtErr = FALSE;
		 }

             |   RealNumList COMMA NAME EQUALS RealNum {
                   if (!stmtErr)
		     if (!RealSyms.add($3, $5))
		       delete $5;             // some error adding symbol
		   stmtErr = FALSE;
		 }
             ;

 RealNum:        NUMBER {
                   currRealNum = new RandomReal;
		   currRealNum->dist = const_real;
		   currRealNum->val = $1;
		   $$ = currRealNum;     // push RandomReal on semantic stack  
		 }

             |   NAME {
                   currRealNum = (RandomReal*) RealSyms.lookup($1);
		   if (currRealNum == NULL) { // no such symbol?
		     sprintf(errmsg, "undefined real number:  %s", $1);
		     yyerror(errmsg);
		   }
		   else
		     $$ = currRealNum;     // push RandomReal on semantic stk
		 }

             |   UNIFORM LPAREN NUMBER COMMA NUMBER RPAREN {
                   currRealNum = new RandomReal;
		   currRealNum->dist = uniform_real;
		   currRealNum->uniformLow = $3;
		   currRealNum->uniformHigh = $5;
		   $$ = currRealNum;     // push RandomReal on semantic stack  
		 }

             |   RANDOM LPAREN NUMBER COMMA NUMBER RPAREN {
                   currRealNum = new RandomReal;
		   currRealNum->dist = rand_real;
		   currRealNum->uniformLow = $3;
		   currRealNum->uniformHigh = $5;
		   $$ = currRealNum;     // push RandomReal on semantic stack  
		 }

             |   EXPNTL LPAREN NUMBER RPAREN {
                   currRealNum = new RandomReal;
		   currRealNum->dist = expntl_real;
		   currRealNum->mean = $3;
		   $$ = currRealNum;     // push RandomReal on semantic stack  
		 }

             |   NORMAL LPAREN NUMBER COMMA NUMBER RPAREN {
                   currRealNum = new RandomReal;
		   currRealNum->dist = nrm_real;
		   currRealNum->mean = $3;
		   currRealNum->variance = $5;
		   $$ = currRealNum;     // push RandomReal on semantic stack  
		 }

             |   ProbDist {
	           currProbDist = $1;    // grab distr from top of stack
	           if (currProbDist->allReals()) { // are all elements reals? 
	             currRealNum = new RandomReal;
		     currRealNum->dist = userDist_real;
		     currRealNum->userDist = currProbDist;
		     $$ = currRealNum;   // push RandomReal on semantic stack
		     }
		   else {                // at least one object not a real
		     yyerror("one or more undefined real nums in distribution");
		     delete currProbDist; // trash it
		   }
		 }
             ;

 FileStmt:       FILEKW {currFileInfo.type = data_file;} FileList Semi
             |   INDEXKW {currFileInfo.type = index_file;} FileList Semi
             ;

 FileList:       FileListEl
             |   FileList COMMA FileListEl
             ;

 FileListEl:     NAME EQUALS File { 
                   CreateFile($1, $3, 0); 
		   ResetFile();           // clear accumulated attributes
		 }

             |   NUMBER EQUALS File { 
	           currFile = $3;
		   if (currFile->dist != const_file)
		     yyerror("only constant files may be unnamed");
		   else {
		     if (notafile((int) $1))
		       CreateFile(NULL, $3, (int) $1); 
		     else
		       yyerror("cannot reuse an existing data/index file ID");
		   }
		   ResetFile();           // clear accumulated attributes
		 }

             |   NUMBER DOTDOT NUMBER EQUALS File {
		   if ($5->dist != const_file)
		     yyerror("only constant files may be unnamed");
		   else {
		     if ($3 > $1) 
		       for (i = (int) $1; i <= (int) $3; i++) {
			 if (stmtErr) break; // dont cascade errors
			 if (notafile(i)) {
			   currFile = new RandomFile; // replicate file
			   currFile->dist = const_file;
			   CreateFile(NULL, currFile, i);
			 }
			 else
			   yyerror("cannot reuse an existing data/index file ID");
		       }
		     else {
		       sprintf(errmsg, "%d..%d is an invalid range", (int) $1, 
			       (int) $3);
		       yyerror(errmsg);
		     }
		   }
		   ResetFile();           // clear accumulated attributes
		   ($5)->fileNo = -1;     // mess up fileNo so destructor only
		                          // cleans up rand file and nothing
					  // else by mistake
		   delete $5;             // only used as a model for clones
		 }
             ;

 File:           NUMBER {
		   if (notafile((int) $1)) {
		     sprintf(errmsg, "undefined file:  %d", (int) $1);
		     yyerror(errmsg);
		   }
		   else 
		     $$ = ConstFiles[(int) $1]; // push file on semantic stack
		 }

             |   NAME {
                   currFile = (RandomFile*) FileSyms.lookup($1); // find symbol
		   if (currFile == NULL){ // no such symbol?
		     sprintf(errmsg, "undefined file:  %s", $1);
		     yyerror(errmsg);
		   }
		   else
		     $$ = currFile;       // push file on semantic stack
		 }

             |   ProbDist {
	           currProbDist = $1;    // grab distr from top of stack
	           if (currProbDist->allFiles(currFileInfo.type)) {
	             currFile = new RandomFile;
		     currFile->dist = userDist_file;
		     currFile->userDist = currProbDist;
		     $$ = currFile;       // push file on semantic stack
		   }
		   else {                 // at least one bad thing in dist
		     delete currProbDist; // trash it
		   }
		 }

             |   UNIFORM LPAREN NUMBER COMMA NUMBER RPAREN {

                   // Check existence of all files in distribution

		   for (i = (int) $3; i <= (int) $5; i++) {
		     if (notafile(i)) {
		       sprintf(errmsg, "range includes an undefined file:  %d",
			       i);
		       yyerror(errmsg);
		       break;
		     }
		     if (ConstFiles[i]->type != currFileInfo.type) {
		       yyerror("range includes both index and data files");
		       break;
		     }
		   }

		   if (i > (int) $5) {        // all files checked out
		     currFile = new RandomFile;
		     currFile->dist = random_file;
		     currFile->randomLow = (int) $3;
		     currFile->randomHigh = (int) $5;
		     $$ = currFile;       // push file on semantic stack
		   }
		 }

             |   LESSTHAN FileAttrList GREATERTHAN {
                   currFile = new RandomFile;
		   currFile->dist = const_file;
		   $$ = currFile;       // push file on semantic stack
		 }
             ;

 FileAttrList:   FileAttr
             |   FileAttrList COMMA FileAttr
  
             |   FileAttrList error {yyerror("bad file attribute"); }
             ;

 FileAttr:       ID EQUALS NUMBER {
		   if (notafile((int) $3))
		     currFileInfo.fileNo = (int) $3;
		   else
		     yyerror("cannot reuse an existing data/index file ID");
		 }

              |  TUPLESIZE EQUALS NUMBER {
		   if ($3 <= 0.0)
		     yyerror("tuplesize must be > 0");
		   else
		     currFileInfo.tupleSize = (int) $3;
		 }

              |  NUMTUPLES EQUALS NUMBER {
		   if ($3 <= 0.0)
		     yyerror("number of tuples must be > 0");
		   else
		     currFileInfo.numTuples = (int) $3;
		 }

             |   NODES EQUALS { currIntSet = new IntegerSet(NumDiskNodes); }
                 IntSet {
		     
		   // Verifies the numbers are reasonable for node IDs
		     
		   if (currIntSet->setType == random_intset)
		     yyerror("random set of nodes not allowed");
		   else {

		     for (i = 0; i < currIntSet->card; i++) 
		       if (notanode(currIntSet->set[i])) break;
		     
		     // Should really check for duplicate node IDs too !!!!
		       
		     if (i >= currIntSet->card) { // all Node IDs check out
		     
		       // Translates list of node ids into list of node objects

		       for (i = 0; i < currIntSet->card; i++) 
			 currIntSet->set[i] = (int) Nodes[currIntSet->set[i]];
					   
		       currFileInfo.storageNode = (Node**) currIntSet->set;
		       currFileInfo.numNodes = currIntSet->card; 
		     }
		     else {
		       yyerror("one or more invalid nodes specified");
		       delete currIntSet;
		     }
		   }
		 }

             |   DISKS EQUALS {currIntSet = new IntegerSet(DisksPerNode); }
                 IntSet {
		     
		   // Verifies the numbers are reasonable for disk IDs

		   if (currIntSet->setType == random_intset)
		     yyerror("random set of nodes not allowed");
		   else {

		     for (i = 0; i < currIntSet->card; i++) 
		       if (notadisk(currIntSet->set[i])) break;

		     // Should really check for duplicate disk IDs too !!!!

		     if (i >= currIntSet->card) { // all disk IDs check out
		       currFileInfo.storageDisk = currIntSet->set;
		       currFileInfo.numDisks = currIntSet->card; //# of disks
		     }
		     else {
		       yyerror("one or more invalid disks specified");
		       delete currIntSet;
		     }
		   }
		 }

             |   KEYSIZE EQUALS NUMBER {
	           currFileInfo.keySize = (int) $3;
	         }

             |   TYPE EQUALS CLUSTEREDKW {
	           currFileInfo.indexType = CLUSTERED;
	         }

             |   TYPE EQUALS UNCLUSTEREDKW {
	           currFileInfo.indexType = UNCLUSTERED;
	         }

             |   FILEKW EQUALS NAME {
                   currFile = (RandomFile*) FileSyms.lookup($3); // find symbol
		   
		   if (currFile == NULL){ // no such symbol?
		     sprintf(errmsg, "undefined file:  %s", $3);
		     yyerror(errmsg);
		   }
		   else {
		     currFileInfo.indexedFile = currFile;
		     
		     if (currFile->dist != const_file)
		       yyerror("idices may only be created on constant files");

		     if (currFile->type != data_file)
		       yyerror("idices may only be created on data files");
		   }
	         }

             |   FILEKW EQUALS NUMBER {
		   if (notafile((int) $3)) {
		     sprintf(errmsg, "undefined file:  %d", (int) $3);
		     yyerror(errmsg);
		   }
		   else {
		     currFileInfo.indexedFile = ConstFiles[(int) $3];
		     
		     if (currFileInfo.indexedFile->dist != const_file)
		       yyerror("idices may only be created on constant files");

		     if (currFileInfo.indexedFile->type != data_file)
		       yyerror("idices may only be created on data files");
		   }
	         }

             ;

 IntSet:         LBRACE IntSetList RBRACE
             ;

 IntSetList:     NUMBER { currIntSet->add((int) $1); } 

             |   IntSetList COMMA NUMBER { currIntSet->add((int) $3); } 
   
             |   NUMBER DOTDOT NUMBER {
	           if ((int) $3 >= (int) $1)
		     for (i = (int) $1; i <= (int) $3; i++)
		       if (currIntSet->add(i) != 0) break;
		 }

             |   IntSetList COMMA NUMBER DOTDOT NUMBER { 
	           if ((int) $5 >= (int) $3)
		     for (i = (int) $3; i <= (int) $5; i++)
		       if (currIntSet->add(i) != 0) break;
		 }

             |   RealNum SPLAT RealNum {
	           currIntSet->setType = random_intset; // change from default
		   delete currIntSet->set; // kill fixed set storage
		   currIntSet->set = NULL;
		   currIntSet->randSet = new RandomSet(currIntSet->max);
	           currIntSet->randSet->elements = $1;
	           currIntSet->randSet->cardinality = $3;
		 }

             |   NUMBER error {yyerror("bad set definition"); } 
             |   IntSetList error {yyerror("bad set definition"); }
             ;
          
 JobClassStmt:   JOBCLASS JobClassList Semi;

 JobClassList:   NAME EQUALS JobClass 
                   { CreateJobClass($1, $3); }

             |   JobClassList COMMA NAME EQUALS JobClass 
                   { CreateJobClass($3, $5); }

             |   NAME error {yyerror("bad job class statement"); }
             |   JobClassList error {yyerror("bad job class statement"); }
             ;

 JobClass:       LESSTHAN {currJClass = new JobClass;} 
                 JClassDefList GREATERTHAN {
		   $$ = currJClass;      // push job class on semantic stack
		 }
             ;

 JClassDefList:  JClassDefStmt
             |   JClassDefList COMMA JClassDefStmt
  
             |   JClassDefList error {yyerror("bad job class definition list");}
             ;

 JClassDefStmt:  POPULATION EQUALS NUMBER {
		   if ($3 <= 0.0)
		     yyerror("population must be > 0");
		   else
		     currJClass->population = (int) $3;
		 }

             |   PRIMARY {

	           /* note we cannot set stopCount to the population at this
		      point (which is what it is supposed to be), since 
		      the POPULATION= keyword may not have been recognized
		      yet.  Instead, we just change the value to something
		      non-negative, and the JobClass::start method will 
		      set its value to equal the job class population before
		      kicking off the terminals of this class. */

                   currJClass->stopCount = 0;
		 }

             |   MAXMPL EQUALS NUMBER {
	           currJClass->specifiedMaxMPL = (int) $3;
		 }

             |   THINKSECS EQUALS RealNum {
	           currJClass->thinksecs = $3;
		 }

             |   PLANS EQUALS ProbDist {
		   currProbDist = $3;
		   if (currProbDist->allPlans())  // are all objects plans?
		     currJClass->plans = currProbDist;       
		   else {
		     yyerror("one or more undefined plans in distribution");
		     delete currProbDist;         // trash it
		     }
		 }
             ;

 PlanStmt:       PLAN PlanList Semi
             ;

 PlanList:       NAME PlanAttrList EQUALS Plan 
                   { CreatePlan($1, $4, $2); }
             |   PlanList COMMA NAME PlanAttrList EQUALS Plan
                   { CreatePlan($3, $6, $4); }

             |   NAME error {yyerror("bad plan statement"); }
             |   PlanList error {yyerror("bad plan statement"); }
             ;

 /* If more plan attributes are defined, then this production should be a
    real list, which would return some structure containing all the attributes
    as its value.  Right now, it just returns a boolean, since there is only
    one plan attribute (stats) which has a boolean value. */

 PlanAttrList:   LPAREN STATS RPAREN { $$ = TRUE; }

             |   { $$ = FALSE; }
             ;

 Plan:           Query {
                   $$ = $1;  // just pass it on up
		 }

             |   UPDATE File WHERE Query {
	           yyerror("statement not supported yet");
		 }
             ;

 Query:          LPAREN Query RPAREN {
                   $$ = $2;             // just pass it on up
		 }

             |   NAME {
		   if (($$ = (RandomPlan*) PlanSyms.lookup($1)) == NULL) {
		     sprintf(errmsg, "undefined plan:  %s", $1);
		     yyerror(errmsg);
		   }
		 }

             |   ProbDist {
                   currProbDist = $1;
	           if (currProbDist->allPlans()) { // are all objects plans?
		     currPlan = new RandomPlan;
		     currPlan->dist = userDist_plan;
		     currPlan->userDist = currProbDist;
		     $$ = currPlan;
		   }
		   else {                 
		     yyerror("one or more undefined plans in distribution");
		     delete currProbDist;     // trash it
		     }
		 }

             |   QueryOp { 
	           $$ = $1;  // just pass it on up
		 }
             ;

 QueryOp:        PROJ Sites MemReq LPAREN Query COMMA RealNum RPAREN {
                   // proj(Query, retain %) 
	           
		   yyerror("statement not supported yet");
                 }
 
             |   ScanType Sites MemReq LPAREN File RPAREN {
                   //  no selection predicate, no projection 
		   
		   if ($1 == file_scan && $5->type != data_file)
		     yyerror("cannot perform a file scan on an index file");
		   if ($1 == index_scan && $5->type != index_file)
		     yyerror("cannot perform an index scan on a data file");

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = SELECT;
		   currPlan->file = $5;
		   currPlan->sel = new RandomReal;
		   currPlan->sel->dist = const_real;
		   currPlan->sel->val = 1.0;
		   if ($1 == index_scan) { //need to default index selectivity?
		     currPlan->isel = new RandomReal;
		     currPlan->isel->dist = const_real;
		     currPlan->isel->val = 1.0;
		   }
		   currPlan->proj = new RandomReal;
		   currPlan->proj->dist = const_real;
		   currPlan->proj->val = 1.0;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
                 }

             |   ScanType Sites MemReq LPAREN File COMMA RealNum RPAREN {
                   // fscan(File, selectivity %) 
		   
		   if ($1 == file_scan && $5->type != data_file)
		     yyerror("cannot perform a file scan on an index file");
		   if ($1 == index_scan && $5->type != index_file)
		     yyerror("cannot perform an index scan on a data file");
		   if ($1 == index_scan)
		     yyerror("index scan requires 2 selectivities:  operator sel and index sel");

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = SELECT;   
		   currPlan->file = $5;
		   currPlan->sel = $7;
		   currPlan->proj = new RandomReal;
		   currPlan->proj->dist = const_real;
		   currPlan->proj->val = 1.0;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
                 }

             |   ScanType Sites MemReq LPAREN File COMMA RealNum COMMA 
                 RealNum RPAREN{
                   // fscan(File, selectivity %, project retain %) -or-
                   // iscan(File, op selectivity %, index selectivity %)
		   
		   if ($1 == file_scan && $5->type != data_file)
		     yyerror("cannot perform a file scan on an index file");
		   if ($1 == index_scan && $5->type != index_file)
		     yyerror("cannot perform an index scan on a data file");

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = SELECT;   
		   currPlan->file = $5;
		   currPlan->sel = $7;
		   if ($1 == file_scan)
		     currPlan->proj = $9;
		   else {
		     currPlan->isel = $9;  // must default projection to 100%
		     currPlan->proj = new RandomReal;
		     currPlan->proj->dist = const_real;
		     currPlan->proj->val = 1.0;
		   }
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
                 }

             |   ScanType Sites MemReq LPAREN File COMMA RealNum COMMA RealNum 
                 COMMA RealNum RPAREN {
                   // iscan(File, op sel %, index sel %, proj %)
		   
		   if ($1 == file_scan)
		     yyerror("illegal file scan syntax");
		   if ($1 == index_scan && $5->type != index_file)
		     yyerror("cannot perform an index scan on a data file");

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = SELECT;   
		   currPlan->file = $5;
		   currPlan->sel = $7;
		   currPlan->isel = $9;
		   currPlan->proj = $11;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
                 }

             |   STOREOP Sites MemReq LPAREN Query RPAREN {
	           // store(query) - write query result in new file on disk 0

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = STORE;
		   currPlan->lchild = $5;  // unary ops use left child
		   currPlan->disk = new RandomReal; // default is random disk
		   currPlan->disk->dist = rand_real; 
		   currPlan->disk->uniformLow = 0;
		   currPlan->disk->uniformHigh = DisksPerNode-1;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
		 }

             |   STOREOP Sites MemReq LPAREN Query COMMA RealNum RPAREN {
	           // store(query, disk #) - write query result in new file

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = STORE;
		   currPlan->lchild = $5;  // unary ops use left child
		   currPlan->disk = $7;    
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
		 }

             |   TERMSEND Sites MemReq LPAREN Query RPAREN {
	           // termsend(query) - send query result to terminal

		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = TERMOP;
		   currPlan->lchild = $5;  // unary ops use left child
		   $$ = currPlan;

		   // site and memory specs are ignored

		   if ($2 != NULL) delete $2;
		   if ($3 != NULL) delete $3;
		   
		 }

             |   AGGR Sites MemReq LPAREN Query COMMA RealNum RPAREN {
                   // aggr(Query, cost/tuple) 
	           
		   yyerror("statement not supported yet");
	         }

             |   SORT Sites MemReq LPAREN Query RPAREN {
                   // sort(query) 
	           
		   yyerror("statement not supported yet");
	         }

             |   SORTUNIQUE Sites MemReq LPAREN Query RPAREN {
                   // sortunique(query) 
	           
		   yyerror("statement not supported yet");
	         }

             |   UNIQUE Sites MemReq LPAREN Query RPAREN {
                   // unique(query) 
	           
		   yyerror("statement not supported yet");
	         }

             |   JoinType Sites MemReq LPAREN Query COMMA Query COMMA RealNum 
                 RPAREN {
                   // join(query, query, selectivity %) 
		   
		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = JOIN;   // more specific please!!!
		   currPlan->lchild = $5;
		   currPlan->rchild = $7;
		   currPlan->sel = $9;
		   currPlan->proj = new RandomReal; // default proj% to 100%
		   currPlan->proj->dist = const_real;
		   currPlan->proj->val = 1.0;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
	         }

             |   JoinType Sites MemReq LPAREN Query COMMA Query COMMA RealNum 
                 COMMA RealNum RPAREN {
                   // join(query, query, selectivity %, project retain %) 
		   
		   currPlan = new RandomPlan;
		   currPlan->dist = const_plan;
		   currPlan->opCode = JOIN;   // more specific please!!!
		   currPlan->lchild = $5;
		   currPlan->rchild = $7;
		   currPlan->sel = $9;
		   currPlan->proj = $11;
		   if ($2 != NULL) {       // if site assmt expr present
		     currPlan->sites = ($2)->set;
		     currPlan->sizeSites = ($2)->card;
		     currPlan->siteSet = ($2)->randSet;
		   }
		   currPlan->memReq = $3;
		   $$ = currPlan;
	         }

             |   XPROD Sites MemReq LPAREN Query COMMA Query RPAREN {
                   // xprod(query, query) 
	           
		   yyerror("statement not supported yet");
	         }

             |   UNION Sites MemReq LPAREN Query COMMA Query RPAREN {
                   // union(query, query) 
	           
		   yyerror("statement not supported yet");
	         }

             |   INTERSECTION Sites MemReq LPAREN Query COMMA Query COMMA 
                 RealNum RPAREN {
                   // intersect(query, query, % of union) 
	           
		   yyerror("statement not supported yet");
	         }

             |   DIFFERENCE Sites MemReq LPAREN Query COMMA Query COMMA 
                 RealNum RPAREN {
                   // difference(query, query, % of first file retained) 
	           
		   yyerror("statement not supported yet");
	         }

             |   PROJ error {yyerror("bad project syntax");}
             |   ScanType error {yyerror("bad scan syntax");}
             |   AGGR error {yyerror("bad aggregation syntax");}
             |   SORT error {yyerror("bad sort syntax");}
             |   SORTUNIQUE error {yyerror("bad sortunique syntax");}
             |   UNIQUE error {yyerror("bad unique syntax");}
             |   JoinType error {yyerror("bad join syntax");}
             |   XPROD error {yyerror("bad xprod syntax");}
             |   UNION error {yyerror("bad union syntax");}
             |   INTERSECTION error {yyerror("bad intersection syntax");}
             |   DIFFERENCE error {yyerror("bad difference syntax");}
             ;

 JoinType:       HASHJOIN  { $$ = hash_join; }
             |   INDEXJOIN { $$ = index_join; }
             |   NLJOIN    { $$ = nl_join; }
             |   MERGEJOIN {
	           $$ = merge_join;
		   printf("Merge join?  This is Wisconsin, not IBM!");
		 }
             ;

 ScanType:       INDEXSCAN { $$ = index_scan; }
             |   FILESCAN  { $$ = file_scan; }
             ;

 Sites:          LBRACE { currIntSet = new IntegerSet(NumDiskNodes); }
                 IntSetList RBRACE {

		   // Verifies the numbers are reasonable for node IDs
		   // Should really check for duplicate node IDs too !!!!

		   if (currIntSet->setType == constant_intset) {

		     for (i = 0; i < currIntSet->card; i++) 
		       if (notanode(currIntSet->set[i])) break;

		     if (i < currIntSet->card) { // some node didnt check out
		       yyerror("one or more invalid nodes specified");
		       delete currIntSet;
		       $$ = NULL;           // tell op to take defaults
		     } else {
		       $$ = currIntSet;     // give op the set of sites
		     }
		   } else {
		     $$ = currIntSet;       // cant verify random sets
		   }
		 }

             |   {$$ = NULL;} /* user wants default site assignment */
             ;
  
 MemReq:         LBRACK RealNum COMMA RealNum RBRACK {
                   yyerror("obsolete memory syntax");
		 }

             |   LBRACK RealNum MemUnit RBRACK {
                   currMemReq = new MemReq;
		   currMemReq->min = currMemReq->max = $2;
		   currMemReq->minUnits = currMemReq->maxUnits = $3;
		   $$ = currMemReq;
		 }

             |   LBRACK RealNum MemUnit COMMA RealNum MemUnit RBRACK {
                   currMemReq = new MemReq;
		   currMemReq->min = $2;
		   currMemReq->minUnits = $3;
		   currMemReq->max = $5;
		   currMemReq->maxUnits = $6;
		   $$ = currMemReq;
		 }

             |   {$$ = NULL;} /* user wants default memory requirements */
             ;

 MemUnit:        BPOOL {$$ = BUFPOOL_UNITS;}
             |   MIN {$$ = MINREQ_UNITS;}
             |   BYTES {$$ = BYTE_UNITS;}
             |   KBYTES {$$ = KBYTE_UNITS;}
             |   MBYTES {$$ = MBYTE_UNITS;}

 NodeStmt:       NODES EQUALS LESSTHAN NodeAttrList GREATERTHAN SEMI
             ;

 NodeAttrList:   NodeAttr
             |   NodeAttrList COMMA NodeAttr
  
             |   NodeAttrList error {yyerror("bad node definition syntax"); }
             ;

 NodeAttr:       POPULATION EQUALS NUMBER {
                   if ((int) $3 > MAX_NODES-1 || (int) $3 <= 0) {
     		     sprintf(errmsg, "number of disk nodes must be between 0 and %d", MAX_NODES);
		     yyerror(errmsg);
		   }
		   else {
                     NumDiskNodes = (int) $3; 
		     NumNodes = NumDiskNodes+1;
		   }
		 }
		     
             |   DISKS EQUALS NUMBER { 
	           if ((int) $3 > MAX_DISKSPERNODE || (int) $3 <= 0) {
		     sprintf(errmsg, "number of disks per node must be between 0 and  %d", MAX_DISKSPERNODE);
		     yyerror(errmsg);
		   }
		   else 
		     DisksPerNode = (int) $3; 
		 }
		     
             |   CPUS EQUALS NUMBER { 
	           if ($3 <= 0.0)
		     yyerror("number of cpus per node must be > zero");
		   else
		     CPUsPerNode = (int) $3;
		 }

             |   CPUMIPS EQUALS NUMBER {
	           if ($3 <= 0.0)
		     yyerror("MIPS must be > zero");
		   else
		     Mips = $3;
		 }

             |   KBYTES EQUALS NUMBER { 
	           if ($3 <= 0.0)
		     yyerror("node memory must be > zero");
		   else
		     KBytesPerNode = (int) $3;
		 }

             |   PAGESIZE EQUALS NUMBER { 
	           if ($3 <= 0.0)
		     yyerror("page size must be > zero");
		   else
		     PageSize = (int) $3;
		 }

             |   STARTIO_INST EQUALS NUMBER { Inst_StartIO = $3; }

             ;

 DiskStmt:       DISKS EQUALS LESSTHAN DiskAttrList GREATERTHAN SEMI
             ;

 DiskAttrList:   DiskAttr
             |   DiskAttrList COMMA DiskAttr
  
             |   DiskAttrList error {yyerror("bad disk definition syntax"); }
             ;

 DiskAttr:       PAGESPERCYL EQUALS NUMBER {
                   if ($3 <= 0.0)
		     yyerror("pages per cylinder must be > 0");
		   else
		     PagesPerCyl = (int) $3;
		 }

             |   SEEKFACTOR EQUALS NUMBER { 
                   if ($3 <= 0.0)
		     yyerror("seek factor must be > 0");
		   else
		     SeekFactor = $3; 
		 }

             |   MINLATENCY EQUALS NUMBER { MinLatency = $3; }

             |   ROTATION EQUALS NUMBER { 
                   if ($3 <= 0.0)
		     yyerror("rotation speed must be > 0");
		   else
		     DiskRotation = $3; 
		 }

             |   SETTLE EQUALS NUMBER { SettleTime = $3; }

             |   XFERRATE EQUALS NUMBER { XferRate = $3; }

             |   CACHEPAGES EQUALS NUMBER { DiskCachePages = (int) $3; }

             |   CACHECONTEXTS EQUALS NUMBER { DiskCacheContexts = (int) $3; }
             ;

 Semi:	         SEMI { yyerrok; stmtErr = FALSE; }
             ;
%%

#include "lex.yy.c"

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       yyerror
//  
//    Function:   Does general formatting of all error messages.
//
//    Change Activity:
//
//      10/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void yyerror(char *s) {
  extern int yynerrs;
  
  yynerrs++;
  stmtErr = TRUE;
  yywhere();
  fputs(s, yyerfp);
  putc('\n', yyerfp);
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       yywhere
//  
//    Function:   Prints the location pinpointing part of parser err messages.
//
//   Change Activity:
//
//      10/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void yywhere() {
  bool colon = FALSE;
  
  // Prints file name part of message

  if (SourceFileName && *SourceFileName && strcmp(SourceFileName, "\"\"")) {
    char *cp=SourceFileName;
    
    int len = strlen(SourceFileName);
    
    if (*cp == '"')
      ++cp, len -= 2;
    if (strncmp(cp, "./", 2) == 0)
      cp += 2, len -= 2;
    fprintf(yyerfp, "file %.*s", len, cp);
    colon = TRUE;
  }

  // Figures out and prints line number part of message

  if (yylineno > 0) {
    if (colon)
      fputs(", ", yyerfp);
    fprintf(yyerfp, "line %d", yylineno - (*yytext == '\n' || ! *yytext));
    colon = TRUE;
  }

  // Prints a nearby token where error occurred, if it can

  if (*yytext) {
    register int i;
    
    for (i=0; i<20; i++)
      if (!yytext[i] || yytext[i] == '\n')
	break;
    if (i) {
      if (colon)
	putc(' ', yyerfp);
      fprintf(yyerfp, "near \"%.*s\"", i, yytext);
      colon = TRUE;
    }
  }

  if (colon)
    fputs(": ", yyerfp);
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       yywrap
//  
//    Function:   Gets called by lex when EOF is detected.
//
//    Change Activity:
//
//      10/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

int yywrap() {
  if (Flags[TRACE_PARSER]) {
    printf("\n\nParser symbol table dump:\n-------------------------\n");
    RealSyms.print(4);
    FileSyms.print(4);
    PlanSyms.print(4);
    JobClassSyms.print(4);
    printf("\n");
  }

  return(1); // tell lex to go ahead and quit 
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       ResetFile
// 
//    Function:   Sets all accumulated file attributes to invalid values so
//                the next file definition can be checked for missing/invalid
//                attributes.
//
//    Change Activity:
//
//      11/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void ResetFile() {
  currFileInfo.fileNo = currFileInfo.numTuples =
    currFileInfo.tupleSize = currFileInfo.numDisks =
      currFileInfo.keySize = currFileInfo.indexType = 
	currFileInfo.numNodes = -1;
  currFileInfo.storageNode = NULL;
  currFileInfo.storageDisk = NULL;
  currFileInfo.indexedFile = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       CreateFile
// 
//    Function:   Handles a "file =" statement reduction.  Does final error 
//                checks and default handling, creating a real file or index
//                for a random file whose distribution = const_file.
//
//    Change Activity:
//
//      10/91 - Original version.  KPB
//   11/01/91 - Added support for indices.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void CreateFile(char* sym, RandomFile* f, int fileID) {
  char* fileName;
  char  fileDigits[10];
  int   numDisks, numNodes, indexType, fileNum, keySize;

  // Builds a file name for error msgs

  if (sym != NULL)                     // named file?
    fileName = sym;
  else {
    sprintf(fileDigits, "%d", fileID);
    fileName = &(fileDigits[0]);
  }

  if (!stmtErr)
    f->type = currFileInfo.type;       // set index or data file

  // Validity checks and determines file/index ID for constant files

  if (!stmtErr && f->dist == const_file) {
    if (sym != NULL) {                 // named file?
      if (currFileInfo.fileNo < 0) {
	sprintf(errmsg, "id must be specified for named files/indices (%s)", 
		fileName);
	yyerror(errmsg);
      }
      else {
	fileNum = currFileInfo.fileNo; // file id is from id= keyword
      }
    }
    else {                             // unnamed file
      if (currFileInfo.fileNo >= 0) {
	sprintf(errmsg, "id cannot be specified for unnamed files/indices (%s)",
		fileName);
	yyerror(errmsg);
      }
      else {
	fileNum = fileID;              // file id is from left side of = sign
      }
    }
	  
    f->fileNo = fileNum;               // set in RandomFile so notafile() works
  }

  // Handles creation of constant data files

  if (!stmtErr && f->dist == const_file && currFileInfo.type == data_file) { 
					    
    // Checks that required parameters are specified 

    if (currFileInfo.numTuples < 0) {
      sprintf(errmsg, "numtuples must be specified for file %s", fileName);
      yyerror(errmsg);
    }
    
    if (currFileInfo.keySize >= 0) 
      yyerror("keysize parameter doesn't make sense for a data file");

    if (currFileInfo.indexType >= 0)
      yyerror("type parameter doesn't make sense for a data file");

    if (currFileInfo.indexedFile != NULL)
      yyerror("file parameter doesn't make sense for a data file");

    // If no errors so far, takes appropriate defaults and cReates a data file
	
    if (!stmtErr) {
      if (currFileInfo.numNodes < 0) {
//	if (!Quiet) 
//	  printf("File %s defaults to full declustering (nodes %d..%d)\n", 
//		 fileName, 0, NumDiskNodes-1);
	numNodes = NumDiskNodes;
	currFileInfo.storageNode = Nodes;
      }
      else
	numNodes = currFileInfo.numNodes;
      
      if (currFileInfo.numDisks < 0) {
//	if (!Quiet) 
//	  printf("File %s defaults to one disk per node (disk 0)\n", fileName);
	numDisks = 1;
	currFileInfo.storageDisk = new int; // get a one element int list
	*(currFileInfo.storageDisk) = 0;    // disk zero of the node
      }
      else
	numDisks = currFileInfo.numDisks;
      
      if (currFileInfo.tupleSize < 0) {
//	if (!Quiet)
//	  printf("File %s tuple size defaults to %d bytes\n",
//		 fileName, DEFAULT_TUPSIZE);
	currFileInfo.tupleSize = DEFAULT_TUPSIZE;
      }
      
      f->file = new File(fileNum,
			 numNodes,
			 currFileInfo.storageNode,
			 numDisks,
			 currFileInfo.storageDisk,
			 currFileInfo.numTuples,
			 currFileInfo.tupleSize);

      /* Since every constant file must be locatable by number as well as by
	 name, places both the "real" file object AND the corresponding 
	 random file object (distribution = constant) in well known locations
	 which are indexed by file number.  */
      
      ConstFiles[fileNum] = f;              // store in a well known location
      Relations[fileNum] = f->file;         // store in a well known loc
    }
  } // constant data file
    
  // Handles creation of constant index files

  if (!stmtErr && f->dist == const_file && currFileInfo.type == index_file) { 
					    
    // Checks that required parameters are specified

    if (currFileInfo.indexedFile == NULL) {
      sprintf(errmsg, "file must be specified for index %s", fileName);
      yyerror(errmsg);
    }
    
    if (currFileInfo.numNodes >= 0)
      yyerror("nodes parameter doesn't make sense for an index file");

    if (currFileInfo.numDisks >= 0)
      yyerror("disks parameter doesn't make sense for an index file");

    if (currFileInfo.tupleSize >= 0)
      yyerror("tuplesize parameter doesn't make sense for an index file");

    if (currFileInfo.numTuples >= 0)
      yyerror("numtuples parameter doesn't make sense for an index file");

    // If no errors so far, takes appropriate defaults and cReates an index
	
    if (!stmtErr) {
      if (currFileInfo.indexType < 0) {
//	if (!Quiet) printf("Index file %s defaults to unclustered\n", fileName);
	indexType = UNCLUSTERED;
      }
      else
	indexType = currFileInfo.indexType;
      
      if (currFileInfo.keySize < 0) {
//	if (!Quiet) printf("Index file %s defaults to keysize 4\n", fileName);
	keySize = 4;
      }
      else
	keySize = currFileInfo.keySize;
    
      f->index = new Index(fileNum,
			  currFileInfo.indexedFile->file,
			  keySize,
			  indexType);
      
      /* Since every constant file must be locatable by number as well as by
	 name, places both the "real" file object AND the corresponding 
	 random file object (distribution = constant) in well known locations
	 which are indexed by file number.  */
      
      ConstFiles[fileNum] = f;            // store in a well known location
      Indices[fileNum] = f->index;        // store in a well known loc
    }
  } // constant index file
    
  /* If no previous errors, and this is a named file, attempts to add the 
     name/object pair to the file symbol table, backing out everything if
     some error occurs adding the symbol. */

  // want to make up symbols for unnamed files??

  if (!stmtErr && sym != NULL) {
    if (!FileSyms.add(sym, f)) {          // if some error adding symbol
      if (f->dist == const_file && f->type == data_file) { // const data file?
        ConstFiles[f->fileNo] = NULL;
	Relations[f->fileNo] = NULL;
        delete f->file;                   // delete real data file 
      }

      if (f->dist == const_file && f->type == index_file) { //const index file?
        ConstFiles[f->fileNo] = NULL;
	Indices[f->fileNo] = NULL;
        delete f->index;                  // delete real index file 
      }

      delete f;                           // delete random file
    }
  }
	
  stmtErr = FALSE;                        // clear err indicator
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       CreateJobClass
// 
//    Function:   Handles a "class =" statement reduction, does final error
//                checks and defaulting, and if OK, adds the job class 
//                definition to the symbol table.
//
//    Change Activity:
//
//      10/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void CreateJobClass(char* sym, JobClass* jc) {

  if (!stmtErr) {
    if (jc->plans == NULL) {
      sprintf(errmsg, "no plan distribution for job class %s", sym);
      yyerror(errmsg);
    }
    
    if (jc->population == -1) {
//    if (!Quiet) printf("Job class %s population defaults to 1.\n", sym);
      jc->population = 1;
    }
    
    if (jc->thinksecs == NULL) {
//    if (!Quiet) printf("Job class %s think time defaults to 0.\n", sym);
      jc->thinksecs = new RandomReal;
      jc->thinksecs->dist = const_real;
      jc->thinksecs->val = 0.0;
    }
  }

  if (stmtErr)               // if previous err with this class definition
    delete jc;
  else
    if (!JobClassSyms.add(sym, jc))
      delete jc;             // some error adding symbol
    else {                   // everything ok, job class is defined
      jc->name = sym;        // put symbolic name inside jobclass object
      JobClasses[NumClasses++] = jc; // stuff in well known location
      NumTerminals += jc->population; // accumulate total terminal population
    }

  stmtErr = FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//
//    Name:       CreatePlan
// 
//    Function:   Handles a "plan =" statement reduction.
//
//    Change Activity:
//
//      10/91 - Original version.  KPB
//
///////////////////////////////////////////////////////////////////////////////

void CreatePlan(char* sym, RandomPlan* p, bool stats) {
  
  if (!stmtErr) {

    // Determines type of scheduling based on plan naming convention - KLUDGE!

    for (i = 0; i < MAX_SCHEDTYPES; i++)
      if (strncmp(sym, PlanPrefixes[i], strlen(PlanPrefixes[i])) == 0) {
	p->schedType = i;
	break;
      }
    
    // Default to left deep if no recognizable plan prefix is present

    if (i >= MAX_SCHEDTYPES)
      p->schedType = LEFTDEEP;

    // Unless the plan is already named, saves the symbolic name for this 
    // plan inside itself for debugging.  Plans record their first given name.
    // This MUST be performed before calling the activateStats() method, since
    // it must know which plans are named versus unnamed!!!!

    if (p->name == NULL) p->name = sym;

    // If statistics are desired for this plan name, sets up to do so.

    if (stats) p->activateStats(p);

    if (!PlanSyms.add(sym, p))
      delete p;                // some error adding symbol
  }
  stmtErr = FALSE;
}















