
///////////////////////////////////////////////////////////////////////////////
//
//  Name:        language.h
//
//  Title:       Parallel Relational Plan Language (PRPL) class definitions
//
//               Defines data objects and types that are shared between 
//               the PRPL parser and other modules (specifically the terminals)
//
//               o ProbDist
//               o RandomReal
//               o RandomSet
//               o RandomFile
//               o RandomPlan
//               o SymTable
//
//  WARNING!  WARNING!  DANGER WILL ROBINSON!!!!
//               
//               Any new fields which are added to RandomX objects must also
//               be dealt with in the clone methods for each of these objects!
//               The clone methods "replicate" random objects (and in some
//               cases, its subobjects) for each terminal instance. "Replicate"
//               is in quotes, because although some elements of these objects
//               must, in fact, be duplicated for each terminal, others must be
//               shared by all terminal instances. 
//
//  Change Activity:
//
//     10/01/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef languageh
#define languageh 

class QueryQueue;
class PlanDef;
class File;
class Index;
class QueryPlan;

///////////////////////////////////////////////////////////////////////////////

typedef enum probDist {determ_dist, rand_dist};
typedef enum distObjType {name_obj, number_obj}; 
typedef enum fileType {data_file, index_file};

typedef struct distel {   // an element of a distribution or sequence
  void* thing;            // the object itself
  distObjType type;       // type of object 
  float prob;             // probability of this element for a distribution
  float cumProb;          // cumulative probability from first el to this one
  int repFactor;          // repetition factor for a deterministic seq element
  int repCount;           // number of times this element has been emitted
  distel* next;           // next element in list
} distEl;

// A user specified probability distribution([...]) or deterministic sequence
// (<...>)

class ProbDist : public Object {
  private:
    probDist dist;        // random or deterministic sequence?
    int elCount;          // number of elements in distribution/sequence
    float totProb;        // sum of the probabilities specified till now
    distEl* firstDistEl;  // pointer to a first dist or sequence element
    distEl* lastDistEl;   // pointer to a last dist or sequence element
    distEl* lastChosenEl; // pointer to last chosen sequence element

    // internal, generalized add element to distribution list
    void add(void* thing, float prob, int repFactor, distObjType t);

  public:
    ProbDist(probDist d);
    ~ProbDist(){};
    void add(char* name); // adds a name to a dist or a sequence
    void add(float num);  // adds a number to a dist or a sequence
    void add(char* name, float prob); // adds a name/prob to a distribution
    void add(float num, float prob);  // adds a number/prob to a distribution
    void add(char* name, int repFactor); // adds a name*n to a determ seq
    void add(float num, int repFactor);  // adds a num*n to a deterministic seq
    bool spread();        // Spreads unused probability evenly across elements

    /* Note the following constraint checking functions are meant to be called
       AFTER the ENTIRE distribution list has been recognized.  They cannot
       be done while building the list, since the parser is bottom up, and
       doesn't know the constraints until higher up in the parse tree. */

    bool allFiles(fileType ftype); // TRUE if all objects files of type ftype
    bool allReals();      // TRUE if all objects check out as real numbers
    bool allPlans();      // TRUE if all objects check out as plans

    // choose picks an object, sets t to its type and sets reset to TRUE if
    // a deterministic sequence has wrapped back to the first element.

    void* choose(distObjType& t, bool& reset); 

    ProbDist* clone();   // replicates a probdist object
    void print(int indent);
};

///////////////////////////////////////////////////////////////////////////////

typedef enum realDist 
  {const_real, uniform_real, rand_real, expntl_real, nrm_real, userDist_real};

class RandomReal : public Object { // represents a fixed or random real number
  public:
    realDist dist;           
    float val;            // its value if a constant
    float uniformLow;     // low value if uniform/random distributed
    float uniformHigh;    // high value if uniform/random distributed
    float mean;           // mean if exponential or normally distributed
    float variance;       // variance if normally distributed
    ProbDist* userDist;   // ptr to user specified distribution 

    RandomReal(){};
    ~RandomReal(){};
    float choose();       // returns a randomly selected real num
    void print(int indent);
    RandomReal* clone();  // replicates a random real object
};

///////////////////////////////////////////////////////////////////////////////

class RandomSet : public Object { // represents a random set of integers
  public:
    int max;              // max allowed cardinality and element value
    RandomReal* elements; // distribution to choose elements
    RandomReal* cardinality; // distribution to choose cardinality

    RandomSet(int mx) { max = mx; };
    ~RandomSet(){};
    int* choose(int& card); // returns a set of integers of size card
    void print(int indent);
    RandomSet* clone();   // replicates a random set object
};

///////////////////////////////////////////////////////////////////////////////

typedef enum fileDist {const_file, random_file, userDist_file};

class RandomFile : public Object {  // represents a fixed or random relation
  public:
    fileType type;        // index_file or data_file
    fileDist dist;
    int fileNo;           // index/data file number for constant files
    File* file;           // ptr to file object if a "constant" data/index file
    Index* index;         // ptr to index object if a "constant" index file 
    int randomLow;        // low value if uniformly distributed
    int randomHigh;       // high value if uniformly distributed
    ProbDist* userDist;   // ptr to user specified distribution 

    RandomFile(){};
    ~RandomFile();
    void* choose(fileType& ftype);  // returns a randomly selected file 
    void print(int indent);
    RandomFile* clone();  // replicates a random file object
};

///////////////////////////////////////////////////////////////////////////////

class MemReq {             // place to store [min, max] memory req construct
  public:
    RandomReal* min;
    int minUnits;			      
    RandomReal* max;
    int maxUnits;			      

    MemReq(){ min = NULL; max = NULL; };
    ~MemReq(){};
};

class JobClass;
typedef enum planDist {const_plan, userDist_plan};

class RandomPlan : public Object { // represents a fixed or random query plan
  private:
    QueryPlan* fillin(QueryPlan* parent, JobClass* jc); // fill in a const plan

  public:
    planDist dist;
    char* name;           // valid if this is the root on rhs of a plan stmt
    RandomPlan* statsPlan;// plan object to record stats in if stats == TRUE 
    int schedType;        // technique to schedule these types of plans
    int opCode;           // physical algebra operator
    RandomReal* sel;      // operator selectivity, if applicable
    RandomReal* isel;     // index selectivity, if applicable
    RandomReal* proj;     // % projection, if applicable
    RandomReal* cost;     // per tuple operator cost (# instr), if applicable
    RandomReal* disk;     // disk # for query result file 
    RandomFile* file;     // file, if applicable
    RandomPlan* lchild;   // left child, if applicable (unary ops use lchild) 
    RandomPlan* rchild;   // right child, if applicable
    ProbDist* userDist;   // ptr to user specified distribution 
    MemReq* memReq;       // memory requirements of operator
    table *bufRefs;       // table of buffer references for this plan
    table *bufHits;       // table of buffer hits for this plan
    table *bufDirtys;     // table of buffer pages dirtied for this plan
    table *diskReads;     // table of disk reads by this plan
    table *diskCacheHits; // table of disk cache hits by this plan
    table *CPUInstr;      // table of CPU instructions used by this plan
    table *respTimes;     // tbl of resp times for this plan (root nodes only)
    table *memAlloc;      // memory allocation for this class

    // Set of sites can to run the op at can be specified as a constant or
    // random set.  Both fields NULL mean take the default site assignment.

    int* sites;
    int sizeSites;				   
    RandomSet* siteSet;

    RandomPlan();
    ~RandomPlan(){};
    QueryPlan* choose(QueryPlan* parent, JobClass* jc); //stamp out a real plan
    void print(int indent);
    RandomPlan* clone();  // replicates a random plan object
    void writeStats();    // write out plan related statistics
    void writeStats(char* symbol);  // write out plan related statistics
    void activateStats(RandomPlan* flaggedPlan); // turn on plan based stats
}; 

typedef enum joinTypes {hash_join, index_join, nl_join, merge_join};
typedef enum scanTypes {index_scan, file_scan};

///////////////////////////////////////////////////////////////////////////////

class JobClass : public Object { // represents a class of terminals
  public:
    char* name;           // class name (symbol table entry)
    RandomReal* thinksecs;// think time in seconds
    int population;       // number of terminals in this job class
    ProbDist* plans;      // distribution of plans this terminal class submits
    int completions;      // used to calculate throughput for this class
    table *respTimes;     // table of response time stats for this class
    table *bufRefs;       // table of buffer references for this job class
    table *bufHits;       // table of buffer hits for this job class
    table *bufDirtys;     // table of buffer pages dirtied for this job class
    table *diskReads;     // table of disk reads by this class
    table *diskCacheHits; // table of disk cache hits by this class
    table *CPUInstr;      // table of CPU instructions used by this class
    table *diskResp;      // disk req resp time (including queueing)
    table *armResp;       // disk arm resp time (no queueing)
    table *memAlloc;      // memory allocation for this class
    table *buckets;       // number of hjoin buckets for this class
    int stopCount;        // cnt of terms that must complete b4 simulation is
                          // terminated.  Ignored if < 0 (non stopping class)
    int specifiedMaxMPL;  // maximum MPL allowed for this job class
    int actualMaxMPL;     // actual max MPL observed in the system
    int execMPL;          // num of queries executing in the class
    int currentMPL;       // num of queries admitted in the class
    QueryQueue *queryWaitQueue;  // job class MPL waiting queue
    qtable *memoryWaitTable; // table to collect stats on memory waitingQ
    double timeWeightedMPL;  // maintained by scheduler
    double prevMPLChange;    // last time a query admitted/departed

    JobClass(); 
    ~JobClass(){};
    void start();         // instatiates/starts all instances of a job class
    void writeStats();    // write out job class statistics
    void print(int indent);
}; 

///////////////////////////////////////////////////////////////////////////////

// Symbol table stuff.  Note that a particular object may have multiple names.

typedef struct symEntry { // a symbol table entry
    char* name;           // name of the symbol
    Object* thing;        // object that symbol refers to
    symEntry* next;       // ptr to next symbol or NULL for end of list
} symTabEntry;
  
class SymTable {          // symbol table entry
  private:
    symTabEntry* head;    // single head, single thead list for now

  public:
    SymTable();
    ~SymTable(){};
    bool add(char* name, Object* value); // TRUE if insert sucessful
    Object* lookup(char* name); // returns an object, given its name
    void print(int indent);
    void writeStats();    // calls a writeStats method on every object in table
};

extern SymTable RealSyms; // real num symbol table
extern SymTable FileSyms; // file symbol table
extern SymTable PlanSyms; // plan symbol table
extern SymTable JobClassSyms; // terminal symbol table
extern RandomFile *ConstFiles[MAX_FILES]; 

#define    notafile(fID)  ((fID) >= MAX_FILES || (fID) < 0 || ConstFiles[(fID)] == NULL || ConstFiles[(fID)]->fileNo != (fID))

#define    notadisk(dID)  ((dID) >= DisksPerNode ||(dID) < 0)

#define    notanode(nID)  ((nID) >= NumNodes || (nID) < 0 || Nodes[(nID)] == NULL || Nodes[(nID)]->id != (nID))

#endif
