///////////////////////////////////////////////////////////////////////////////
//
//  Name:        mailbox.c
//
//  Title:       UW database simulator mailbox class implementations.
//
//               The mbox class is an enhancement to the CSIM mailbox class 
//               which allows us to charge CSIM CPU facilities for message 
//               receive protocol pathlength costs.  It is a generic class and
//               is not intended for external use outside of its subclasses.
//
//               The MailBox and TermMailBox classes are futher refinements
//               which just indicates whether the owner is a terminal or a
//               node (send/receive processing is different for each case).
//
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "node.h"
#include "op.h"
#include "msg.h"
#include "mailbox.h"

///////////////////////////////////////////////////////////////////////////////
//
//  mbox::recv         Issues a CSIM receive on the mailbox, and if the mailbox
//                     owner is a node (versus a terminal), then charges the 
//                     owning node's CPU for the message protocol pathlength.
//
///////////////////////////////////////////////////////////////////////////////

Msg* mbox::recv() { 
  Msg* m;

  // Waits for an incoming message
  if (Flags[TRACE_MSG]) 
    	printf("mbox::recv - issuing mailbox::receive on mbox %x\n", this);
  mailbox::receive((int *) &m);

  return(m);
}

///////////////////////////////////////////////////////////////////////////////
// 
//  Mailbox::Mailbox   Associates a mailbox with a node.
// 
///////////////////////////////////////////////////////////////////////////////

Mailbox::Mailbox(char* name, Node* n, Op* o): (name) {
  owningNode = n;
  owningOp = o;
  type = NODE;
  if (Flags[TRACE_MSG]) 
    printf("Mailbox::Mailbox - instantiating Mailbox %x (%s)\n", this, name);
}

///////////////////////////////////////////////////////////////////////////////
//
//  Mailbox::recv     Issues a CSIM receive on the mailbox associated with
//                      a node (versus a terminal) and then charges the 
//                     owning node's CPU for the message protocol pathlength.
//
///////////////////////////////////////////////////////////////////////////////

Msg* Mailbox::recv() { 
  Msg* m;
  float netCost;
  extern facility Network;

  // Waits for an incoming message
  if (Flags[TRACE_MSG]) 
    	printf("mbox::recv - issuing mailbox::receive on mbox %x\n", this);

  mailbox::receive((int *) &m);

  // If the message was not short circuited, then simulates an appropriate
  // wire delay for the receiver.  This delay is on the receiver's end so
  // that the sender is not blocked, and can overlap processing with its 
  // receivers (pipelining).

  if (m->sourceBox->owningNode != m->destBox->owningNode) {
    if (m->numBytes <= 0) {
      printf("mailbox::receive - < 0 bytes (%d), msg id %d\n", 
              m->numBytes, m->id);
      netCost = 0.55;
    }
    else if (m->numBytes < 65) netCost = 0.60; // milliseconds
    else netCost = 0.000582 * m->numBytes + 0.835;

////////////////// the above formula approximates the following measured costs.
//
//    	   else if (numBytes < 1025 ) netCost = 1.45;
//    	   else if (numBytes < 2049 ) netCost = 2.00;
//    	   else if (numBytes < 5010 ) netCost = 3.80;
//    	   else if (numBytes <= 8192) netCost = 5.60;
//    	   else 
//	   {
//     	      printf("msg::send - too many bytes (%d), msg id %d\n", 
//		numBytes, id);
//      	      netCost = 6.10;
//    	   }
////////////////////////

    // if appropriate delay time has not passed make the reciver wait
    if ((clock - m->timestamp) < netCost)
        Network.use(netCost - (clock - m->timestamp));  //consume n/w bandwidth
  }

  // Accounts for protocol costs on the receivers's CPU

  owningNode->cpu.use(((float) PROTOCOLCOST) / (Mips * 10e3)); // milliseconds
  if (owningOp != NULL) owningOp->cnt_CPUInstr += PROTOCOLCOST;
  return(m);
}

///////////////////////////////////////////////////////////////////////////////
// 
//  TermMailbox::TermMailbox   Associates a mailbox with a terminal
// 
///////////////////////////////////////////////////////////////////////////////

TermMailbox::TermMailbox(char* name, Terminal* t): (name) {
  owningTerm = t;
  type = TERMINAL;
  if (Flags[TRACE_MSG]) 
    printf("TermMailbox::TermMailbox - instantiating TermMailbox %x (%s)\n", 
	   this, name);
}













