// MISC constants

#define MAXBUCKETS 150
#define MAXJOINS 11
#define MAXDISKS MAX_NODES*MAX_DISKSPERNODE
#define STORESIZE 330
#define JOINSIZE 346
#define SPLITENTRYSIZE 32
#define SCANSIZE 422 + (MAX_NODES*SPLITENTRYSIZE)
#define DONESIZE 24
#define NEWTASKSIZE 4
#define NOTIFYSIZE 4
#define NEWSPLITSIZE 140 + (MAX_NODES*SPLITENTRYSIZE)
#define HTSTATUSSIZE 16

// various opcodes
#define SELECT       1
#define STORE        2
#define JOIN         3
#define SPOOL        4
#define DELAY        5
#define TERMOP       6

// various operator types
#define REMOTE 0
#define LEFTLOCAL 1
#define RIGHTLOCAL 2
#define XMED 3

// various Scheduler type classifications
#define LEFTDEEP 1
#define RIGHTDEEPONE 2
#define RIGHTDEEPALL 3
#define RIGHTDEEPSTATIC1 4
#define RIGHTDEEPSTATICa 5
#define RIGHTDEEP4 6
#define RIGHTDEEP5 7
#define RIGHTDEEP6 8
#define BUSHYMAX 9
#define BUSHY1 10
#define BUSHY2 11
#define BUSHY3 12
#define BUSHY4 13
#define BUSHY5 14
#define BUSHY6 15
#define BUSHY7 16
#define BUSHY8 17                 /* don't forget to update MAX_SCHEDTYPES! */

#define MAX_SCHEDTYPES 18         /* don't forget to update this! */

// Array of plan name prefixes which correspond to the previous scheduling
// disciplines.  Their position in the array must correspond to the value
// of the scheduling discipline as defined above.  Need something explicit
// and more general than this to specify scheduling in a plan!  KLUDGE

#ifdef isparser 
  static char PlanPrefixes[MAX_SCHEDTYPES][10] = {
    "1reserved",
    "ld",
    "rd1",
    "rda",
    "rds1",
    "rdsa",
    "rd4",
    "rd5",
    "rd6",
    "bsm",
    "bs1",
    "bs2",
    "bs3",
    "bs4",
    "bs5",
    "bs6",
    "bs7",
    "bs8"
  };
#else
  extern char PlanPrefixes[MAX_SCHEDTYPES][10];
#endif

// various JOIN phases, operator initialized to 1
#define BUILD 1
#define PROBE 2

// Codes which indicate how to interpret minMem and maxMemNeeded fields

#define BUFPOOL_UNITS 1
#define MINREQ_UNITS  2
#define BYTE_UNITS    3
#define KBYTE_UNITS   4
#define MBYTE_UNITS   5

class SplitTable;
class Mailbox;
class File;
class Index;
class JobClass;
class RandomPlan;

class QueryPlan : public Object {
  public:
    int		transId;
    char*       name;           // valid if this node was stamped from the root
                                // node of a PRPL "plan x = ..." stmt
    JobClass*   jobClass;       // pointer to job class object
    RandomPlan* statsPlan;      // random plan in which to store plan oriented 
                                // stats.  Stats are ignored (tossed) if NULL
    int         schedType;      // type of scheduling for this query
    int		opCode;		// an opcode
    QueryPlan*	parent;
    TIME	startTime;
    int		outTupSize;	// size of result tuples in bytes
    int		noActiveDisk;	// number of nodes operator will be active on
    bool	activeSites[MAX_NODES]; // true if operator is to run on a node
    Mailbox*	sites[MAX_NODES]; // port of the process running the operator
    int		activeCnt;	// number of operators currently active
    int		tupleCnt;	// number of tuples actually processed ???
    double      execTime;       // estimated execution time of the operator
    int         numOutTuples;   // number of estimated output tuples of the op
      
// delete the next two when scheduler is fixed!!!
    int         maxPagesNeeded; // maximum memory needed by the operator
    int         minPagesNeeded; // minimum memory needed by the operator

    int         totalMinPagesNeeded[MAX_NODES];  // min. numer of pages needed
    int         totalMaxPagesNeeded[MAX_NODES];  // max. number of pages needed
    float       minMemNeeded;   // minimum memory requested by the operator
    int         minMemUnits;    // how to interpret mimMemNeeded value
    float       maxMemNeeded;   // maximum memory requested by the operator
    int         maxMemUnits;    // how to interpret maxMemNeeded value

    // The following statistics are initially accumulated by the buffer and 
    // disk code on behalf of a particular operator instance.  The operator
    // destructors take responsibility for accumulating their statistics in 
    // their corresponding query plan node before deleting themselves.
    // The accumStats() method on QueryPlan objects recursively accumulates
    // statistics for child plan nodes into the parent.  A terminal then 
    // takes responsibility for accumulating the statistics from the root node
    // of a query plan into the job class the query was issued under.

    int         cnt_bufRef;     // count of buffer references by this plan node
    int         cnt_bufHit;     // count of buffer hits by this plan node
    int         cnt_bufDirty;   // count of buf pages dirtied by this plan node
    int         cnt_diskRead;   // count of disk reads by this plan node
    int         cnt_diskCacheHit; // count of disk cache hits by this plan node
    int         cnt_CPUInstr;   // count of CPU instructions used by this node

    QueryPlan();
    ~QueryPlan(){};
    void print(int indent);
    void printName(int indent);
    virtual void accumStats(){}; // rollup stats from child nodes into this one
    void accumPlanStats();      // accumulates plan-based stats in randomPlan
    void pullStats(QueryPlan* source); // pull stats from source to this node
    QueryPlan* findRoot();  // return the root of the query tree

};

//////////////////////////////////////////////////////////////////////////////

class SelectPlan: public QueryPlan {
private:
    static class SelectPlan* freeSelectPlan;	// list of free requests
  public:
    SelectPlan* nextPlan;
    File*	file;		
    Index*      index;             // if != NULL, index to use for scan on file
    float	selFactor;	   // selectivity factor of operator
    float	indexSelectivity;  // sel factor of the index (iscans only)
    int		inTupSize;
    SplitTable*	splitTable;	
    int         selectType;
    
    SelectPlan();
    ~SelectPlan();
    void print(int indent);
    void* operator new(size_t size);		       // new is overloaded
    void operator delete(void* oldPlan, size_t size);   // delete is overloaded
    void accumStats();          // rollup stats from child nodes into this one
};

//////////////////////////////////////////////////////////////////////////////

class JoinPlan: public QueryPlan {
private:
    static class JoinPlan* freeJoinPlan;	// list of free requests
  public:
    JoinPlan*   nextPlan;
    float	selFactor;	// selectivity factor of operator
    int		inStreamsL;	// number of eof's expected from left child
    int		inStreamsR;	// number of eof's expected from right child
    int		inTupSizeL;	// size of tuples (in bytes) from left child
    int		inTupSizeR;	// size of tuples (in bytes) from right child
    SplitTable*	splitTable;	
    QueryPlan*	leftChild;
    QueryPlan*	rightChild;
    File*	xmedFiles[MAX_NODES];
    int         phase;          // indicates join phase (BUILD or PROBE)
    double      estBuildCost;   // estimated cost of the build operation
    int         numBuckets;     // number of buckets. KLUDGE for now assume a
                                // fixed number of buckets for the join
    float       weightFactor;   // extra entries to be added for sending more
                                //    tuples to bucket 0
    int         bucketKMemory;  //  memory needed for buckets 2..n
    int         skewNode;       // KLUDGE FOR SKEW:indicates the join skew node
    int         skewEntry;      // KLUDGE FOR SKEW:indicates entry to be used
                                //  in split table (set in makesched.c)

    JoinPlan();
    ~JoinPlan();
    void print(int indent);
    void* operator new(size_t size);		       // new is overloaded
    void operator delete(void* oldPlan, size_t size);   // delete is overloaded
    void accumStats();         // rollup stats from child nodes into this one
};

//////////////////////////////////////////////////////////////////////////////

class StorePlan: public QueryPlan {
  private:
    static class StorePlan* freeStorePlan;	// list of free requests
  public:
    StorePlan*  nextPlan;
    File*	file;		
    int		diskId;		// which disk the store should use
    int		inStreams;	// number of eof's expected from child
    int		inTupSize;	// size of tuples (in bytes) from child
    TIME	writeTime;      // time first tuple was written by any node
    QueryPlan*	child;
    int         storeType;

    StorePlan();
    ~StorePlan(){};
    void print(int indent);
    void* operator new(size_t size);		       // new is overloaded
    void operator delete(void* oldPlan, size_t size);   // delete is overloaded
    void accumStats();          // rollup stats from child nodes into this one
};


//////////////////////////////////////////////////////////////////////////////
// Simulates a user terminal - just takes tuples as input with no action
//
class TermPlan: public QueryPlan {
  private:
    static class TermPlan* freeTermPlan;	// list of free requests
  public:
    QueryPlan*	child;
    TermPlan*   nextPlan;
    int		inStreams;	// number of eof's expected from child
    int		inTupSize;	// size of tuples (in bytes) from child

    TermPlan();
    ~TermPlan(){};
    void print(int indent);
    void* operator new(size_t size);		       // new is overloaded
    void operator delete(void* oldPlan, size_t size);   // delete is overloaded
    void accumStats();          // rollup stats from child nodes into this one
};
