#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "msg.h"
#include "mailbox.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "query.h"
#include "sched.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        gen_sched.c
//
//  Function:    a generic query scheduler
//                 
//  Change Activity:
//
//
///////////////////////////////////////////////////////////////////////////////

extern Node* Nodes[];
extern Mailbox* GlobalSchedPort;
extern int NumDiskNodes;

extern void startQuery(Node*, QueryMsg*);
extern planCost(Schedule*, Schedule*, QueryPlan*);
extern int actualResourceAllocation(QueryMsg*, Schedule*&, Schedule*&);
void DeleteSched(Schedule*&);
extern MemoryQueue *GlobalQueryWaitQueue;
extern double TimeWeightedMPL;
extern int MaxMPL;
extern int BufPages[];
extern double PrevMPLChange;
extern int CurrentMPL;

allocateResources(QueryMsg*, Schedule*&, Schedule*&);
void QueryScheduler(Node* , QueryMsg*, Schedule*, Schedule*);
void startQuery(Node*, QueryMsg*);
void processOperatorDone(CntlMsg*, int&);
void genericScheduler(Schedule*, Schedule*, Mailbox*, int transId);
int activate(QueryPlan* , Mailbox*, int);
void startOperator(QueryPlan* , Mailbox*, int);
int getPortIds(QueryPlan*, Mailbox*);
void fillSplitAddrs (QueryPlan*, SplitTable*, int);
SplitTable* copySplitTable(SplitTable*);
void sendCntlMsg(QueryPlan* , int, Mailbox* );
void sendSplitTblMsg(QueryPlan* , SplitTable* , Mailbox* );


allocateResources(QueryMsg* reqMsg, Schedule*& schedule, 
		  Schedule*& waitSchedule) 
{
     int e;

     // form the schedules. also does the resource allocation
     e = actualResourceAllocation(reqMsg, schedule, waitSchedule);

     // if the query not scheduled properly, decrease class mpl
     if (e != OK)
     {
         reqMsg->query->jobClass->currentMPL--;

         // Send "the results" back to the terminal, reusing the request pkt
         reqMsg->send(reqMsg->sourceBox, GlobalSchedPort);

         // destroy the old schedules
	 DeleteSched(schedule);
	 DeleteSched(waitSchedule);


     }
     return(e);
}

//////////////////////////////////////////////////////////////////////
//
// Generic Query Scheduler
//   In - 1) node - 
//        2) reqMsg - scheduler request message, contains pointer to 
//                      query tree  
//   Action
//     calls the appropriate query schedule generator and then calls 
//     the generic scheduler to actually schedule all the operators
//
///////////////////////////////////////////////////////////////////////
void QueryScheduler(Node* node, QueryMsg* reqMsg, 
		    Schedule* schedule, Schedule* waitSchedule) 
{

    Mailbox* 	qschedMbox;
    int nodeNo, numSites, totalMemory;
    QueryReq* nextQuery;
    JobClass* queryClass;

    create("QueryScheduler");

    if (Flags[TRACE_SCHED]) {
    	printf("query scheduler process started\n");
    	reqMsg->print("query scheduler got");
    };

    // check the global wait queue to see if another query can run
    // done here because only after resource allocation do we know 
    // the actual number of pages left in the buffer pool
    //  this check enables multiple queries to start after one query completes
    //  as each starting query after doing it's own resource allocation will
    //  also check the next waiting query
    GlobalQueryWaitQueue->checkQueue();

    queryClass = reqMsg->query->jobClass;

    // add the weighted sum for the last interval to the total
    TimeWeightedMPL += (clock - PrevMPLChange) * CurrentMPL;
    queryClass->timeWeightedMPL += (clock - queryClass->prevMPLChange) 
	        * queryClass->execMPL;
    queryClass->prevMPLChange = clock;

    CurrentMPL++;
    if (CurrentMPL > MaxMPL) MaxMPL = CurrentMPL;
    queryClass->execMPL++;
    if (queryClass->execMPL > queryClass->actualMaxMPL)
	        queryClass->actualMaxMPL = queryClass->execMPL;


    // update MPL and last change time
    PrevMPLChange = clock;  

    // get a mail mailbox for operators to communicate with
    qschedMbox = new Mailbox("qschedMbox", node);

    //
    //  cost the query plan
    //
    //      planCost(schedule, waitSchedule, reqMsg->query);  


    //
    // call the generic query scheduler
    //
    genericScheduler(schedule, waitSchedule, qschedMbox, 
		     reqMsg->query->transId);


    //after the query completes update the bufPages array
    //also update memory usage statistics by calculating the
    //average amount of memory given to the query over all the sites
    totalMemory = numSites = 0;
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    {
	BufPages[nodeNo] += reqMsg->query->totalMaxPagesNeeded[nodeNo];
	totalMemory += reqMsg->query->totalMaxPagesNeeded[nodeNo];
	if (reqMsg->query->totalMaxPagesNeeded[nodeNo] > 0) numSites++;
    }
    if (reqMsg->query->statsPlan != NULL)
       reqMsg->query->statsPlan->memAlloc->record((int)(totalMemory * 1.0
                                                           / numSites));
    queryClass->memAlloc->record((int) (totalMemory * 1.0 / numSites));


    // if the root of the query is a store
    //    destroy the final result file created by the query
    if (reqMsg->query->opCode == STORE)
    {
        StorePlan* sp = (StorePlan*) reqMsg->query;
	if (sp->opCode == STORE && sp->storeType == REMOTE)
	      DeleteFile(sp->file->fileNo);
    }

    // check the global wait queue to see if another query can run now
    //    after buffers have been given back
    GlobalQueryWaitQueue->checkQueue();


    // add the weighted sum for the last interval to the total
    TimeWeightedMPL += (clock - PrevMPLChange) * CurrentMPL;
    queryClass->timeWeightedMPL += (clock - queryClass->prevMPLChange) 
             * queryClass->execMPL;

    // update MPL and last change time
    PrevMPLChange = clock;  
    queryClass->prevMPLChange = clock;  
   
    // decrement the class MPL
    queryClass->currentMPL--;
    queryClass->execMPL--;
    CurrentMPL--;


    // if queries waiting in the job class queue start the next query
    if (queryClass->queryWaitQueue->nonempty())
    {
       nextQuery = queryClass->queryWaitQueue->dequeue(NULL);
       if (Flags[TRACE_SCHED])
            printf("next job class query started");
       startQuery(nextQuery->node, nextQuery->reqMsg);
       delete nextQuery;
    }

    // Send "the results" back to the terminal, reusing the request pkt
    reqMsg->send(reqMsg->sourceBox, GlobalSchedPort);

    DeleteSched(schedule);
    DeleteSched(waitSchedule);
    delete qschedMbox;
}



/////////////////////////////////////////////////////////////////////////////
//
// process the completion message received from one of the running operators
//     increment the numOpsDone counter if the operator has completed
//
/////////////////////////////////////////////////////////////////////////////
void processOperatorDone(CntlMsg *cMsg, int& numOpsDone)
{

    JoinPlan* joinPlan;
    StorePlan* storePlan;

    // decrement number of active instances of operator
    cMsg->query->activeCnt--;
    cMsg->query->tupleCnt += cMsg->intData;

    if (cMsg->query->opCode == STORE) 
    {
        // get the response time from the store
        if (cMsg->floatData < ((StorePlan *) cMsg->query)->writeTime)
        {
	    ((StorePlan* ) cMsg->query)->writeTime = cMsg->floatData;
        }

        storePlan = (StorePlan*) cMsg->query;

	if (storePlan->storeType == XMED)
	{
	    //add file name to parent intermediate select op
	    if (storePlan->parent->opCode != JOIN)
	    {
	        printf("error - parent of spool not join");
	    }
	    else
	        ((JoinPlan*) storePlan->parent)->xmedFiles[cMsg->myNodeNum] =
		    cMsg->fileName;
	}

    }

    if (Flags[TRACE_SCHED])
    {
        cMsg->query->printName(5);
        printf(" instance has completed %d\n", cMsg->query->activeCnt );
    }

 
    // if no more active instances of the operator
    if (cMsg->query->activeCnt == 0) 
    {
        switch(cMsg->query->opCode)
	{
	    case JOIN:
	        joinPlan = (JoinPlan *) cMsg->query;
	        if (Flags[TRACE_SCHED]) 
		  printf(" JOIN operator has completed phase %d\n", 
			 joinPlan->phase );
		//
	        // if phase just completed is BUILD 
		//      set phase to PROBE
		// else
		//       PROBE phase has completed
		//
	        if (joinPlan->phase == BUILD)
		    joinPlan->phase = PROBE;
		else
		    // set the active count back to the original
		    joinPlan->activeCnt = joinPlan->noActiveDisk;
		break;
	    case SELECT:
	        if (Flags[TRACE_SCHED]) 
		    printf(" SELECT operator has completed\n");
	        break;
	    case STORE:
	        if (Flags[TRACE_SCHED]) 
		{
		   printf(" STORE operator has completed\n");
		   if ((Flags[TRACE_TERM]) && (storePlan->storeType == REMOTE))
		        printf("\ntime to first write tuple %f\n", 
			       ((StorePlan*) cMsg->query)->writeTime);
		}
	        break;
	    case TERMOP:
	        if (Flags[TRACE_SCHED]) 
		    printf(" TERMOP operator has completed\n");
	        break;
	}

	// increase number of operators done
	numOpsDone++;
    }

    
}



/////////////////////////////////////////////////////////////////////
//
//  genericScheduler
//     In - 1) schedule
//          2) qschedBox - mailbox for communication
//          3) transId - transaction ID
//     Action - traverses the schedule L->R and top->bottom and
//           activates all the operators. once all the operators in
//           a row have been activates waits for them to finish and
//           processes the next row
//
//           Q - change row-wise synchronization?
//
/////////////////////////////////////////////////////////////////////
void genericScheduler(Schedule* schedule, Schedule* waitSchedule,
		      Mailbox* qschedMbox, int transId)
{

    int totalOprs, numOpsDone ;
    Schedule *scheduleRow, *waitRow;
    Operator* opr;
    CntlMsg *cMsg ;


    //
    // for each row in the schedule
    //
    for(scheduleRow = schedule, waitRow = waitSchedule;
        scheduleRow != NULL, waitRow != NULL; 
        scheduleRow = scheduleRow->nextRow, waitRow = waitRow->nextRow)
    {


       //
       // for each operator in the pipeline group activate the operator
       //
       for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
            // activate the operator
	    activate(opr->oprPlan, qschedMbox, transId);

       //
       // count total number of operators to wait for
       // have to check for active cnts because the operator may have 
       // already finished while starting some other operators
       //
       totalOprs = 0 ;
       for(opr = waitRow->oprRow; opr != NULL; opr = opr->nextOpr)
           if (opr->oprPlan->activeCnt > 0)
	       totalOprs++;

       //
       // wait for all the operators to finish
       //
       numOpsDone = 0;

       // while all operators not finished
       while (numOpsDone < totalOprs)
       {
       
           // block waiting for next message
           cMsg = (CntlMsg*) qschedMbox->recv();
           if (cMsg->type != CONTROL)
	       printf("expected CONTROL message in getPortIds. got type=%d\n",
		      cMsg->type);
	   else
	   {
	       switch (cMsg->opCode)
	       {
	           case OPDONE:
	           case TBLSTATUS:
		       processOperatorDone(cMsg, numOpsDone);
		       break;
		   default:
		       printf("Unexpected control msg type received %d", 
			      cMsg->opCode);
		       printf("from node %d\n", 
			      cMsg->queryOp, cMsg->sourceBox->owningNode->id);

		       break;
	       }
	       CntlMsg:: free(cMsg);  // release message buffer
	   }
       }
    }
}



////////////////////////////////////////////////////////////////////////
//
// activate
//
//   Action - activate a node in a query tree
//
////////////////////////////////////////////////////////////////////////
activate(QueryPlan* currentNode, Mailbox* qschedMbox, int transId)
{
    JoinPlan*  jnNode;
    StorePlan* strNode;
    TermPlan* termNode;
    SplitTable *leftSplitTable, *rightSplitTable, *splitTable;
    int		cnt;

    switch(currentNode->opCode) 
    {
        case JOIN:

	    if (((JoinPlan*)currentNode)->phase == BUILD)
	    {
	        if (Flags[TRACE_SCHED]) 
		    printf("Starting BUILD phase of JOIN\n");
 
		// update the num buckets statistics if this is the first
		// bucket
		((JoinPlan*) currentNode)->jobClass->buckets->record(((JoinPlan*) currentNode)->numBuckets);

	        startOperator(currentNode, qschedMbox, transId); 

	        // get the private ports of the various processes involved
	        cnt = getPortIds (currentNode, qschedMbox);

	        // build a split table for left child
	        leftSplitTable = new SplitTable (cnt, 
				    ((JoinPlan*) currentNode)->inTupSizeL);

	        // build a split table for right child
	        rightSplitTable = new SplitTable (cnt, 
				    ((JoinPlan*) currentNode)->inTupSizeR);
		
	        // fill in the addresses
	        fillSplitAddrs (currentNode, leftSplitTable, cnt);
	        fillSplitAddrs (currentNode, rightSplitTable, cnt);

	        // copy split table to the left child 
	        jnNode = (JoinPlan*) currentNode;
	        if (jnNode->leftChild->opCode == SELECT)
		    ((SelectPlan*) jnNode->leftChild)->splitTable = 
		                             copySplitTable(leftSplitTable);
	        else 
		    if (jnNode->leftChild->opCode == JOIN)
		        ((JoinPlan*) jnNode->leftChild)->splitTable =
		                             copySplitTable(leftSplitTable);
		    else 
		        printf("in activate, funny opcode=%d\n",
			       currentNode->opCode);

 	        // copy split table to the right child 
	        if (jnNode->rightChild->opCode == SELECT) 
		    ((SelectPlan*) jnNode->rightChild)->splitTable =
		                             copySplitTable(rightSplitTable);
		else 
		    if (jnNode->rightChild->opCode == JOIN)
		        ((JoinPlan*) jnNode->rightChild)->splitTable =
		                             copySplitTable(rightSplitTable);
		    else 
		        printf("in activate, funny opcode=%d\n",
			       currentNode->opCode);
	      
		 delete leftSplitTable; 
		 delete rightSplitTable; 
	    }
	    else
	    { 
	        if (Flags[TRACE_SCHED]) 
		    printf("Starting PROBE phase of JOIN\n");

	        // join operator has already been started 
	        // so build phase should be over by the time we get here
	        // ignore overflow of joins but here is where it is handled
	        // in the real gamma system

	        // send split tables to all sites
	        sendSplitTblMsg(currentNode , 
				((JoinPlan*) currentNode)->splitTable, 
				qschedMbox);
	    }
            break;

        case SELECT:
	    if (Flags[TRACE_SCHED]) printf("Starting SELECT \n");

	    // start the select operator
	      startOperator(currentNode, qschedMbox, transId); 

	    // get the private ports of the various operators involved
	      cnt = getPortIds (currentNode, qschedMbox);
            break;

        case TERMOP:
	    if (Flags[TRACE_SCHED]) printf("Starting TERMOP \n");
	    termNode = (TermPlan*) currentNode;

	    // start the terminal operator
	    startOperator(currentNode, qschedMbox, transId); 

	    // get the private ports of the various processes involved
	    cnt = getPortIds (currentNode, qschedMbox);

	    // build a split table to feed data 
	    splitTable = new SplitTable (cnt, termNode->child->outTupSize);
		
	    // fill in the addresses
	    fillSplitAddrs (currentNode, splitTable, cnt);
	    
	    // copy split table to the child 
            if (termNode->child->opCode == SELECT)
		((SelectPlan*) termNode->child)->splitTable = 
		                                 copySplitTable(splitTable);
	    else 
		if (termNode->child->opCode == JOIN)
		      ((JoinPlan*) termNode->child)->splitTable = 
			                         copySplitTable(splitTable);
		else 
		      printf("in activate, funny opcode=%d\n",
			     currentNode->opCode);

	    delete splitTable;
	    break;

        case STORE:
	    if (Flags[TRACE_SCHED]) printf("Starting STORE \n");
	    strNode = (StorePlan*) currentNode;

	    // start the store operator
 	    startOperator(currentNode, qschedMbox, transId); 

	    // get the private ports of the various processes involved
	    cnt = getPortIds (currentNode, qschedMbox);

	    // build a split table to feed data 
	    splitTable = new SplitTable (cnt, strNode->inTupSize);
		
	    // fill in the addresses
	    fillSplitAddrs (currentNode, splitTable, cnt);

	    // copy split table to the child 
	    if (strNode->child->opCode == SELECT)
	        ((SelectPlan*) strNode->child)->splitTable = 
		                            copySplitTable(splitTable);
	    else 
	        if (strNode->child->opCode == JOIN)
		    ((JoinPlan*) strNode->child)->splitTable = 
			                          copySplitTable(splitTable);
		else 
		    printf("in activate, funny opcode=%d\n",
			   currentNode->opCode);
	    delete splitTable;
            break;
    }
    return(OK);
}


//////////////////////////////////////////////////////////////////////////////
//
//
//  start the operator (qry) on all sites where activeSite is TRUE
//
/////////////////////////////////////////////////////////////////////////////
void startOperator(QueryPlan* qry, Mailbox* qschedMbox, int transId)
{
    Mailbox *nodeMgr;
    QueryMsg *qMsg;
    int	i, cnt;

    // initialize some fields in the actual plan
    qry->activeCnt = 0;
    qry->tupleCnt = 0;
    qry->transId = transId;

    cnt=0;
    for (i=0; i <= NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    // operator is to run on this node 
            // get the mail box of the node manager on the appropriate node 
            nodeMgr = Nodes[i]->nodeMgrPort;
            qMsg = QueryMsg::alloc(); // get query message
	    qMsg->query = qry;
	    // if it is an intermediate select send xmed file name with msg
            if ((qry->opCode == SELECT) && 
		(((SelectPlan*) qry)->selectType == XMED))
	        qMsg->file = ((JoinPlan*) qry->parent)->xmedFiles[i];
	    qMsg->send(nodeMgr, qschedMbox);
	    cnt++;
	}
    }
    qry->activeCnt = cnt;
}

/////////////////////////////////////////////////////////////////////////////
//
// get the private ports of all the processes executing an operator
// return count of number of active processes
//
/////////////////////////////////////////////////////////////////////////////
int getPortIds (QueryPlan* qry, Mailbox* qschedMbox)
{
    int		replyCnt;  // number of responses received
    CntlMsg	*cMsg;
    int		cnt;
    int		nodeId;
    int         dummy;     // dummy argument for processOperatorDone

    cnt = qry->activeCnt;
    replyCnt = 0;

    // loop until the message ports of all the operator processes are received
    while (replyCnt < cnt)
    {
	// block waiting for next message
	cMsg = (CntlMsg*) qschedMbox->recv();
	if (cMsg->type != CONTROL)
	    printf("expected CONTROL message in getPortIds. got type=%d\n",
			cMsg->type);
	else
	{
	    switch (cMsg->opCode)
	    {
	    case NEWTASK:
		    // update list of active sites
		replyCnt++;
		nodeId = cMsg->intData;  // kind of a kludge but ok
		qry->sites[nodeId] = cMsg->sourceBox; // extract opr's port

//                if (Flags[TRACE_SCHED])
//                    printf("reply recieved replycnt %d from node %d\n", 
//                           replyCnt, cMsg->sourceBox->owningNode->id);

	        break;
	    case OPDONE:
                if (cMsg->query == qry)
                {
//                    if (Flags[TRACE_SCHED])
//                    printf("reply recieved replycnt %d from node %d\n", 
//                           replyCnt, cMsg->sourceBox->owningNode->id);
                }

            // should we do this for tblstatus also ????
            case TBLSTATUS:
                processOperatorDone(cMsg, dummy);
	        break;
	    default:
	        printf("Unexpected control message type received = %d\n", 
				cMsg->opCode);
		break;
	    }
	    CntlMsg:: free(cMsg);  // release message buffer
	}
    }

//    if (Flags[TRACE_SCHED])
//         printf("getPortIds done replycnt %d\n", replyCnt);

    return(replyCnt);
}
	
/////////////////////////////////////////////////////////////////////////////
//
// Use the mailbox's of the operators executing the parent operator
// to construct a split table for the child operator
// Return the number of entries in the splitTable
//
/////////////////////////////////////////////////////////////////////////////
void fillSplitAddrs (QueryPlan* qry, SplitTable* splitTable, int size)
{
    // qry points at the parent's query plan
    // splitTable points at the child's split table
    // size is the size of the splitTable

    int		i;
    int		splitCnt;

    splitCnt = 0;
    for (i=0; i  <= NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	      // parent operator is being run on node i
	      splitTable->splitArray[splitCnt].numTuples = 0;
	      splitTable->splitArray[splitCnt].destNode = qry->sites[i];
	      splitCnt++;
	}
    }
    if (splitCnt != size) 
	printf("\n\nmajor problem in fillSplitAddres, size=%d, splitCnt=%d\n\n"
	       ,size, splitCnt);
    splitTable->splitCnt = splitCnt;
}


// make a copy of a split table
SplitTable* copySplitTable(SplitTable* from)
{
    int 	i;
    SplitTable*	to;

    to = new SplitTable (from->splitCnt, from->tupleSize);
    to->splitCnt = from->splitCnt;
    for (i=0;i<from->splitCnt;i++) to->splitArray[i] = from->splitArray[i];
    return(to);
}


//
// ************************************************
//    old code. this procedure not used anymore
// ************************************************
// Send a control message to all processes executing an operator
//
void sendCntlMsg(QueryPlan* qry, int msgType, Mailbox* qschedMbox)
{
    CntlMsg 	*cMsg;
    int		i;
	
    for (i=0; i <= NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    cMsg = CntlMsg::alloc(); // get control message
	    cMsg->opCode = msgType;  
	    cMsg->send (qry->sites[i], qschedMbox);
	}
    }
}

// send the split table to all sites where qry running
void sendSplitTblMsg(QueryPlan* qry, SplitTable* splitTable,
	Mailbox* qschedMbox)
{
    CntlMsg 	*cMsg;
    int		i;
	
    for (i=0; i <= NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    qry->activeCnt++;
	    cMsg = CntlMsg::alloc(); // get control message
	    cMsg->opCode = NEWSPLIT;  
	    cMsg->splitTable = copySplitTable(splitTable);
	    cMsg->send (qry->sites[i], qschedMbox);
	}
    }
}

