#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "query.h"
#include "language.h"
#include "db.h"
#include "mailbox.h"
#include "msg.h"
#include "sched.h"
#include <math.h>


///////////////////////////////////////////////////////////////////////////////
//
//  Name:        makesched.c
//
//  Author:      MM
//
//  Function:    has procedures to form different types of schedules and
//               perform resource allocation on the query plan
//
///////////////////////////////////////////////////////////////////////////////

extern "C" {
  double ceil(double x);
  double floor(double x);
  double sqrt(double x);
  void exit(int);
}


#define MAX(A,B) ( ( (A) > (B) ) ? (A) : (B) )



extern MemoryQueue *GlobalQueryWaitQueue;
extern int BufPages[];
extern SkewFactor; // KLUDGE FOR SKEW
void DeleteSched(Schedule*&);

initialSchedule(QueryMsg*, Schedule*&, Schedule*&);
initQueryRequirements(QueryPlan*, bool, bool staticSched = FALSE);
makeSchedule(QueryPlan*, Schedule*&, Schedule*&, bool);
initResourceAllocation(QueryMsg*, Schedule*, Schedule*);
CanRun(QueryPlan*);
actualResourceAllocation(QueryMsg*, Schedule*&, Schedule*&);
void addOpr(Operator*&, QueryPlan*, int bucketNum = 1, int phase = -1);
void addLastRow(Schedule*&, Schedule*&, QueryPlan*, int bucketNum = 1);
void ldExec(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, 
	    Schedule*&, Schedule*&);
void rdExec(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, 
	    Schedule*&, Schedule*&, Schedule*&);
void rdExecAll(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, 
	    Schedule*&, Schedule*&, Schedule*&);
void bushyExec(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, Schedule*&, 
		Schedule*&, Schedule*&, bool);
void staticExec(QueryPlan*, Schedule*&, Schedule*&, Schedule*&, Schedule*&, 
		Schedule*&, Schedule*&, bool, bool, int memLeft[], 
		int totalMemory[]);

/////////////////////////////////////////////////////////////////////////////
//
//   initialSchedule
//     given the root of the query tree
//     1) does initial resource requirement calculation
//     2) creates a new schedule
//     3) checks to see if the query can run, if not it is sent to the global 
//           wait queue
//
//////////////////////////////////////////////////////////////////////////////
initialSchedule(QueryMsg* reqMsg, Schedule*& schedule, Schedule*& waitSchedule)
{
  int e;

  // set the number of buckets and max and min pages needed by each operator
  if ((reqMsg->query->schedType == RIGHTDEEPSTATIC1) ||
      (reqMsg->query->schedType == RIGHTDEEPSTATICa))
    e = initQueryRequirements(reqMsg->query, TRUE, TRUE);
  else
    e = initQueryRequirements(reqMsg->query, TRUE);
  if (e  != OK) return(NOTOK);

  // form the initial schedule
  if ((e = makeSchedule(reqMsg->query, schedule, waitSchedule, TRUE)) != OK) 
      return(NOTOK);

  // perform initial query resource allocation and check if the query can run
  //   if it cannot the query is sent to the GlobalQueryWaitQueue
  if ((e = initResourceAllocation(reqMsg, schedule, waitSchedule)) != OK) 
      return(NOTOK);

  return(OK);

}

/////////////////////////////////////////////////////////////////////////
//
//  set the maximum and minimum memory requirement for the query plan
//      pointed to by root. also calculates the number of buckets to be
//      used for every join. firstTime TRUE if query requirements being
//      found for the first time.
//
/////////////////////////////////////////////////////////////////////////
initQueryRequirements(QueryPlan* root, bool firstTime, bool staticSched)
{

    SelectPlan *selPlan = NULL;
    JoinPlan *joinPlan = NULL;
    int numInTuples, optMaxPagesNeeded, optMinPagesNeeded;
    int Memory, Buckets, e;

    // KLUDGE FOR SKEW
    bool found = FALSE;
    int tryNode;

    if (root != NULL)
    {
        switch(root->opCode)
	{
  	  case SELECT:
	      selPlan = (SelectPlan*) root;

	      // find the number of output tuples of the select
	      if (selPlan->index != NULL)
		  selPlan->numOutTuples = (int) ceil(selPlan->file->numTuples 
			   * selPlan->indexSelectivity * selPlan->selFactor);
	      else
		selPlan->numOutTuples = (int) ceil(selPlan->file->numTuples * 
						 selPlan->selFactor);

	      // *****KLUDGES(!)*****
	      // 1) the select and store operator do not reserve pages
	      //    while executing. so they take out an option to ensure
	      //    that they are guaranteed at least two pages at every site
	      //    so the total number of buffer pages is decremented by 2 
	      //     for each select or store
	      //
	      // 2) min requirement also set to 2 though they can work with 
	      //    just one page each. this causes complications in 
	      //    resourceAllocation where we need to find the memory to be 
	      //    allocated to the join


	      // set minimum and maximum memory requirements for the select
	      root->maxPagesNeeded = 2;
              root->minPagesNeeded = 2;

	      if (Flags[TRACE_SCHED])
		printf("SELECT maxPages %d minPages %d numOutTuples %d\n", 
		       root->maxPagesNeeded, root->minPagesNeeded, 
		       root->numOutTuples);

	      break;
          case TERMOP:
	      // process the child of the terminal operator
	      e = initQueryRequirements(((TermPlan*) root)->child,firstTime,
					staticSched);
              if (e != OK) return(NOTOK);

	      // find the number of output tuples of the operator
	      root->numOutTuples = ((TermPlan*) root)->child->numOutTuples;
	      root->maxPagesNeeded = 0;
	      root->minPagesNeeded = 0;

	      if (Flags[TRACE_SCHED])
		printf("TERMOP maxPages %d minPages %d, numOutTuples %d\n", 
		       root->maxPagesNeeded, root->minPagesNeeded, 
		       root->numOutTuples);
	      break;
    
  	  case STORE:
	      // process the child of the store operator
	      e = initQueryRequirements(((StorePlan*) root)->child,firstTime,
					staticSched);
              if (e != OK) return(NOTOK);
	      
	      // find the number of output tuples of the store
	      root->numOutTuples = ((StorePlan*) root)->child->numOutTuples;

              // see KLUDGE above
	      // set minimum and maximum memory requirements for the store
	      root->maxPagesNeeded = 2;
	      root->minPagesNeeded = 2;

	      if (Flags[TRACE_SCHED])
		printf("STORE maxPages %d minPages %d, numOutTuples %d\n", 
		       root->maxPagesNeeded, root->minPagesNeeded, 
		       root->numOutTuples);
	      break;

  	  case JOIN:
	      joinPlan = (JoinPlan*) root;


	      // call routine recursively for the children of the join node
              e = initQueryRequirements(joinPlan->leftChild, firstTime,
					staticSched);
              if (e != OK) return(NOTOK);

	      e = initQueryRequirements(joinPlan->rightChild, firstTime,
					staticSched);
              if (e != OK) return(NOTOK);

	      //set the number of input and output tuples of the join node
	      joinPlan->numOutTuples = 
		(int) floor(joinPlan->rightChild->numOutTuples * 
			    joinPlan->selFactor) ;
	      numInTuples = joinPlan->leftChild->numOutTuples ;

	      // find the total number of pages needed to fit all of the 
	      // build hash table in memory
	      optMaxPagesNeeded = 
		   MAX((int) ceil(1.0 * numInTuples * joinPlan->inTupSizeL * 
			   FUDGEFACTOR / (joinPlan->noActiveDisk * PageSize)),
		       3);

	      //set the optimally minimum number of pages needed
	      //  minpages = sqrt( |R| * F ) + 1
	      optMinPagesNeeded = (int) ceil(sqrt(optMaxPagesNeeded)) + 1;

              if (optMinPagesNeeded > BufSize)
	      {
		  printf("ERROR: Join min. memory (%d) > BufSize (%d),");
		  printf(" Query rejected\n", optMinPagesNeeded, BufSize);
		  exit(NOTOK);
//		  return(NOTOK);
	      }

              if (staticSched == FALSE) 
	      {
		
		 // min and max pages needed is set only the first time this 
		 //   procedure is called. the second time only the number of 
		 //   buckets is calculated
		 if (firstTime)
		 {


		     // KLUDGE FOR SKEW
		     //////////////////////////////////////////
		     // if skew present choose a random node from the nodes
		     // the join operator runs on. set the skew node parameter
	             // accordingly
		     // 
		     if (SkewFactor != 1)
		     {
		       joinPlan->skewNode = 0;
		       joinPlan->skewEntry = -1;
		       while( !(found) )
		       {
			   tryNode = random(0, MAX_NODES - 1);
			   if (joinPlan->activeSites[tryNode])
			   {
			     joinPlan->skewNode = tryNode;

			     // find the number of this node among all the 
			     // active nodes e.g. if join runs on sites 
			     // 25..30, site chosen is 26, then skew node 
			     // will be set to 2 as 26 is the second active 
			     // join node
			     for(int nodeNum = 0; nodeNum < tryNode; nodeNum++)
			       if (joinPlan->activeSites[nodeNum])
				 joinPlan->skewEntry++;
			     if (Flags[TRACE_SCHED])
			       printf("skewNode found %d", joinPlan->skewNode);
			     found = TRUE;
			   }
		       }
		     }
		     ///////////////////////////////////////////


		     //
		     // set the min mem needed by the join
		     //
		     if (joinPlan->minMemUnits == MINREQ_UNITS)
		         joinPlan->minPagesNeeded = 
			     (int) ceil(optMinPagesNeeded * 
					joinPlan->minMemNeeded);
		     else if (joinPlan->minMemUnits == KBYTE_UNITS)
		         joinPlan->minPagesNeeded = 
			     (int) ceil(joinPlan->minMemNeeded * 1024 / 
					PageSize);
		     else // BUFPOOL_UNITS
		         joinPlan->minPagesNeeded = 
			     (int) ceil(BufSize * joinPlan->minMemNeeded);

		     //
		     // set the max mem needed by the join
		     //
		     if (joinPlan->maxMemUnits == MINREQ_UNITS)
			 joinPlan->maxPagesNeeded = 
			   (int) ceil(optMinPagesNeeded * 
				      joinPlan->maxMemNeeded);
		     else if (joinPlan->maxMemUnits == KBYTE_UNITS)
		         joinPlan->maxPagesNeeded = 
			     (int) ceil(joinPlan->maxMemNeeded * 1024 / 
					PageSize);
		     else // BUFPOOL_UNITS
		         joinPlan->maxPagesNeeded = 
			     (int) ceil(BufSize * joinPlan->maxMemNeeded);
                                                                               
		     if (joinPlan->maxPagesNeeded > optMaxPagesNeeded)
		         joinPlan->maxPagesNeeded = optMaxPagesNeeded;

		     if (joinPlan->maxPagesNeeded > BufSize)
		         joinPlan->maxPagesNeeded = BufSize;

		     if (joinPlan->maxPagesNeeded < optMinPagesNeeded)
		     {
		         printf("ERROR: max mem (%d) < min needed (%d)\n",
				joinPlan->maxPagesNeeded, optMinPagesNeeded);
			 printf("Query rejected");
			 exit(NOTOK);
		     }

		     if (joinPlan->minPagesNeeded < optMinPagesNeeded)
		     {
			 printf("\nLess min. memory (%d) than needed (%d).", 
				joinPlan->minPagesNeeded, optMinPagesNeeded);
			 printf("set to min needed\n\n");
			 joinPlan->minPagesNeeded = optMinPagesNeeded;
		     }
		     else if (joinPlan->minPagesNeeded < 3)
		     {
			 printf("\nLess memory (%d) than needed (%d).", 
				joinPlan->minPagesNeeded, optMinPagesNeeded);
			 printf(" Join needs at least 3 pages\n\n");
			 exit(NOTOK);
		     }
		     else 
		       if (joinPlan->minPagesNeeded > joinPlan->maxPagesNeeded)
		       {
			   printf("\nMore min memory given (%d) than max (%d)",
			   joinPlan->minPagesNeeded, joinPlan->maxPagesNeeded);
			   printf(" => min set to (%d)\n", 
				  joinPlan->maxPagesNeeded );
			   joinPlan->minPagesNeeded = joinPlan->maxPagesNeeded;
		       }
		 }

	         //
	         //set the number of buckets to be used for this join
	         //
	         //  if optimal pages > max pages allocated
	         //        multi bucket join so calculate the number of buckets
	         //  else (* all hash table fits in memory*)
	         //       so number of bucket is 1
	         //
	         if (optMaxPagesNeeded > joinPlan->maxPagesNeeded)
		   //
		   // for a multi-bucket join, reserve 1 extra page for
		   //   each extra bucket
		   if (joinPlan->maxPagesNeeded > 3)
		   {

		     Memory  = joinPlan->maxPagesNeeded;

		     //       |R| * F - M
		     //  B = ---------------
		     //          M -  1
		     Buckets = (int) ceil((1.0 * 
			(optMaxPagesNeeded - Memory)) / (1.0 * ( Memory - 1)));

                     // formula assumes bucket 0 to bucket B
		     //    here we start from 1 onwards
                     joinPlan->numBuckets = 1 + Buckets;
		   
		     joinPlan->bucketKMemory = 
		        (int) ceil((1.0 * optMaxPagesNeeded - Memory + 
				    Buckets) / (1.0 * Buckets));

		     // weight factor = extra entries to be added to split tbl
		     //                 to send more tuples to bucket 0
		     //  = (bucket1 mem / mem for other buckets)-1
		     //           (M - B)
                     //  = ---------------------    - 1
                     //    ((|R| * F - (M - B))/B)
		     //
		     joinPlan->weightFactor = 
		      ((1.0 * (Memory - Buckets) * Buckets)
		       / (1.0 * optMaxPagesNeeded - Memory + Buckets)) - 1;

		     // safeguard against any strange floating point
		     //   calculation error
		     if (joinPlan->weightFactor < 0.0)
		         joinPlan->weightFactor = 0.0;

		     if (Flags[TRACE_SCHED])
		       printf("weightFactor %f opt %d memory %d bucketk %d buckets %d\n", 
			    joinPlan->weightFactor, optMaxPagesNeeded, Memory, 
			    joinPlan->bucketKMemory, Buckets);

		   }
	           else
		   {
		       printf("\n\nERROR:multi-bucket join, at least 3 pages");
		       printf("required in memory\n");
		       joinPlan->print(5);
		       exit(NOTOK);
//		       return(NOTOK);
		   }
		     
	         else
		   joinPlan->numBuckets = 1;
	      }
              else // staticSched == TRUE
	      {
		  joinPlan->maxPagesNeeded = joinPlan->minPagesNeeded = 
		      optMaxPagesNeeded;
		  if (joinPlan->maxPagesNeeded > BufSize)
		  {
		      printf("static sched: more max mem needed (%d) than buffer size (%d)",
			     joinPlan->maxPagesNeeded, BufSize);
		      exit(NOTOK);
//		      return(NOTOK);
		  }
		  joinPlan->numBuckets = 1;
	      }

	      if (Flags[TRACE_SCHED])
	      {
 		  printf("JOIN maxPgs %d minPgs %d numOutTups %d numBkts %d\n" 
			 ,root->maxPagesNeeded, root->minPagesNeeded, 
			 root->numOutTuples, joinPlan->numBuckets);
	      }
              break;

	  default:
	      printf("\n\nERROR: unknown operator type %d\n", root->opCode);
              exit(NOTOK);
//	      return(NOTOK);
	}
    }
    return(OK);
}






//calculate the total resources required to run the query. flag an error if 
//    more than buffer size required
initResourceAllocation(QueryMsg* reqMsg, Schedule* schedule, 
		       Schedule* waitSchedule)
{
  int 
      // total no. of minimum pages for operators on one site per scheduler row
      minPagesNeeded[MAX_NODES],
      
      // total no. of maximum pages for operators on one site per scheduler row
      maxPagesNeeded[MAX_NODES],
      nodeNo;

  Schedule *scheduleRow, *waitScheduleRow;
  Operator *opr;

    // initialize the totalMinPagesNeeded and totalMaxPagesNeeded arrays to 0
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    {
          reqMsg->query->totalMinPagesNeeded[nodeNo] = 0;
	  reqMsg->query->totalMaxPagesNeeded[nodeNo] = 0;
    }

    // find the maximum and minimum pages needed per node to run the query
    //   to find this both schedule and waiting schedule are used. schedule
    //   is used to find the new operators being launched. there memory is 
    //   added to the total memory needed. waiting schedule tells us what 
    //   operators are completing and so there memory is removed from the 
    //   total requirement
   
    // initialize the minPagesNeeded and maxPagesNeeded arrays to 0
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
            minPagesNeeded[nodeNo] = maxPagesNeeded[nodeNo] = 0;

    // for each row in the schedule find the max memory needed for each node
    for(scheduleRow = schedule, waitScheduleRow = waitSchedule; 
	scheduleRow != NULL, waitScheduleRow != NULL; 
	scheduleRow = scheduleRow->nextRow, 
	waitScheduleRow = waitScheduleRow->nextRow)
    {
	//
	// find the total memory needed per site for this scheduler row 
	//
        for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	    switch(opr->oprPlan->opCode)
	    {
		case SELECT:
	            for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		        if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
			    minPagesNeeded[nodeNo] += 
                                                opr->oprPlan->minPagesNeeded;
			    maxPagesNeeded[nodeNo] += 
                                                opr->oprPlan->maxPagesNeeded;
			}
		    break;
		case STORE:
		    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		        if (opr->oprPlan->activeSites[nodeNo] == TRUE) {
			    minPagesNeeded[nodeNo] += 
			                        opr->oprPlan->minPagesNeeded;
			    maxPagesNeeded[nodeNo] += 
                                                opr->oprPlan->maxPagesNeeded;
			}
		    break;
		case JOIN:
		    // add join's memory if the operator seen for the first 
		    // time i.e. for the build phase of the first bucket
		    if ((opr->phase == BUILD) && (opr->bucketNum = 1))
		      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			{
			  if ((reqMsg->query->schedType != RIGHTDEEPSTATIC1) &&
			      (reqMsg->query->schedType != RIGHTDEEPSTATICa))
			  {
			        minPagesNeeded[nodeNo] += 
				                 opr->oprPlan->minPagesNeeded;
			  }
			  else
			      minPagesNeeded[nodeNo] = 
				   MAX(minPagesNeeded[nodeNo],
				       opr->oprPlan->maxPagesNeeded);
				 
			      maxPagesNeeded[nodeNo] += 
				                 opr->oprPlan->maxPagesNeeded;
			}
		    break;
	      case TERMOP:
		    break;
	    }
       
	//
	// if for any site maxPagesNeeded(minPagesNeeded) for this row is greater
        //     than the maximum(minimum) values seen so far in the schedule
        //     reset totalMaxPagesNeeded(totalMinPagesNeeded)
	//
      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
      {
       if (minPagesNeeded[nodeNo] > reqMsg->query->totalMinPagesNeeded[nodeNo])
	  reqMsg->query->totalMinPagesNeeded[nodeNo] = minPagesNeeded[nodeNo];
       if (maxPagesNeeded[nodeNo] > reqMsg->query->totalMaxPagesNeeded[nodeNo])
	  reqMsg->query->totalMaxPagesNeeded[nodeNo] = maxPagesNeeded[nodeNo];
      }

	//
	// if the operator is to complete remove the memory from minPagesNeeded
	//   and maxPagesNeeded.
	// if the operator is not completing now, it's memory should stay as a
        //     part of the aggregate memory requirement per node
        //    
        for(opr = waitScheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	    switch(opr->oprPlan->opCode)
	    {
		case SELECT:
		      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		        if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			{
			     minPagesNeeded[nodeNo] -= 
                                               opr->oprPlan->minPagesNeeded;
			     maxPagesNeeded[nodeNo] -= 
                                               opr->oprPlan->maxPagesNeeded;
			}
	            break;
		case STORE:
		        for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		          if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			  {
			       minPagesNeeded[nodeNo] -= 
				               opr->oprPlan->minPagesNeeded;
			       maxPagesNeeded[nodeNo] -= 
                                               opr->oprPlan->maxPagesNeeded;
			  }
		    break;
		case JOIN:
		    // add join's memory if the operator seen for the last time
		    //    i.e. for the probe phase of the last bucket
		    if ((opr->phase == PROBE) && 
                        (opr->bucketNum == 
                                 ((JoinPlan*)opr->oprPlan)->numBuckets))
		       for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			 if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			 {
 			  if ((reqMsg->query->schedType != RIGHTDEEPSTATIC1) &&
			      (reqMsg->query->schedType != RIGHTDEEPSTATICa))

			        minPagesNeeded[nodeNo] -= 
				                 opr->oprPlan->minPagesNeeded;

			      maxPagesNeeded[nodeNo] -= 
				                 opr->oprPlan->maxPagesNeeded;
			 }
		    break;
                case TERMOP:
                    break;
	    }
    }

    //
    // if the minimum memory requirement of the query is greater than the 
    //     total buffer size then discard the query
    //
    if ((reqMsg->query->schedType != RIGHTDEEPSTATIC1) &&
        (reqMsg->query->schedType != RIGHTDEEPSTATICa))
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
        if (reqMsg->query->totalMinPagesNeeded[nodeNo] > BufSize)
	{
	    printf("\n\nMin. memory requirement of query (%d) > total buffer",
                   reqMsg->query->totalMinPagesNeeded[nodeNo]);
            printf("size (%d) for site %d\n\n", BufSize, nodeNo);
            exit(NOTOK);
//	    return(NOTOK);
	}

    return(OK);  

}


//////////////////////////////////////////////////////////////////////////
//
// check if the query can be executed given the present BufPages
//
//////////////////////////////////////////////////////////////////////////
CanRun(QueryPlan* query)
{
    int nodeNo;

    // for each site
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
        //
        // if available buffers less than minimum needed
	//        send the query to the GlobalQueryWaitQueue
	//
        if (BufPages[nodeNo] < query->totalMinPagesNeeded[nodeNo])
        {
             if (Flags[TRACE_SCHED])
	     {
                 printf("Not enough pages on site %d. needed %d, avail %d\n", 
			nodeNo, query->totalMinPagesNeeded[nodeNo], 
			BufPages[nodeNo]);
	     }
	     return(NOTOK);

        }

    return(OK);
}





///////////////////////////////////////////////////////////////////////////////
//
// perform actual resource allocation for the query plan using the available 
//     buffers at different sites. 
//     schedule and waitSchedule point to the schedules created using 
//     the optimal memory requirement
//
//     if the available buffers are less than the minimum needed
//         query is sent to the GlobalQueryWaitQueue
//     else
//         if the available buffers are less than the maximum needed
//               query is rescheduled
//         else
//               dispatched for execution as is
//     
//////////////////////////////////////////////////////////////////////////////
actualResourceAllocation(QueryMsg* reqMsg, Schedule*& schedule, 
			 Schedule*& waitSchedule)
{

    int
	rowPagesAlloc[MAX_NODES],     // pages allocated to the query per site
        totalMinJoinPages[MAX_NODES], // total no. of pages needed for 
	                              // concurrent joins per site
	copyMinJoinPages[MAX_NODES],  // scratch array
        numJoins, nodeNo, reschedule, joinMemory, e, joinNum ;

    Schedule *scheduleRow, *waitScheduleRow;
    Operator *opr;

    QueryPlan *concJoins[MAXJOINS], *qPlan;

    // for each site
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
    {
	//
	// if the available pages less than the maximum pages needed
	//     set pagesAllocated accordingly and mark query for rescheduling
	//   NOTE: as memory available has changed the num of buckets for the
	//          joins may also have changed
	//
	if (BufPages[nodeNo] < reqMsg->query->totalMaxPagesNeeded[nodeNo])
	{

             if (Flags[TRACE_SCHED])
	         printf("Less than max. pgs on site %d. needed %d, avail %d\n",
			nodeNo, reqMsg->query->totalMaxPagesNeeded[nodeNo], 
			BufPages[nodeNo]);

	     // note the buf pages allocated to the site for this site
	     reqMsg->query->totalMaxPagesNeeded[nodeNo] = BufPages[nodeNo];

	     // mark the query for rescheduling
	     reschedule = TRUE;
	 }
    }

    if (Flags[TRACE_SCHED])
    {
        printf("\n\tBufferPages left and allocated to query for each site\n");
        for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	   printf("\t\tsite %d BufPages %d allocated %d\n", nodeNo, 
		BufPages[nodeNo], reqMsg->query->totalMaxPagesNeeded[nodeNo]);
    }

    // if query marked for rescheduling
    if (reschedule == TRUE)
    {
      if ((reqMsg->query->schedType != RIGHTDEEPSTATIC1) &&
          (reqMsg->query->schedType != RIGHTDEEPSTATICa))
      {
        if (Flags[TRACE_SCHED])
             printf("\nQuery being rescheduled\n\n");

	numJoins = 0 ;
	for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	      totalMinJoinPages[nodeNo] = 0;

        // for every scheduler row
        for(scheduleRow = schedule, waitScheduleRow = waitSchedule; 
	    scheduleRow != NULL, waitScheduleRow != NULL; 
	    scheduleRow = scheduleRow->nextRow, 
	    waitScheduleRow = waitScheduleRow->nextRow)
        {
	    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	        // need to copy to another array because may change in 
		// the code below
	        rowPagesAlloc[nodeNo] = 
		      reqMsg->query->totalMaxPagesNeeded[nodeNo];

	    // find the number of join operators in the row and take out 
	    // the pages needed for select and store from the pages available 
	    // per site
            for(opr = scheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	        switch(opr->oprPlan->opCode)
		{
		    case SELECT:
                       for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			   if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			       rowPagesAlloc[nodeNo] -= opr->oprPlan->minPagesNeeded;
		       break;
		    case STORE:
                       for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			   if (opr->oprPlan->activeSites[nodeNo] == TRUE)
		               rowPagesAlloc[nodeNo] -= opr->oprPlan->minPagesNeeded;
		       break;
		    case JOIN:
		       // if join seen for the first time
		       if ((opr->phase == BUILD) && (opr->bucketNum = 1))
		       {
			  // add join to list of concurrently running joins
			  concJoins[++numJoins] = opr->oprPlan;

			 //
			 // for every site on which the join runs, add 
			 //    min memory needed to total min mem for joins
			 //
                         for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			     if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			         
			           totalMinJoinPages[nodeNo] += 
				               opr->oprPlan->minPagesNeeded;
		       }
		       break;
		}

	    // KLUDGE*******
	    //  divide the total memory in the row among the joins proportional
	    //  to the minimum memory requirement of the join
	    //  this is an arbitrary choice. other choices include size of the
	    //  join, size of the result, size of inner relation etc.
	    //
	    for(joinNum = 1; joinNum <= numJoins; joinNum++)
	    {
	        qPlan = concJoins[joinNum];
		for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		    if (qPlan->activeSites[nodeNo] == TRUE)
		    {
		       joinMemory =  
			  (int) (((1.0 * qPlan->minPagesNeeded) / 
			   totalMinJoinPages[nodeNo]) * rowPagesAlloc[nodeNo]);
		       
		       // if the memory given < maximum needed
		       //     change the allocated memory to the new value
		       if (joinMemory < qPlan->maxPagesNeeded)
			     qPlan->maxPagesNeeded = joinMemory;
		    }  

		if (joinMemory < qPlan->minPagesNeeded)
		{
		    printf("\n\nERROR:join mem (%d) < min req (%d) \n", 
			   joinMemory, qPlan->minPagesNeeded);
		    exit(NOTOK);
		}
 	    }

	    // if a join completes then change list of concurrent joins and
	    //  total min. join mem
            for(opr = waitScheduleRow->oprRow; opr != NULL; opr = opr->nextOpr)
	       if ((opr->oprPlan->opCode == JOIN) && ((opr->phase == PROBE) && 
                   (opr->bucketNum == ((JoinPlan*)opr->oprPlan)->numBuckets)))
	       {
		   // find the join in the concJoin array
		   for(joinNum = 1; joinNum <= numJoins; joinNum++)
		      if (concJoins[joinNum] == opr->oprPlan)
			  break;
		   // shift the subsequent join plans once to the left
		   for (;joinNum <= (numJoins - 1); joinNum++)
		      concJoins[joinNum] = concJoins[joinNum + 1];
		   //reduce the number of joins
		   numJoins--;

		   // reset the total join minimum page requirement
                   for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
			if (opr->oprPlan->activeSites[nodeNo] == TRUE)
			     totalMinJoinPages[nodeNo] -= 
			                opr->oprPlan->minPagesNeeded;
	       }
	}
	
	// perform the calculation of the memory required by the query again
	//   due to overlap the memory allocated to joins may have changed thus
	//   changing the memory requirement of the query
	if ((e = initResourceAllocation(reqMsg,schedule, waitSchedule)) != OK) 
	  return(NOTOK);
	
	// recalculate only the number of buckets for the join
	if ((reqMsg->query->schedType == RIGHTDEEPSTATIC1) ||
	    (reqMsg->query->schedType == RIGHTDEEPSTATICa))
	    e = initQueryRequirements(reqMsg->query, FALSE, TRUE);
	else
	    e = initQueryRequirements(reqMsg->query, FALSE);
	if (e  != OK) return(NOTOK);
      }

	// make a new schedule given the new number of buckets for the joins
        // destroy the old schedules and form the new schedules
	DeleteSched(schedule);
	DeleteSched(waitSchedule);
	schedule = NULL;
	waitSchedule = NULL;
        if ((e = makeSchedule(reqMsg->query, schedule, waitSchedule, FALSE)) != OK) 
	     return(NOTOK);
    }

    // for each site reduce the number of free buffer pages by the amount 
    //    allocated to the query
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
         BufPages[nodeNo] -= reqMsg->query->totalMaxPagesNeeded[nodeNo];

    return(OK);
}


////////////////////////////////////////////////////////////////////////////
//
// form the schedule and the wait schedule given the root of the query plan
//
////////////////////////////////////////////////////////////////////////////
makeSchedule(QueryPlan* root, Schedule*& schedule, Schedule*& waitSchedule,
	     bool flag)
{

    Schedule   *lastWaitRow = NULL, *insertRow = NULL, *insertWaitRow = NULL;
    Schedule   *lastRow = NULL;
    int nodeNo, allocMem[MAX_NODES], joinMem[MAX_NODES];

    // call the appropriate scheduler depending on scheduler type
    switch(root->schedType)
    {
      case RIGHTDEEPSTATIC1:
      case RIGHTDEEPSTATICa:
          if (flag)
	      bushyExec(root,schedule, waitSchedule, insertRow, insertWaitRow, 
			lastRow, lastWaitRow, TRUE);
	  else
	  {
	      // copy total memory requirement
	      // schedule the query by calling the static scheduler
	      // reduce memory by 2 to account for the stores
	      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		  allocMem[nodeNo] = root->totalMaxPagesNeeded[nodeNo] - 2;

	      for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		printf("\tsite %d start %d\n", nodeNo,
		       root->totalMaxPagesNeeded[nodeNo]);

	      if (root->schedType == RIGHTDEEPSTATIC1)
	 	  staticExec(root, schedule, waitSchedule, insertRow, 
			     insertWaitRow, lastRow, lastWaitRow, TRUE,
			     TRUE, allocMem, root->totalMaxPagesNeeded);
	      else
	 	  staticExec(root, schedule, waitSchedule, insertRow, 
			     insertWaitRow, lastRow, lastWaitRow, FALSE,
			     TRUE, allocMem, root->totalMaxPagesNeeded);
	  }
	  break;

      case RIGHTDEEPONE:
          // schedule the query by calling bushyExec
	  bushyExec(root, schedule, waitSchedule, insertRow, insertWaitRow, 
		    lastRow, lastWaitRow, TRUE);
	  break;

      case LEFTDEEP:
      case RIGHTDEEPALL:
      case BUSHYMAX:
          // schedule the query by calling bushyExec 
	  bushyExec(root, schedule, waitSchedule, insertRow, insertWaitRow, 
		    lastRow, lastWaitRow, FALSE);
	  break;
      default:
	  printf("\nERROR: unknown sched type %d\n", root->schedType);
	  exit(NOTOK);
//	  return(NOTOK);
    }
    if (Flags[TRACE_SCHED])
    {
      printf("Sending Schedule\n\n");
      schedule->print();
      printf("\nWaiting Schedule\n\n");
      waitSchedule->print();
      printf("\n");
    }
    return(OK);

}


//////////////////////////////////////////////////////////////////
//
//  addOpr
//    In - 1) operator - pointer to a schedule row (dependency group)
//                     changed by the procedure
//         2) opr - pointer to operator to be added
//    Action - adds the operator to the schedule row, the first
//             parameter is changed by this action
//
//////////////////////////////////////////////////////////////////
void addOpr(Operator*& oprRow, QueryPlan* opr, int bucketNum, int phase)
{
  Operator* newOperator;

  newOperator = new Operator();
  newOperator->oprPlan = opr;
  newOperator->nextOpr = oprRow;
  newOperator->bucketNum = bucketNum;
  if (phase != -1)
    newOperator->phase = phase;
  oprRow = newOperator;
}


////////////////////////////////////////////////////////////////////////////
//
//  addLastRow
//    adds the currentNode to the lastRow in the schedule
//    marks operator as processing 'bucketNum' bucket
//    if lastRow = NULL, adds a new last row
//
////////////////////////////////////////////////////////////////////////////
void addLastRow(Schedule*& schedule, Schedule*& lastRow, 
		QueryPlan* currentNode, int bucketNum)
{

  Schedule* newRow;

  // add new last row;
  if (lastRow == NULL)
  {
    lastRow  = new Schedule();
    schedule = lastRow;
  }
  else // add currentNode to the last row
  {
    newRow = new Schedule();
    lastRow->nextRow = newRow;
    lastRow = newRow;
  }
  // add node to last row;
  addOpr(lastRow->oprRow, currentNode, bucketNum);

  return;
}



////////////////////////////////////////////////////////////////////////
//
//  ldExec
//
//    In -  currentNode - pointer to the root of the query tree
//          schedule - pointer to the current schedule
//          lastRow - pointer to the last row in the schedule
//          waitSchedule - pointer to the waiting schedule
//          lastWaitRow - pointer to the last row in the waitSchedule
//          insertRow - row where operator to be added in schedule
//
//    Action - a recursive routine which traverses the query tree and
//             creates a left deep execution schedule
//
////////////////////////////////////////////////////////////////////////
void ldExec(QueryPlan* currentNode, 
	    Schedule*& schedule, Schedule*& lastRow,
	    Schedule*& waitSchedule, Schedule*& lastWaitRow,
	    Schedule*& insertRow)
{

  JoinPlan* joinNode;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:

      // add to last row in schedule
      addLastRow(schedule, lastRow, currentNode);

      // add to last row in wait schedule
      addLastRow(waitSchedule, lastWaitRow, currentNode);

      insertRow = lastRow ;

      break;

    case STORE:
      // process the child
      ldExec(((StorePlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;


    case TERMOP:
      // process the child
      ldExec(((TermPlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;

    case JOIN:

      joinNode = (JoinPlan*) currentNode;

      // process the left child of join
      ldExec(joinNode->leftChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow);

      // add node to last Row in waitSchedule
      addOpr(lastWaitRow->oprRow, currentNode, 1, BUILD);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, BUILD);

      // process right child of join
      ldExec(joinNode->rightChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow);

      // add node to last row in waitSchedule
      addOpr(lastWaitRow->oprRow, currentNode, 1, PROBE);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, PROBE);

      break;
  }

  return;

}






/////////////////////////////////////////////////////////////////////////////
//
//  rdExec
//    In -  currentNode - pointer to the root of the query tree
//          schedule - pointer to the current schedule
//          lastRow - pointer to the last row in the schedule
//          waitSchedule - pointer to the waiting schedule
//          lastWaitRow - pointer to the last row in the waitSchedule
//          insertRow - row where operator to be added in schedule
//          insertWaitRow - row where operator to be added in waitSchedule
//
//         
//    Action - a recursive routine which traverses the query tree and
//             creates a right deep execution schedule
//             both schedule and lastRow are changed
//
//    the resulting schedule carries out the hash table builds from the
//       base relation sequentially. 
//       different form rdExecAll which carries out all of them in parallel
//
//////////////////////////////////////////////////////////////////////////////
void rdExec(QueryPlan* currentNode, 
	    Schedule*& schedule, Schedule*& lastRow,
	    Schedule*& waitSchedule, Schedule*& lastWaitRow,
	    Schedule*& insertRow, Schedule*& insertWaitRow)
{

  Schedule* newRow;
  QueryPlan* parentNode;
  JoinPlan* joinNode;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:
      // if left child of join then
      if (currentNode->parent->opCode == JOIN)
      {
	parentNode = currentNode->parent;
	if (((JoinPlan*) parentNode)->leftChild == currentNode)
	{
	  // add new first row to schedule
	  newRow = new Schedule();
	  newRow->nextRow = schedule;
	  schedule = newRow;
	  if (lastRow == NULL) lastRow = newRow;

	  // add node to first row in schedule
	  addOpr(schedule->oprRow, currentNode);

	  // add new first row to waitSchedule
	  newRow = new Schedule();
	  newRow->nextRow = waitSchedule;
	  waitSchedule = newRow;
	  if (lastWaitRow == NULL) lastWaitRow = newRow;

	  // add node to first row in wait schedule
	  addOpr(waitSchedule->oprRow, currentNode);

	  //
	  //set the insert row (insertWaitRow) to be the lastRow (lastWaitRow)
	  //    to insert the parent of the select operator here 
	  //
	  insertRow = schedule;
	  insertWaitRow = waitSchedule;
	}
	else
	  // if right child of join then
	    if (((JoinPlan*) parentNode)->rightChild == currentNode)
	    {
	      // add opr to last row in schedule
	      addLastRow(schedule, lastRow, currentNode);

	      // add opr to last row in wait schedule
	      addLastRow(waitSchedule, lastWaitRow, currentNode);

	      //set the insert row (insertWaitRow) to be the lastRow 
	      //    (lastWaitRow) to insert the parent of the select 
	      //    operator here 

	      insertRow = lastRow;
	      insertWaitRow = lastWaitRow;
	    }
	    else
	      // error
		printf("Unexpected Select operator in RD tree");
      }
      break;

    case STORE:

      // process the child
      rdExec(((StorePlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;

    case TERMOP:

      // process the child
      rdExec(((TermPlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;

    case JOIN:

      joinNode = (JoinPlan*) currentNode;

      // process the left child of join
      rdExec(((JoinPlan*) currentNode)->leftChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node to insert row in waitSchedule
      addOpr(insertWaitRow->oprRow, currentNode, 1, BUILD);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, BUILD);
      
      // process right child of join
      rdExec(((JoinPlan*) currentNode)->rightChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node to insert row in waitSchedule
      addOpr(insertWaitRow->oprRow, currentNode, 1, PROBE);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, PROBE);

      //set the last wait row to be the insertwait row to insert the
      //    the parent of the join operator here (after all the buckets
      //    have been processed)
      insertWaitRow = lastWaitRow;

      break;
  }
  return;

}




//////////////////////////////////////////////////////////////////////////////
//
//  rdExecAll
//    In -  currentNode - pointer to the root of the query tree
//          schedule - pointer to the current schedule
//          lastRow - pointer to the last row in the schedule
//          waitSchedule - pointer to the waiting schedule
//          lastWaitRow - pointer to the last row in the waitSchedule
//          insertRow - row where operator to be added in schedule
//          insertWaitRow - row where oprerator to be added in waitSchedule
//
//
//    Action - a recursive routine which traverses the query tree and
//             creates a right deep execution schedule
//             here all the selects are started in parallel
//             both schedule and lastRow are changed
//
//
//    the resulting schedule carries out the hash table builds from all the
//       base relation in parallel.
//       different form rdExec which carries them out sequentially
//
//    does not work with multiple buckets right now
//
////////////////////////////////////////////////////////////////////////////
void rdExecAll(QueryPlan* currentNode, 
	    Schedule*& schedule, Schedule*& lastRow,
	    Schedule*& waitSchedule, Schedule*& lastWaitRow,
	    Schedule*& insertRow, Schedule*& insertWaitRow)
{

  Schedule* newRow;
  QueryPlan* parentNode;
  JoinPlan* joinNode;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:
      // if left child of join then
      if (currentNode->parent->opCode == JOIN)
      {
	parentNode = currentNode->parent;
	if (((JoinPlan*) parentNode)->leftChild == currentNode)
	{
	  if (schedule == NULL)
	  {
	    schedule = new Schedule();
	    lastRow = schedule;
	  }
	  // add node to first row
	  addOpr(schedule->oprRow, currentNode);

	  if (waitSchedule == NULL)
	  {
	    waitSchedule = new Schedule();
	    lastWaitRow = waitSchedule;
	  }
	  // add node to first row
	  addOpr(waitSchedule->oprRow, currentNode);

	  insertRow = schedule;
	  insertWaitRow = waitSchedule;

	}
	else
	  // if right child of join then
	    if (((JoinPlan*) parentNode)->rightChild == currentNode)
	    {
	      newRow = new Schedule();
	      lastRow->nextRow = newRow;
	      lastRow = newRow;

	      // add node to lastRow
	      addOpr(lastRow->oprRow, currentNode);

	      newRow = new Schedule();
	      lastWaitRow->nextRow = newRow;
	      lastWaitRow = newRow;

	      // add node to lastWaitRow
	      addOpr(lastWaitRow->oprRow, currentNode);

	      insertRow = lastRow;
	      insertWaitRow = lastWaitRow;
	    }
	    else
	      // error
		printf("Unexpected Select operator in RD tree");
      }
      break;

    case STORE:
      // process the child
      rdExecAll(((StorePlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);
      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;

    case TERMOP:
      // process the child
      rdExecAll(((TermPlan*) currentNode)->child, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);
      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode);

      // add node in last row in wait Schedule
      addOpr(lastWaitRow->oprRow, currentNode);

      break;
    case JOIN:

      joinNode = (JoinPlan*) currentNode;

      // process the left child of join
      rdExecAll(joinNode->leftChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node to last row in waitSchedule
      addOpr(insertWaitRow->oprRow, currentNode, 1, BUILD);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, BUILD);

      
      // process right child of join
      rdExecAll(joinNode->rightChild, schedule, lastRow,
	     waitSchedule, lastWaitRow, insertRow, insertWaitRow);

      // add node to last row in waitSchedule
      addOpr(insertWaitRow->oprRow, currentNode, 1, PROBE);

      // add node in insert row in schedule
      addOpr(insertRow->oprRow, currentNode, 1, PROBE);

      //set the last wait row to be the insertwait row to insert the
      //    the parent of the join operator here (after all the buckets
      //    have been processed)
      insertWaitRow = lastWaitRow;

      break;
  }
  return;

}

// add a new first row to schedule
void addFirstRow(Schedule*& schedule, Schedule*& lastRow)
{
  Schedule *newRow = new Schedule();
  newRow->nextRow = schedule;
  schedule = newRow;
  if (lastRow == NULL)
     lastRow = schedule;
}

// add a new last row to the schedule
Schedule* addNewLastRow(Schedule*& schedule, Schedule*& lastRow)
{
    Schedule *newRow = new Schedule();
    newRow->nextRow = NULL;
    if (schedule == NULL)
      schedule = newRow;
    else
        lastRow->nextRow = newRow;
    lastRow = newRow;

    return(newRow);
}

//check if secondRow appears after firstRow in schedule
bool after(Schedule* firstRow, Schedule* secondRow, Schedule*& lastRow)
{
    if ((secondRow == NULL) || (firstRow == lastRow))
        return(FALSE);
    else
    {
        // check rows succeeding first row
        for(Schedule *row = firstRow->nextRow; row != NULL; row = row->nextRow)
	  if (row == secondRow)
	      return(TRUE);
	return(FALSE);
    }
}


// add opr to the schedule row. set bucketNum (default 1) 
//   and phase (ignored if opr not a join)
void addOpr2(Schedule* row, QueryPlan* opr, int bucketNum = 1, int phase = -1)
{
  Operator* newOperator;

  newOperator = new Operator();
  newOperator->oprPlan = opr;
  newOperator->nextOpr = row->oprRow;
  newOperator->bucketNum = bucketNum;
  if (opr->opCode == JOIN)
    newOperator->phase = phase;
  row->oprRow = newOperator;
}



///////////////////////////////////////////////////////////////////////////////
//
//  bushyExec
//   In - 1) currentNode - pointer to the root of the query tree
//        2) schedule - pointer to the schedule
//        3) waitSchedule - pointer to the waitSchedule
//        4) insertRow - ptr to row in schedule where op to be inserted
//        5) lastRow - last row in schedule
//        6) waitInsertRow - ptr to row in waitSchedule where op to be inserted
//        7) lastRow - last row in waitSchedule
//        8) flag - TRUE if right-deep-one schedule to be build (else FALSE)
//                right-deep-one sched -> one scan at a time
//                                        compared to all in parallel
//
//    Action - a recursive routine which traverses the query tree and
//             creates a bushy execution schedule, it generates a
//             bushy tree with the maximal parallelism possible.
//             as lef
//
//////////////////////////////////////////////////////////////////////////////
void bushyExec(QueryPlan* currentNode, Schedule*& schedule, 
	       Schedule*& waitSchedule, Schedule*& insertRow, 
	       Schedule*& lastRow, Schedule*& waitInsertRow, 
	       Schedule*& lastWaitRow, bool flag)
{

  JoinPlan *joinNode;
  Schedule *origInsertRow, *origWaitInsertRow;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:
      // if right child of join
      //    add to new last row
      if ((currentNode->parent->opCode == JOIN) &&
  	  (((JoinPlan*) currentNode->parent)->rightChild == currentNode))
      {
	   insertRow = addNewLastRow(schedule, lastRow);
	   waitInsertRow = addNewLastRow(waitSchedule, lastWaitRow);
      }
      else  // left child of join or child of store/termop
      {
	  // if right-deep-one schedule form new row for select
	  if (flag)
	  {
	      addFirstRow(schedule, lastRow);
	      addFirstRow(waitSchedule, lastWaitRow);
	  }
	  else // add to first row
	  {
	      if (schedule == NULL)
		  addFirstRow(schedule, lastRow);
	      if (waitSchedule == NULL)
		  addFirstRow(waitSchedule, lastWaitRow);
	  }
	  insertRow = schedule;
	  waitInsertRow = waitSchedule;
      }
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case STORE:
      bushyExec(((StorePlan*) currentNode)->child, schedule, waitSchedule,
		insertRow, lastRow, waitInsertRow, lastWaitRow, flag);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case TERMOP:
      bushyExec(((TermPlan*) currentNode)->child, schedule, waitSchedule,
		insertRow, lastRow, waitInsertRow, lastWaitRow, flag);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;

    case JOIN:
      joinNode = (JoinPlan*) currentNode;

      // make a copy as level may change below
      origInsertRow = insertRow;
      origWaitInsertRow = waitInsertRow;

      // process the left child of join
      bushyExec(joinNode->leftChild, schedule, waitSchedule, insertRow, 
		lastRow, waitInsertRow, lastWaitRow, flag);

      // add join node (BUILD) to both schedules
      addOpr2(insertRow, joinNode, 1, BUILD);
      addOpr2(waitInsertRow, joinNode, 1, BUILD);

      // if new rows before original rows in schedule, restore rows
      if (after(insertRow, origInsertRow, lastRow )) 
	  insertRow = origInsertRow;
      if (after(waitInsertRow,origWaitInsertRow, lastWaitRow)) 
	  waitInsertRow = origWaitInsertRow;

      // process right child of join
      bushyExec(joinNode->rightChild, schedule, waitSchedule, insertRow, 
		lastRow, waitInsertRow, lastWaitRow, flag);

      // add join node (PROBE) to both schedules
      addOpr2(insertRow, joinNode, 1, PROBE);
      addOpr2(waitInsertRow, joinNode, 1, PROBE);

      waitInsertRow = lastWaitRow;

      break;
  }
}






//////////////////////////////////////////////////////////////////////////
//
// check if the parent and child can execute concurrently given total memory
//   oneScan is used to indicate if extra memory for select needs to be
//   added
//
//////////////////////////////////////////////////////////////////////////
bool canRunJoin(JoinPlan* parent, int totalMemory[MAX_NODES], bool oneScan)
{
    int nodeNo, memNeeded;

    // for each site
    for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
      if (parent->activeSites[nodeNo])
      {
	// add memory for associated select if more that one scan allowed
	memNeeded = parent->maxPagesNeeded + (!oneScan) * 2 ;
        if (memNeeded  > (totalMemory[nodeNo]))
        {
             if (Flags[TRACE_SCHED])
	     {
                 printf("Not enough pages on site %d. needed %d, avail %d\n", 
			nodeNo, memNeeded, totalMemory[nodeNo]);
	     }
	     return(FALSE);

        }
      }
    return(TRUE);
}


//////////////////////////////////////////////////////////////////////////////
//
//  createXmedSelect - make an intermediate select node to be linked to 
//    to the spool node created earlier
//
//////////////////////////////////////////////////////////////////////////////
SelectPlan* createXmedSelect(JoinPlan* joinNode, StorePlan* storeNode)
{
  SelectPlan *sp;
  int i;
  
  sp = new SelectPlan;
  sp->selectType = XMED;

  // connect to the parent join operator and the intermediate store
  sp->parent = joinNode;
  sp->jobClass = joinNode->jobClass;

  // all tuples qualify for the join so selectivity = 1
  sp->selFactor = 1;
  
  // input tuple size is same as output tuple size of associated spool node
  sp->inTupSize = storeNode->outTupSize;
  sp->outTupSize = sp->inTupSize;

  // Runs the operator only where the parent join node active
  sp->noActiveDisk = sp->parent->noActiveDisk;
  for (i = 0; i < NumDiskNodes; i++) 
    sp->activeSites[i] = sp->parent->activeSites[i] ;

  return(sp);
}


// make a local spool node - (used to store intermediate files in the
//   tree
StorePlan* createLocalSpool(JoinPlan* joinNode)
{
  StorePlan *st;
  int i;
  
  st = new StorePlan;
  st->storeType = XMED;

  // connect store node to the parent join node
  st->parent = joinNode;
  st->jobClass = joinNode->jobClass;

  // link spoolnode to join and link the right child of join to the
  // spool node (if operator to be generalized needs to be changed)
  st->child = joinNode->rightChild;
  joinNode->rightChild->parent = st;

  // initialize other fields of the store operator
  st->inTupSize = st->child->outTupSize;
  st->outTupSize = st->inTupSize;
  st->inStreams = st->child->noActiveDisk;
  
  // Runs the operator only where the parent join node active
  st->noActiveDisk = joinNode->noActiveDisk;
  for (i = 0; i < NumDiskNodes; i++) 
    st->activeSites[i] = joinNode->activeSites[i] ;
  return(st);
}




//Schedule the query using static right-deep scheduling
void staticExec(QueryPlan* currentNode, Schedule*& schedule, 
	       Schedule*& waitSchedule, Schedule*& insertRow, 
	       Schedule*& lastRow, Schedule*& waitInsertRow, 
	       Schedule*& lastWaitRow, bool oneScan, bool runJoinTogether, 
	       int memLeft[MAX_NODES], int totalMemory[MAX_NODES])
{

  SelectPlan *sp;
  StorePlan *stp;
  JoinPlan *joinNode;
  int nodeNo;
  Schedule *origInsertRow, *origWaitInsertRow;

  if (currentNode == NULL) return;

  switch(currentNode->opCode)
  {
    case SELECT:
        if ((currentNode->parent->opCode == JOIN) &&
	    (((JoinPlan*) currentNode->parent)->rightChild == currentNode))
        {
	    insertRow = addNewLastRow(schedule, lastRow);
	    waitInsertRow = addNewLastRow(waitSchedule, lastWaitRow);
        }
        else
	{
	    if (runJoinTogether)
	    {
 	        // if one-scan schedule add new first row
	        if (oneScan)
		{
		    addFirstRow(schedule, lastRow);
		    addFirstRow(waitSchedule, lastWaitRow);
		}
		else // add to first row
		{
		    if (schedule == NULL)
		        addFirstRow(schedule, lastRow);
		    if (waitSchedule == NULL)
		        addFirstRow(waitSchedule, lastWaitRow);

		    // as select to run together with join, reduce memory
		    // allocated to join by 2
		    for (nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
		        if (currentNode->activeSites[nodeNo])
			  memLeft[nodeNo] -= currentNode->maxPagesNeeded;
		}
		insertRow = schedule;
		waitInsertRow = waitSchedule;
	    }
	    else // join to run separate so add new last row
	    {
	        insertRow = addNewLastRow(schedule, lastRow);
		waitInsertRow = addNewLastRow(waitSchedule, lastWaitRow);
	    }
	}
	addOpr2(insertRow, currentNode);
	addOpr2(waitInsertRow, currentNode);
	break;
    case STORE:
      staticExec(((StorePlan*) currentNode)->child, schedule, waitSchedule,
		 insertRow, lastRow, waitInsertRow, lastWaitRow, oneScan,
		 runJoinTogether, memLeft, totalMemory);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;
    case TERMOP:
      staticExec(((TermPlan*) currentNode)->child, schedule, waitSchedule,
		 insertRow, lastRow, waitInsertRow, lastWaitRow, oneScan,
		 runJoinTogether, memLeft, totalMemory);
      addOpr2(insertRow, currentNode);
      addOpr2(waitInsertRow, currentNode);
      break;
    case JOIN:
      joinNode = (JoinPlan*) currentNode;

      // process right child of join
      staticExec(joinNode->rightChild, schedule, waitSchedule, insertRow, 
		 lastRow, waitInsertRow, lastWaitRow, oneScan, runJoinTogether,
		 memLeft, totalMemory);
      
      // make a copy as level may change below
      origInsertRow = insertRow;
      origWaitInsertRow = waitInsertRow;

      runJoinTogether = TRUE;

      // if right child join and both joins cannot run concurrently 
      if ((joinNode->rightChild->opCode == JOIN)
	  && ((runJoinTogether = canRunJoin(joinNode, memLeft, oneScan)) != 
	       TRUE))
      {
	   // reset the maximum pages available after removing the pages
	   //    needed for this join
	   for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	       if (joinNode->activeSites[nodeNo])
	           memLeft[nodeNo] = totalMemory[nodeNo] - 
		                                    joinNode->maxPagesNeeded;

	   // add intermediate spool node to both schedules
	   stp = createLocalSpool(joinNode);
	   addOpr2(insertRow, stp);
	   addOpr2(waitInsertRow, stp);

	   // process left child of join
	   staticExec(joinNode->leftChild, schedule, waitSchedule, 
		      insertRow, lastRow, waitInsertRow, lastWaitRow, oneScan,
		      FALSE, memLeft, totalMemory);

	   // add join node (BUILD) to both schedules
	   addOpr2(insertRow, joinNode, 1, BUILD);
	   addOpr2(waitInsertRow, joinNode, 1, BUILD);
	     
	   // add intermediate select node to both schedules
	   sp = createXmedSelect(joinNode, stp);
	   insertRow = addNewLastRow(schedule, lastRow);
	   waitInsertRow = addNewLastRow(waitSchedule, lastWaitRow);
	   addOpr2(insertRow, sp);
	   addOpr2(waitInsertRow, sp);
      }
      else // right child not join OR joins cannot run together
      {
	   // reduce the memory left for the other joins
	   for(nodeNo = 0; nodeNo < NumDiskNodes; nodeNo++)
	   {
	       if (joinNode->activeSites[nodeNo])
		   memLeft[nodeNo] -= joinNode->maxPagesNeeded;
	       printf("\tsite %d memLeft %d\n", nodeNo, memLeft[nodeNo]);
	   }

	   // process left child of join
	   //   if right child select than set oneScan to false
	   //   to form a separate row for the select
	   if (joinNode->rightChild->opCode != JOIN)
	       staticExec(joinNode->leftChild, schedule, waitSchedule, 
			  insertRow, lastRow, waitInsertRow, lastWaitRow, 
			  TRUE, runJoinTogether, memLeft, totalMemory);
	   else
	       staticExec(joinNode->leftChild, schedule, waitSchedule, 
			  insertRow, lastRow, waitInsertRow, lastWaitRow, 
			  oneScan, runJoinTogether, memLeft, totalMemory);

	  // add join node (BUILD) to both schedules
	  addOpr2(insertRow, joinNode, 1, BUILD);
	  addOpr2(waitInsertRow, joinNode, 1, BUILD);
	    
	  // if new rows before original rows in schedule, restore rows
	  if (after(insertRow, origInsertRow, lastRow )) 
	      insertRow = origInsertRow;
	  if (after(waitInsertRow,origWaitInsertRow, lastWaitRow)) 
	      waitInsertRow = origWaitInsertRow;
      }
      // add join node (PROBE) to both schedules
      addOpr2(insertRow, joinNode, 1, PROBE);
      addOpr2(waitInsertRow, joinNode, 1, PROBE);
      break;
  }
}

